#!/usr/bin/perl
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# This Source Code Form is "Incompatible With Secondary Licenses", as
# defined by the Mozilla Public License, v. 2.0.

# This script compiles all the documentation.
#
# Required software:
#
# 1) Sphinx documentation builder (python-sphinx package on Debian/Ubuntu)
#
# 2a) rst2pdf
# or
# 2b) pdflatex, which means the following Debian/Ubuntu packages:
#     * texlive-latex-base
#     * texlive-latex-recommended
#     * texlive-latex-extra
#     * texlive-fonts-recommended
#
# All these TeX packages together are close to a gig :-| But after you've
# installed them, you can remove texlive-latex-extra-doc to save 400MB.

use 5.10.1;
use strict;
use warnings;

use File::Basename;
BEGIN { chdir dirname($0); }

use lib qw(.. ../lib lib);
use open ':encoding(utf8)';

use Cwd;
use File::Copy::Recursive qw(rcopy);
use File::Find;
use File::Path qw(rmtree make_path);
use File::Which qw(which);
use Pod::Simple;
use Pod::ParseLink;

use Bugzilla::Constants qw(BUGZILLA_VERSION bz_locations);

use Pod::Simple::Search;
use Pod::POM::View::Restructured;
use File::Spec;
use Tie::File;

###############################################################################
# Subs
###############################################################################

my $error_found = 0;
sub MakeDocs {
  my ($name, $cmdline) = @_;

  my @args;

  if ($name eq 'HTML') {
    foreach my $fav (grep { $_ =~ /favicon/ } @ARGV) {
      push(@args, qq{SPHINXOPTS='-Dhtml_favicon=../../../$fav'});
    }
  }

  push(@args, $cmdline);

  say "Creating $name documentation ..." if defined $name;
  system('make', @args) == 0 or $error_found = 1;
  print "\n";
}

# Put all the POD in to the API docs
# Duplicate some docs to user guide, admin guide, and web service docs
sub pod2rst {
    my $path = shift;
    my %callbacks = ();#'link' => \&links,);

    say "Converting POD to RST...";
    my $name2path = Pod::Simple::Search->new->inc(0)->verbose(0)->survey(@{['../']});

    my $ind_path = "$path/rst/integrating/api/";
    rmtree($ind_path);
    make_path($ind_path);

    my $FILE;
    open($FILE, '>', "$ind_path/index.rst") || die ("Can't open rst file");
    print($FILE <<'EOI');
.. highlight:: perl

.. _developer

===================================
API Documentation
===================================

You may end hacking the code, here is the API.

.. toctree::
   :maxdepth: 2

EOI

    make_path($ind_path."Bugzilla/Extension");
    my $adminfile   = $ind_path."Bugzilla/Extension/index-admin.rst";
    my $userfile    = $ind_path."Bugzilla/Extension/index-user.rst";
    my $webfile     = $ind_path."Bugzilla/Extension/index-web.rst";
    my $paramsfile  = $ind_path."Bugzilla/Extension/index-params.rst";
    my ($ADMIN, $USER, $WEB, $PARAMS);
    open($ADMIN, ">:utf8", "$adminfile")    || die "$adminfile: $!";
    open($USER, ">:utf8", "$userfile")     || die "$userfile: $!";
    open($WEB, ">:utf8", "$webfile")       || die "$webfile: $!";
    open($PARAMS, ">:utf8", "$paramsfile") || die "$paramsfile: $!";

    foreach my $mod (sort keys %$name2path) {
        if ($mod =~ /^lib::/) {
            say "Deleting %mod";
            delete $name2path->{$mod};
            next;
        }

        if($mod =~ m{extensions::[^:]*::Config}){
            next;
        }

        my $title = $mod;

        my $ext = $title =~ s/^extensions::/Bugzilla::Extension::/;
        $title =~ s/lib:://;
        my $header = "=" x length($title);
        my $module;
        my $abs_path = $name2path->{$mod};
        my $fpath = File::Spec->abs2rel($abs_path, '../');
        my ($volume,$directories,$file) = File::Spec->splitpath( $fpath );
        $directories =~ s/lib\///;
        $directories = $title;
        $directories =~ s{::[^:]+$}{};
        $directories =~ s{::}{/}g;
        my $dir_path = "$ind_path$directories/";
        make_path($dir_path);
        $file =~ s/\.[^.]*$//;
        my $out_file = $file;

        if($ext && $file eq 'Extension') {
            $directories =~ m{/([^/]+$)};
            $out_file = $1;
            $module = $out_file;
            $directories =~ s{/[^/]+$}{};
            $dir_path = "$ind_path$directories/";
            $title =~ s/::Extension$//;
        } elsif($ext) {
            $directories =~ m{Bugzilla/Extension/([^/]+)};
            $module = $1;
        } elsif($title eq 'Bugzilla') {
            $directories = '';
            $dir_path = "$ind_path";
        }
        $out_file .= '.rst';

        my $callbacks = { link => sub {
                my $text = shift;

                my @results = parselink($text);

                # A link to another page.
                if(defined($results[1]) && defined($results[2])){
                    if($results[2] =~ /^http/){
                        return ($results[2], $results[1]);
                    }
                    elsif($results[2] =~ m/Bugzilla/){
                        my $depth = scalar(split('/',$directories));
                        my @split = split('::',$results[2]);
                        my $base = '../' x $depth;
                        my $url = $base . join('/',@split);
                        $url .= '.html';
                        $url .= '#' . $results[3] if defined $results[3];
                        return ($url, $results[1]);
                    }
                    else {
                        # Not a Bugzilla package, link to CPAN.
                        my $url = 'http://search.cpan.org/search?query=' . $results[2] . '&mode=all';
                        return ($url, $results[1]);
                    }
                }
                # A Link within a page
                elsif(defined($results[1]) && defined($results[4])) {
                    my $anchor = $results[1];
                    $anchor =~ s/"//g;
                    return ("#$anchor", $anchor);
                }
                else {
                    die "Don't know how to parse $text";
                }
            }
        };
        my $conv = Pod::POM::View::Restructured->new({namespace => $title,callbacks=>$callbacks});
        my $OUT;
        open($OUT, ">:utf8", "$dir_path$out_file") || die "$dir_path$out_file: $!";
        my $rv = $conv->convert_file($abs_path, $title, $OUT, \%callbacks);
        close($OUT);
        if($directories ne '') {
            print($FILE "   $directories/$out_file\n");
        } else {
            print($FILE "   $out_file\n");
        }

        if($ext) {
            my $api_path  = $path     . "/rst/integrating/api/$directories/";
            my $apifile   = $api_path . "api/v1/index.rst";
            make_path($api_path) unless -d $api_path;

            # Add core doc to user & admin guides
            if($file eq 'Extension') {
                say $ADMIN qq{* :ref:\`$module <$title>\`};
                say $USER qq{* :ref:\`$module <$title>\`};
            }
            # Add conf doc to admin guide
            elsif($file =~ m{^(:?Config|Params)$}) {
                say $PARAMS qq{* :ref:\`$module <$title>\`};
            }
            # Add WebServices doc to API docs
            elsif($title =~ m{WebService}) {
                my $short_title = $title;
                $short_title =~ s/Bugzilla::Extension:://;
                $short_title =~ s/::WebService//;
                $short_title =~ s/::/./g;
                say $WEB qq{* :ref:\`$short_title <$title>\`};
            } 
        }
    }

    close($FILE);
    close($ADMIN);
    close($USER);
    close($WEB);
    close($PARAMS);
}


###############################################################################
# Make the docs ...
###############################################################################

my @langs;
# search for sub directories which have a 'rst' sub-directory
opendir(LANGS, './');
foreach my $dir (readdir(LANGS)) {
    next if (($dir eq '.') || ($dir eq '..') || (! -d $dir));
    if (-d "$dir/rst") {
        push(@langs, $dir);
    }
}
closedir(LANGS);

my $docparent = getcwd();
my $ext_api_path = 'rst/integrating/api/Bugzilla/Extension';
foreach my $lang (@langs) {

    rmtree("$lang/html", 0, 1);
    rmtree("$lang/txt", 0, 1);
    # Collect up local extension documentation into the extensions/ dir.
    rmtree("$lang/$ext_api_path", 0, 1);
    pod2rst("$docparent/$lang");

    next if grep { $_ eq '--pod-only' } @ARGV;

    # Generate extension documentation, both normal and API
    my $ext_dir = bz_locations()->{'extensionsdir'};
    my @ext_paths = grep { $_ !~ /\/create\.pl$/ && ! -e "$_/disabled" }
                    glob("$ext_dir/*");
    my %extensions;
    foreach my $item (@ext_paths) {
        my $basename = basename($item);
        if (-d "$item/docs/$lang/rst") {
            $extensions{$basename} = "$item/docs/$lang/rst";
        }
    }


    foreach my $ext_name (keys %extensions) {
        my $src = $extensions{$ext_name} . "/*";
        my $dst = "$docparent/$lang/$ext_api_path/$ext_name";
        mkdir($dst) unless -d $dst;
        rcopy($src, $dst);
    }

    chdir "$docparent/$lang";

    MakeDocs('HTML', 'html');
    MakeDocs('TXT', 'text');

    if (grep { $_ eq '--with-pdf' } @ARGV) {
        if (which('pdflatex')) {
            MakeDocs('PDF', 'latexpdf');
        }
        elsif (which('rst2pdf')) {
            rmtree('pdf', 0, 1);
            MakeDocs('PDF', 'pdf');
        }
        else {
            say 'pdflatex or rst2pdf not found. Skipping PDF file creation';
        }
    }

    rmtree('doctrees', 0, 1);
}

die "Error occurred building the documentation\n" if $error_found;
