# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# This Source Code Form is "Incompatible With Secondary Licenses", as
# defined by the Mozilla Public License, v. 2.0.

package Bugzilla::Extension::AntiSpam::Config;

use strict;
use warnings;
use 5.10.1;

use Bugzilla::Config::Common;
use Bugzilla::Group;

our $sortkey = 511;

sub get_param_list {
  my ($class) = @_;

  my @param_list = (
    {
      name    => 'antispam_spammer_exclude_group',
      type    => 's',
      choices => \&_get_all_group_names,
      default => 'editbugs',
      checker => \&check_group
    },
    {
      name    => 'antispam_spammer_comment_count',
      type    => 't',
      default => '3',
      checker => \&check_numeric
    },
    {
      name    => 'antispam_spammer_disable_text',
      type    => 'l',
      default => "This account has been automatically disabled as a result of "
        . "a high number of spam comments.<br>\n<br>\n"
        . "Please contact the address at the end of this message if "
        . "you believe this to be an error."
    },
    {
      name    => 'antispam_abusive_comment_count',
      type    => 't',
      default => '5',
      checker => \&check_numeric
    },
    {
      name    => 'antispam_abusive_disable_text',
      type    => 'l',
      default => "This account has been automatically disabled as a result of "
        . "a high number of comments tagged as abusive.<br>\n<br>\n"
        . "Please contact the address at the end of this message if you "
        . "believe this to be an error, or if you would like your account "
        . "reactivated in order to interact within our etiquette "
        . "guidelines."
    },
    {
      name    => 'antispam_multi_user_limit_age',
      type    => 't',
      default => '2',
      checker => \&check_numeric,
    },
    {
      name    => 'antispam_multi_user_limit_count',
      type    => 't',
      default => '5',
      checker => \&check_numeric,
    },
  );

  return @param_list;
}

sub _get_all_group_names {
  my @group_names = map { $_->name } Bugzilla::Group->get_all;
  unshift(@group_names, '');
  return \@group_names;
}

1;

__END__

=head1 Description

Bugzilla::Extension::AntiSpam::Config - A module for specifying parameters to be added to the data/params.json file.

=head1 Parameters

=over 4

=item antispam_spammer_exclude_group

Users in this group will be excluded from automatic disabling.

=item antispam_spammer_comment_count

If a user has made at least this many comments, and either their first
NNN comments or their last NNN comments have been tagged as 'spam', their
account will be automatically disabled.

=item antispam_spammer_disable_text

This message will be displayed to the user when they try to log
in after their account is disabled due to spam.

=item antispam_abusive_comment_count

If a user has made at least this many comments, and either their first
"NNN comments or their last NNN comments have been tagged as 'abusive' or
"'abuse', their account will be automatically disabled.

=item antispam_abusive_disable_text

This message will be displayed to the user when they try to log
in after their account is disabled due to abuse.

=item antispam_multi_user_limit_age

Accounts created fewer than this many days are restricted to CCing/flag/etc
'antispam_multi_user_limit_count' users at once.

=item antispam_multi_user_limit_count

The maximum number of users new accounts can CC/flag/etc at once.  Attempts to
use more than this many users will result in zero users being set.

=item antispam_spamhaus_domain

The domain for a spamhaus DNS server.  Users will be unable to create an account if they are on this list.

=item honeypot_api_key

API Key for L<http://www.projecthoneypot.org>.

=item honeypot_threat_threshold

Users will be unable to create accounts if their honeypot threat score is this value or higher.

=back

