# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;

package Bugzilla::Extension::DependentProducts::Contact;

use Scalar::Util qw(blessed);

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::User;

use base qw(Bugzilla::Object);

use Bugzilla::Extension::DependentProducts::Util;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'dependent_products_contacts';
use constant DB_COLUMNS => qw(
  id
  record_id
  contact
  contact_role
);

use constant UPDATE_COLUMNS => qw(
  record_id
  contact
  contact_role
);

use constant LIST_ORDER => 'id';

use constant VALIDATORS => {
  record_id    => \&_check_record_id,
  contact      => \&_check_contact,
  contact_role => \&_check_contact_role,
};

###############################
####      Validators       ####
###############################

sub _check_record_id {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->record_id : 0;
  $value = trim($value);

  Bugzilla::Extension::DependentProducts::Record->check({id => $value});
  return $value;
}

sub _check_contact {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->contact : 0;
  $value = trim($value);

  my $object = Bugzilla::User->check({id => $value});
  if ((!$old_id || $object->id != $old_id) && !$object->is_enabled) {
    ThrowUserError('value_inactive',
      {class => ref($object), value => $object->name});
  }
  return $object->id;
}

sub _check_contact_role {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->contact_role : 0;
  $value = trim($value);

  if ( $value != DP_CONTACT
    && $value != DP_PROD_MGR
    && $value != DP_DEV_MGR
    && $value != DP_QE_MGR)
  {
    ThrowUserError('invalid_role', {value => $value});
  }
  return $value;
}

###############################
####       Methods         ####
###############################

sub set_record_id    { $_[0]->set('record_id',    $_[1]); return; }
sub set_contact      { $_[0]->set('contact',      $_[1]); return; }
sub set_contact_role { $_[0]->set('contact_role', $_[1]); return; }

###############################
####      Accessors        ####
###############################

sub record_id    { return $_[0]->{record_id}; }
sub contact      { return $_[0]->{contact}; }
sub contact_role { return $_[0]->{contact_role}; }

sub contact_email {
  my $self = shift;
  if (!defined $self->{contact_email}) {
    $self->{contact_email} = Bugzilla::User->new($self->{contact})->email;
  }
  return ($self->{contact_email});
}

1;

__END__

=head1 Description

Bugzilla::Extension::DependentProducts::Contact - This modules maps users to roles on
dependent products. Contacts will be emailed changes when bugs are flagged as
affecting a dependent product.

=head1 Fields

=over 4

=item record_id

An FK to the L<Bugzilla::Extension::DependentProducts::Record> for this contact.

=item contact

An FK to the L<Bugzilla::User> to contact.

=item contact_role

The roel the contact is playing for this contact.

=over 4

=item DP_CONTACT

Contact is the default contact for an affecting product.

=item DP_PROD_MGR

Contact is the Product Manager for the dependent product.

=item DP_DEV_MGR

Contact is the Development Manager for the dependent product.

=item DP_QE_MGR

Contact is the Quality Manager for the dependent product.

=back

=back

=head1 Methods

These methods allow you to set the specified field.

=over 4

=item set_record_id

=item set_contact

=item set_contact_role


=back

=head1 Accessors

These methods allow you to get the specified field.

=over 4

=item record_id

=item contact

=item contact_role

=item contact_email

=back


=cut

