# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::DependentProducts::WebService;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::WebService);

use Bugzilla::Product;
use Bugzilla::Component;
use Bugzilla::User;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trim diff_arrays);
use Bugzilla::WebService::Util qw(validate);
use List::Util qw(any);

use constant PUBLIC_METHODS => qw(
  update
  get
);


# Update a dependent product mapping.
sub update {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "products"});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;
  my $products = $params->{products};

  my $changes = [];

  foreach my $p (@$products) {
    my $product;
    if (defined($p->{id})) {
      $product = Bugzilla::Product->check(int $p->{id});
    }
    elsif (defined($p->{name})) {
      $product = Bugzilla::Product->check({name => $p->{name}});
    }
    else {
      ThrowCodeError('params_required',
        {function => 'DependentProducts.update', params => [qw(id name)]});
    }

    # Make sure the user can admin this product.
    $user->check_can_admin_product($product->name);

    my $product_changes = {};

    my $dep_product = $product->dependent_product;

    # Update the contact details
    foreach my $type (qw(qe_contacts dm_contacts pm_contacts)) {

      # Skip if absent.
      my $h = $p->{$type} // next;

      my $contact_changes = {};
      my $c_add           = [];
      my $c_remove        = [];

      # Getter and setter functions on the dependent product record object.
      my $set_func = "set_$type";
      my $get_func = "${type}_as_email";

      # Apply a list of contacts, replacing what was already there.
      if (defined($h->{set})) {

        # Get the existing list.
        my $emails = $dep_product->$get_func();

        # Set the new names
        $dep_product->$set_func(join(',', @{$h->{set}}));

        # Build the list of changes
        my ($removed, $added) = diff_arrays($emails, $h->{set});
        push @$c_add,    @$added;
        push @$c_remove, @$removed;
      }
      else {
        # Add some new contacts to the list
        if (defined($h->{add})) {
          my $contacts = $dep_product->$get_func // [];
          push @$contacts, @{$h->{add}};
          push @$c_add,    @{$h->{add}};
          $dep_product->$set_func(join(',', @$contacts));
        }

        # Remove existing contacts
        if (defined($h->{remove})) {
          my $contacts = $dep_product->$get_func // [];
          my @users_keep;
          my @users_remove;
          foreach my $c (@$contacts) {
            if (any { $_ eq $c } @{$h->{remove}}) {
              push @users_remove, $c;
            }
            else {
              push @users_keep, $c;
            }
          }
          push @$c_remove, @users_remove;
          $dep_product->$set_func(join(',', @users_keep));
        }
      }
      $contact_changes->{added}   = $c_add           if scalar @$c_add;
      $contact_changes->{removed} = $c_remove        if scalar @$c_remove;
      $product_changes->{$type}   = $contact_changes if scalar keys %$contact_changes;
    }

    if (defined($p->{depends_on})) {
      my $dep_changes = {};
      my $ch_added    = [];
      my $ch_removed  = [];

      my $c = $p->{depends_on};
      if (defined($c->{set})) {

        # The current list of provider products.
        my $pro_products = $product->provider_products;

        my @product_list = map { Bugzilla::Product->check({name => $_}) } @{$c->{set}};

        # Take the existing and new lists, work out what has changed.
        my ($removed, $added) = diff_arrays($pro_products, \@product_list);

        foreach my $rem (@$removed) {
          _remove_dependent_product($product->id, $rem->id);
          push @$ch_removed, $rem->name;
        }
        foreach my $add (@$added) {
          _add_dependent_product($product->id, $add->id);
          push @$ch_added, $add->name;
        }
      }
      else {
        if (defined($c->{add})) {
          foreach my $name (@{$c->{add}}) {
            my $p = Bugzilla::Product->check({name => $name});
            _add_dependent_product($product->id, $p->id);
            push @$ch_added, $p->name;
          }
        }
        if (defined($c->{remove})) {
          foreach my $name (@{$c->{remove}}) {
            my $p = Bugzilla::Product->check({name => $name});
            _remove_dependent_product($product->id, $p->id);
            push @$ch_removed, $p->name;
          }
        }
      }

      $dep_changes->{added}          = $ch_added    if scalar @$ch_added;
      $dep_changes->{removed}        = $ch_removed  if scalar @$ch_removed;
      $product_changes->{depends_on} = $dep_changes if scalar keys %$dep_changes;
    }

    $dep_product->update;

    # If we made any changes, put them in the change log.
    if (scalar keys %$product_changes) {
      $product_changes->{name} = $product->name;
      $product_changes->{id}   = $product->id;
      push @$changes, $product_changes;
    }
  }

  $dbh->bz_commit_transaction;

  return $changes;
}

sub _add_dependent_product {
  my ($product_id, $depends_on_id) = @_;

  my $entry = Bugzilla::Extension::DependentProducts::Record->create(
    {product_id => $product_id, depends_on => $depends_on_id});
  $entry->update;

  return;
}

sub _remove_dependent_product {
  my ($product_id, $depends_on_id) = @_;

  Bugzilla->dbh->do(
    'DELETE FROM dependent_products WHERE product_id = ? AND depends_on = ?',
    undef, $product_id, $depends_on_id);

  return;
}

# Params should be a list of product names.
# This will return a list of products that this product depends on.

# Users will need to call this to get a list of valid products that can be
# used as dependent products.
sub get {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $names = $params->{names} // [];
  my $ids   = $params->{ids}   // [];

  my @products;

  push @products, map { Bugzilla::Product->check({name => $_}) } @$names;
  push @products, map { Bugzilla::Product->check($_) } @$ids;

  my @results;

  foreach my $product (@products) {

    # Skip this product if the user is not allowed to see it.
    next unless $user->can_see_product($product->name);

    # The list of products we depend upon.
    my $dep_products = $product->provider_products;

    # The list of dependent products that the user is allowed to see
    my $vis_dep_products
      = [grep { $user->can_see_product($_) } map { $_->name } @$dep_products];

    my $result
      = {name => $product->name, id => $product->id, products => $vis_dep_products,
      };

    # If the user can admin the product, return contact information
    if ($user->in_group('editcomponents', $product->id)) {
      my $d_product = $product->dependent_product;
      if (defined($d_product)) {
        $result->{dm_contacts} = $d_product->dm_contacts_as_email;
        $result->{pm_contacts} = $d_product->pm_contacts_as_email;
        $result->{qe_contacts} = $d_product->qe_contacts_as_email;
      }
    }

    push @results, $result;
  }

  return \@results;
}

1;


__END__

=head1 NAME

Bugzilla::Extension::DependentProducts::WebService - The API for managing a products
dependent products and contacts

=head1 DESCRIPTION

This API allows you to change the qe, dev and product manager contacts, as well as add and remove
products that depend on another product.

=head1 METHODS

=over

=item B<Name>

DependentProducts.get

=item B<Description>

Returns a list of products that the given product depends on.  Also returns the qe, pm
and dev manager contacts for the given product.

=item B<Params>

=over

=item C<ids>

An array of product ids.

=item C<names>

An array of product names.

=back

=item B<Returns>

An array of hashes is returned. Each hash represents one product that was requested.

The arguments to each hash are explained below.

=over

=item C<name>

The name of the product you requested

=item C<id>

The ID of the product you requested.

=item C<products>

An array of product names.  These products are dependencies of the requested product.

=item C<dm_contacts>

en array of email addresses. Development Manager contact.
You must be in the 'editcomponents' group to view this.

=item C<pm_contacts>

An array of email addresses. Product Manager contacts.
You must be in the 'editcomponents' group to view this.

=item C<qe_contacts>

An array of email addresses. Quality Engineering Manager contacts.
You must be in the 'editcomponents' group to view this.

=back

=back

=over

=item B<Name>

DependentProducts.update

=item B<Description>

Allows the dependent product information to be updated

=item B<Params>

Update takes a hash with a single key named 'products'.  this key contains an array of hashes.

The contents of each hash is described below.

=over

=item C<id>

The ID of the product you are updating.  One of I<id> or I<name> must be supplied

=item C<name>

The name of the product you are updating.  One of I<id> or I<name> must be supplied

=item C<pm_contacts>

Product Manager email addresses.

A hash containing one or more of the keys 'add' 'remove' 'set'. The contents of each are described below.

=item C<dm_contacts>

Development Manager email addresses.

A hash containing one or more of the keys 'add' 'remove' 'set'. The contents of each are described below.

=item C<qe_contacts>

Quality Engineering Manager email addresses.

A hash containing one or more of the keys 'add' 'remove' 'set'. The contents of each are described below.

=over

=item C<add>

An array of email addresses.  These addresses will be appended to the existing list of contacts

=item C<remove>

An array of email addresses.  These addresses will be removed from the existing list of contacts

=item C<set>

An array of email addresses.  These addresses will replace the exsting list of contacts

=back

=item C<depends_on>

The products that depend on this product.

A hash containing one or more of the keys 'add' 'remove' 'set'.  The keys are described below

=over

=item C<add>

A list of product ids or names that will be appended to the existing list of dependencies.

=item C<remove>

A list of product ids or names that will be removed from the existing list of dependencies.

=item C<set>

A list of product ids or names that will replace the existing list of dependencies.

=back

=back

=item B<Returns>

Returns an array of hashes describing the changes made to each product.

The contents of each hash is described below

=over

=item C<id>

The ID of the product that was updated.

=item C<name>

The name of the product that was updated.

=item C<dm_contacts>

A hash containing changes to the development manager contacts.  The format is described below.

=item C<pm_contacts>

A hash containing changes to the product manager contacts.  The format is described below.

=item C<qe_contacts>

A hash containing changes to the QE manager contacts.  The format is described below.

=over

=item C<added>

A list of email addresses that were added to the existing list of contacts.

=item C<removed>

A list of email addresses that was removed from the existing list of contacts.

=back

=back

=back
