# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension);

use Bugzilla::Error;
use Bugzilla::Constants;
use Bugzilla::Token;
use Bugzilla::Util;

use Bugzilla::Extension::ExternalBugs::Bug;
use Bugzilla::Extension::ExternalBugs::Type;
use Bugzilla::Extension::ExternalBugs::Regex;
use Bugzilla::Extension::ExternalBugs::Util qw(resolve_ext_tracker_url);

use Clone qw(clone);
use Scalar::Util qw(tainted blessed looks_like_number);
use List::MoreUtils qw(any none);

our $VERSION = 0.3;

use constant LOCALCONFIG_VARS => ({name => 'ext_logins', default => [],},);

BEGIN {
  *Bugzilla::Bug::add_external_bug    = \&_add_external_bug;
  *Bugzilla::Bug::external_bugs       = \&_external_bugs;
  *Bugzilla::Bug::external_choices    = \&_external_choices;
  *Bugzilla::Bug::internal_links      = \&_internal_links;
  *Bugzilla::Bug::remove_external_bug = \&_remove_external_bug;
  *Bugzilla::User::can_see_ext_bz     = \&_user_can_see_ext_bz;
}

sub _external_bugs {
  my $self = shift;

  if (not defined $self->{external_bugs}) {
    $self->{external_bugs}
      = Bugzilla::Extension::ExternalBugs::Bug->match({bug_id => $self->id});
  }

  return $self->{external_bugs};
}

sub _internal_links {
  my $self = shift;

  if (not defined $self->{internal_links}) {
    my $ext_bz;
    if (Bugzilla->params->{'internal_link_name'}) {
      $ext_bz = Bugzilla::Extension::ExternalBugs::Type->match(
        {description => Bugzilla->params->{'internal_link_name'}});
    }

    return unless (@$ext_bz);

    $self->{internal_links} = Bugzilla::Extension::ExternalBugs::Bug->match(
      {ext_bz_id => $ext_bz->[0]->id, ext_bz_bug_id => $self->id});
  }

  return $self->{internal_links};
}

sub _external_choices {
  my $self = shift;

  if (not defined $self->{external_choices}) {
    $self->{external_choices} = Bugzilla::Extension::ExternalBugs::Type->match();
  }

  return $self->{external_choices};
}

## REDHAT EXTENSION START 1087696
sub _add_external_bug {
  my ($self, $ext_bz) = @_;
  my $current_bugs = $self->external_bugs;

  my $privs = 0;
  $self->check_can_change_field('external_bugs', 0, 1, \$privs)
    || ThrowUserError('illegal_change',
    {field => 'external_bugs', privs => $privs});

  if (defined($ext_bz->{ext_bz_bug_url}) && length($ext_bz->{ext_bz_bug_url})) {
    my $bug_url = $ext_bz->{ext_bz_bug_url};

    my ($tracker_id, $tracker_url) = resolve_ext_tracker_url($bug_url);
    $ext_bz->{ext_bz_id}     = $tracker_id;
    $ext_bz->{ext_bz_bug_id} = $tracker_url;

    # Need to delete this URL or creating the tracker object will fail.
    delete $ext_bz->{ext_bz_bug_url};
  }
  else {
    my $ext_bz_id = $ext_bz->{ext_bz_id};
    detaint_natural($ext_bz_id);
    $ext_bz->{ext_bz_id} = $ext_bz_id;
  }

  # Is the user allowed to add this type
  if (not Bugzilla->user->can_see_ext_bz($ext_bz->{ext_bz_id})) {
    ThrowUserError('external_bug_id_not_valid');
  }

  # Don't do anything if the external bug already exists for this bug
  return if any {
    $_->ext_bz_id == $ext_bz->{ext_bz_id}
      && $_->ext_bz_bug_id eq trim($ext_bz->{ext_bz_bug_id})
  }
  @$current_bugs;

  # Don't do anything if the external bug already added for this bug
  return if any {
    $_->{ext_bz}->{ext_bz_id} == $ext_bz->{ext_bz_id}
      && $_->{ext_bz}->{ext_bz_bug_id} eq trim($ext_bz->{ext_bz_bug_id})
  }
  @{$self->{_added_external_bugs}};

  # Save the external bug details so we can do the create later.
  push @{$self->{_added_external_bugs}}, {bug_id => $self->id, ext_bz => $ext_bz};

  return;
}

sub _remove_external_bug {
  my ($self, $ext_bz_obj) = @_;

  unless (blessed($ext_bz_obj)) {
    $ext_bz_obj = Bugzilla::Extension::ExternalBugs::Bug->new($ext_bz_obj);
  }

  # Nothing to delete?
  return if !$ext_bz_obj;

  my $privs = 0;
  $self->check_can_change_field('external_bugs', 0, 1, \$privs)
    || ThrowUserError('illegal_change',
    {field => 'external_bugs', privs => $privs});

  # Is the user allowed to add this type
  if (not Bugzilla->user->can_see_ext_bz($ext_bz_obj->ext_bz_id)) {
    ThrowUserError('external_bug_id_not_valid');
  }

  # Save the details of the bug we are removing so we can process it later
  push @{$self->{_removed_external_bugs}},
    {id => $self->id, ext_bz => $ext_bz_obj};

  return;
}
## REDHAT EXTENSION END 1087696

################
# Installation #
################

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  ## REDHAT EXTENSION START 406161
  $schema->{external_bugzilla} = {
    FIELDS => [
      id          => {TYPE => 'MEDIUMSERIAL',  NOTNULL => 1, PRIMARYKEY => 1},
      url         => {TYPE => 'varchar(2000)', NOTNULL => 1},
      description => {TYPE => 'varchar(2000)', NOTNULL => 1},
      full_url  => {TYPE => 'varchar(2000)'},
      type      => {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'None'"},
      can_get   => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'},
      can_send  => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'},
      must_send => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'},
      send_once => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'},
      ## REDHAT EXTENSION 1326979: change regex to an FK
      ext_bz_regex_id =>
        {TYPE => 'INT3', REFERENCES => {TABLE => 'ext_bz_regex', COLUMN => 'id'}},
      ## REDHAT EXTENSION START 1538429
      matchrx => {TYPE => 'varchar(64)'},
      format  => {TYPE => 'varchar(64)'},
      ## REDHAT EXTENSION END 1538429
    ],
  };

  $schema->{ext_bz_bug_map} = {
    FIELDS => [
      id     => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE'}
      },
      ext_bz_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'external_bugzilla', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      ext_bz_bug_id   => {TYPE => 'varchar(64)',   NOTNULL => 1},
      ext_status      => {TYPE => 'varchar(64)',   NOTNULL => 1, DEFAULT => "'None'"},
      ext_priority    => {TYPE => 'varchar(64)',   NOTNULL => 1, DEFAULT => "'None'"},
      ext_description => {TYPE => 'varchar(2000)', NOTNULL => 1, DEFAULT => "'None'"},
      ext_last_updated => {TYPE => 'DATETIME'}

    ],
    INDEXES => [
      ext_bz_bug_map_bug_id_idx =>
        {FIELDS => ['bug_id', 'ext_bz_id', 'ext_bz_bug_id'], TYPE => 'UNIQUE'},
      ext_bz_bug_map_ext_bz_bug_id_idx => ['ext_bz_bug_id'],
      ext_bz_bug_map_ext_bz_id_idx     => ['ext_bz_id'],
    ],
  };
  ## REDHAT EXTENSION END 406161
  ## REDHAT EXTENSION START 821293
  $schema->{ext_bz_group_map} = {
    FIELDS => [
      ext_bz_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'external_bugzilla', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      ext_bz_group_map_ext_bz_id_idx =>
        {FIELDS => [qw(ext_bz_id group_id)], TYPE => 'UNIQUE'},
      ext_bz_group_map_group_id_idx => ['group_id'],
    ],
  };
  ## REDHAT EXTENSION END 821293

  ## REDHAT EXTENSION START 1326979
  $schema->{ext_bz_regex} = {
    FIELDS => [
      id          => {TYPE => 'MEDIUMSERIAL',  NOTNULL => 1, PRIMARYKEY => 1},
      name        => {TYPE => 'varchar(2000)', NOTNULL => 1},
      regex       => {TYPE => 'varchar(2000)', NOTNULL => 1},
      description => {TYPE => 'varchar(2000)', NOTNULL => 1},
    ],
    INDEXES => [
      ext_bz_regex_name_idx        => {FIELDS => ['name'],        TYPE => 'UNIQUE',},
      ext_bz_regex_description_idx => {FIELDS => ['description'], TYPE => 'UNIQUE',},
    ],
  };

  ## REDHAT EXTENSION END 1326979

  return;
}

sub install_update_db {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;

  my $ext_bz = new Bugzilla::Field({'name' => 'external_bugzilla.description'});
  if (!$ext_bz) {
    Bugzilla::Field->create({
      name        => 'external_bugzilla.description',
      description => 'Link System Description',
      type        => FIELD_TYPE_UNKNOWN,
      mailhead    => 0,
      enter_bug   => 0,
      obsolete    => 0,
      custom      => 0,
      buglist     => 0,
    });
  }
  elsif ($ext_bz->description =~ /External/) {
    $ext_bz->set_description('Link System Description');
    $ext_bz->update();
  }

  my $ext_bz_bug
    = new Bugzilla::Field({'name' => 'ext_bz_bug_map.ext_bz_bug_id'});
  if (!$ext_bz_bug) {
    Bugzilla::Field->create({
      name        => 'ext_bz_bug_map.ext_bz_bug_id',
      description => 'Link ID',
      type        => FIELD_TYPE_UNKNOWN,
      mailhead    => 0,
      enter_bug   => 0,
      obsolete    => 0,
      custom      => 0,
      buglist     => 0,
    });
  }
  elsif ($ext_bz_bug->description =~ /External/) {
    $ext_bz_bug->set_description('Link ID');
    $ext_bz_bug->update();
  }


  my $ext_bz_status
    = new Bugzilla::Field({'name' => 'ext_bz_bug_map.ext_status'});
  if (!$ext_bz_status) {
    Bugzilla::Field->create({
      name        => 'ext_bz_bug_map.ext_status',
      description => 'Link Status',
      type        => FIELD_TYPE_UNKNOWN,
      mailhead    => 0,
      enter_bug   => 0,
      obsolete    => 0,
      custom      => 0,
      buglist     => 0,
    });
  }
  elsif ($ext_bz_status->description =~ /External/) {
    $ext_bz_status->set_description('Link Status');
    $ext_bz_status->update();
  }


  my $ext_bz_url = new Bugzilla::Field({'name' => 'external_bugzilla.url'});
  if (!$ext_bz_url) {
    Bugzilla::Field->create({
      name        => 'external_bugzilla.url',
      description => 'Link System URL',
      type        => FIELD_TYPE_UNKNOWN,
      mailhead    => 0,
      enter_bug   => 0,
      obsolete    => 0,
      custom      => 0,
      buglist     => 0,
    });
  }
  elsif ($ext_bz_url->description =~ /External/) {
    $ext_bz_url->set_description('Link System URL');
    $ext_bz_url->update();
  }

  $dbh->bz_add_column('ext_bz_bug_map', 'id',
    {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1});

  # dkl - 2010-08-02 Add status, description, type, and last_updated columns
  $dbh->bz_add_column('external_bugzilla', 'type',
    {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => 0});
  $dbh->bz_add_column('external_bugzilla', 'can_get',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'});
  $dbh->bz_add_column('external_bugzilla', 'can_send',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'});
  $dbh->bz_add_column('external_bugzilla', 'must_send',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'});
  $dbh->bz_add_column('external_bugzilla', 'send_once',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => '0'});
  $dbh->bz_add_column('ext_bz_bug_map', 'ext_status',
    {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'None'"});
  $dbh->bz_add_column('ext_bz_bug_map', 'ext_priority',
    {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'None'"});
  $dbh->bz_add_column('ext_bz_bug_map', 'ext_description',
    {TYPE => 'varchar(2000)', NOTNULL => 1, DEFAULT => "'None'"});
  $dbh->bz_add_column('ext_bz_bug_map', 'ext_last_updated', {TYPE => 'DATETIME'});

  ## REDHAT EXTENSION 1541816: drop ext_login and ext_password
  $dbh->bz_drop_column('external_bugzilla', 'ext_login');
  $dbh->bz_drop_column('external_bugzilla', 'ext_password');

  ## REDHAT EXTENSION 1136779 1496286 - Make the description unique.
  $dbh->bz_add_index('external_bugzilla', 'ext_bz_description_idx',
    {TYPE => 'UNIQUE', FIELDS => [qw(description)]});

  ## REDHAT EXTENSION 1538429 Make ETs pretty
  $dbh->bz_add_column('external_bugzilla', 'matchrx', {TYPE => 'varchar(64)'});
  $dbh->bz_add_column('external_bugzilla', 'format',  {TYPE => 'varchar(64)'});

  ## REDHAT EXTENSION 1775514 edit bug page speed
  $dbh->bz_add_index(
    'ext_bz_bug_map',
    'ext_bz_bug_map_ext_bz_bug_id_idx',
    {FIELDS => [qw(ext_bz_bug_id)]}
  );
  $dbh->bz_add_index(
    'ext_bz_bug_map',
    'ext_bz_bug_map_ext_bz_id_idx',
    {FIELDS => [qw(ext_bz_id)]}
  );

  # If we have an internal link and there are any unset values then set them all.
  my $int_ext_bz;
  if (Bugzilla->params->{'internal_link_name'}) {
    $int_ext_bz = Bugzilla::Extension::ExternalBugs::Type->match(
      {description => Bugzilla->params->{'internal_link_name'}});
  }

  if (@$int_ext_bz) {
    my $count = $dbh->selectrow_array(
      q{
        SELECT count(*)
        FROM ext_bz_bug_map
        WHERE ext_bz_id = ?
        AND ext_bz_bug_id ~ '^[0-9]+$'
        AND ext_last_updated is NULL
      }, undef, $int_ext_bz->[0]->id
    );

    if ($count) {
      my $sql = <<'EOS';
UPDATE ext_bz_bug_map
SET
  ext_status = b.bug_status,
  ext_description = b.short_desc,
  ext_priority = b.priority,
  ext_last_updated = NOW()
FROM
  ( SELECT bug_id, bug_status, short_desc, priority FROM bugs) AS b
WHERE ext_bz_id = ?
AND ext_bz_bug_id ~ '^[0-9]+$'
AND ext_bz_bug_id::int = b.bug_id
EOS

      $dbh->do($sql, undef, $int_ext_bz->[0]->id);
    }
  }

  return;
}

#########
# Pages #
#########

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};

  if ($page =~ m{^externalbugs/list\.}) {
    _page_list($vars);
  }
  elsif ($page =~ m{^externalbugs/edit\.}) {
    _page_edit($vars);
  }
  elsif ($page =~ m{^externalbugs/confirm-delete\.}) {
    _page_confirm_delete($vars);
  }
  elsif ($page =~ m{^externalbugs/regexs/list\.}) {
    _page_regexs_list($vars);
  }
  elsif ($page =~ m{^externalbugs/regexs/edit\.}) {
    _page_regexs_edit($vars);
  }
  elsif ($page =~ m{^externalbugs/regexs/confirm-delete\.}) {
    _page_regexs_confirm_delete($vars);
  }

  return;
}

sub _page_list {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  # often used variables
  my $token  = $input->{'token'};
  my $action = $input->{'action'} || '';
  my $id     = $input->{'tracker_id'};

  if ($action eq 'add') {
    check_token_data($token, 'add_external_tracker');

    my $args = {
      url             => $input->{url},
      description     => $input->{description},
      full_url        => $input->{full_url},
      type            => $input->{type},
      ext_bz_regex_id => $input->{regex},
      can_get         => $input->{can_get},
      can_send        => $input->{can_send},
      must_send       => $input->{must_send},
      send_once       => $input->{send_once},
      matchrx         => $input->{matchrx},
      format          => $input->{format},
    };

    if (defined($input->{groups})) {
      $args->{groups} = ref $input->{groups} ? $input->{groups} : [$input->{groups}];
    }

    my $tracker = Bugzilla::Extension::ExternalBugs::Type->create($args);

    delete_token($token);

    $vars->{'message'} = 'external_tracker_created';
    $vars->{'tracker'} = $tracker;
  }

  if ($action eq 'delete') {
    $id || ThrowCodeError('param_required', {param => 'tracker_id'});

    my $tracker = Bugzilla::Extension::ExternalBugs::Type->new($id);

    check_token_data($token, 'delete_external_tracker');
    $tracker->remove_from_db();
    delete_token($token);

    $vars->{'message'} = 'external_tracker_deleted';
    $vars->{'tracker'} = $tracker;
  }

  if ($action eq 'update') {
    $id || ThrowCodeError('param_required', {param => 'tracker_id'});

    my $tracker = Bugzilla::Extension::ExternalBugs::Type->new($id);
    check_token_data($token, 'edit_external_tracker');
    my $groups = [];
    if (defined $input->{groups}) {
      $groups = ref $input->{groups} ? $input->{groups} : [$input->{groups}];
    }
    $tracker->set_description($input->{'description'});
    $tracker->set_url($input->{'url'});
    $tracker->set_full_url($input->{'full_url'});
    $tracker->set_type($input->{'type'});
    $tracker->set_regex($input->{'regex'});
    $tracker->set_can_get($input->{'can_get'});
    $tracker->set_can_send($input->{'can_send'});
    $tracker->set_must_send($input->{'must_send'});
    $tracker->set_send_once($input->{'send_once'});
    $tracker->set_matchrx($input->{'matchrx'});
    $tracker->set_format($input->{'format'});
    $tracker->set_groups($groups);
    $tracker->update();
    delete_token($token);

    $vars->{'tracker'} = $tracker;
    $vars->{'message'} = 'external_tracker_updated';
  }

  $vars->{'trackers'} = Bugzilla::Extension::ExternalBugs::Type->match();
  $vars->{'regexs'}   = Bugzilla::Extension::ExternalBugs::Regex->match();

  return;
}

sub _page_edit {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  my $action = $input->{'action'} || '';

  $vars->{'types'}  = Bugzilla::Extension::ExternalBugs::Type->legal_types;
  $vars->{'groups'} = [Bugzilla::Group->get_all];
  $vars->{'regexs'}
    = Bugzilla->dbh->selectall_hashref('SELECT * FROM ext_bz_regex', 'id');

  if ($action eq 'new') {
    $vars->{'action'} = 'add';
    $vars->{'token'}  = issue_session_token('add_external_tracker');
    return;
  }

  my $id = $input->{'tracker_id'};

  $id || ThrowCodeError('param_required', {param => 'tracker_id'});

  $vars->{'action'}  = 'update';
  $vars->{'tracker'} = Bugzilla::Extension::ExternalBugs::Type->new($id);
  $vars->{'token'}   = issue_session_token('edit_external_tracker');

  return;
}

sub _page_confirm_delete {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  my $id = $input->{'tracker_id'};

  $id || ThrowCodeError('param_required', {param => 'tracker_id'});

  $vars->{'tracker'} = Bugzilla::Extension::ExternalBugs::Type->new($id);
  $vars->{'token'}   = issue_session_token('delete_external_tracker');

  return;
}

sub _page_regexs_list {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  # often used variables
  my $token    = $input->{'token'};
  my $action   = $input->{'action'};
  my $regex_id = $input->{'regex_id'};

  detaint_natural($regex_id);


  if ($action eq 'add') {
    check_token_data($token, 'add_external_tracker_regex');
    my $args = {
      name        => $input->{name},
      regex       => $input->{regex},
      description => $input->{description},
    };

    my $regex = Bugzilla::Extension::ExternalBugs::Regex->create($args);
    delete_token($token);

    $vars->{'message'} = 'external_tracker_regex_created';
    $vars->{'regex'}   = $regex;
  }
  elsif ($action eq 'update') {
    check_token_data($token, 'edit_external_tracker_regex');

    $regex_id || ThrowCodeError('param_required', {param => 'regex_id'});

    my $regex = Bugzilla::Extension::ExternalBugs::Regex->new($regex_id);
    $regex->set_name($input->{'name'});
    $regex->set_regex($input->{'regex'});
    $regex->set_description($input->{'description'});
    $regex->update();
    delete_token($token);
    $vars->{'message'} = 'external_tracker_regex_updated';
    $vars->{'regex'}   = $regex;
  }
  elsif ($action eq 'delete') {
    $regex_id || ThrowCodeError('param_required', {param => 'regex_id'});

    my $regex = Bugzilla::Extension::ExternalBugs::Regex->new($regex_id);

    check_token_data($token, 'delete_external_tracker_regex');
    $regex->remove_from_db();
    delete_token($token);

    $vars->{'message'} = 'external_tracker_regex_deleted';
    $vars->{'regex'}   = $regex;
  }

  $vars->{'regexs'} = Bugzilla::Extension::ExternalBugs::Regex->match();

  return;
}

sub _page_regexs_edit {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  my $action = $input->{'action'} || '';

  if ($action eq 'new') {
    $vars->{'action'} = 'add';
    $vars->{'token'}  = issue_session_token('add_external_tracker_regex');
    return;
  }

  my $regex_id = $input->{'regex_id'};
  detaint_natural($regex_id);
  $regex_id || ThrowCodeError('param_required', {param => 'regex_id'});
  $vars->{'regex'}  = Bugzilla::Extension::ExternalBugs::Regex->new($regex_id);
  $vars->{'action'} = 'update';
  $vars->{'token'}  = issue_session_token('edit_external_tracker_regex');

  return;
}

sub _page_regexs_confirm_delete {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->in_group('editcomponents')
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "externalbugs"});

  my $regex_id = $input->{'regex_id'};
  detaint_natural($regex_id);
  $regex_id || ThrowCodeError('param_required', {param => 'regex_id'});

  $vars->{'token'} = issue_session_token('delete_external_tracker_regex');
  $vars->{'regex'} = Bugzilla::Extension::ExternalBugs::Regex->new($regex_id);

  return;
}

##########
# Config #
##########

sub config {
  my ($self, $args) = @_;
  my $config = $args->{config};
  $config->{ExternalBugs} = "Bugzilla::Extension::ExternalBugs::Config";

  return;
}

sub config_add_panels {
  my ($self, $args) = @_;
  my $modules = $args->{panel_modules};
  $modules->{ExternalBugs} = "Bugzilla::Extension::ExternalBugs::Config";

  return;
}

#############
# Bug Hooks #
#############

sub object_end_of_update {
  my ($self, $args) = @_;

  my ($object, $old_object, $changes) = @$args{qw(object old_object changes)};

  return;
}

sub bug_before_create {
  my ($self, $args) = @_;

  my $params         = $args->{params};
  my $deleted_params = $args->{deleted_params};
  my @external_bugs  = ();

  my @external_ids = grep {/^external_id_\d+$/} keys %$params;
  foreach my $external_id (@external_ids) {
    my ($id) = ($external_id =~ /^external_id_(\d+)$/);
    unless ($params->{$external_id} && $params->{"external_bug_id_$id"}) {

      # Value is empty, just delete it
      delete $params->{$external_id};
      delete $params->{"external_bug_id_$id"};
    }
    else {
      # Check that a value is set
      if (not defined $params->{"external_bug_id_$id"}
        or $params->{"external_bug_id_$id"} eq '0')
      {
        ThrowUserError("external_bug_id_not_specified", {row => $id});
      }

      # Is the user allowed to add this type
      trick_taint($params->{$external_id});
      if (not Bugzilla->user->can_see_ext_bz($params->{$external_id})) {
        ThrowUserError('external_bug_id_not_valid');
      }

      push @external_bugs,
        {
        ext_bz_id     => delete($params->{$external_id}),
        ext_bz_bug_id => delete($params->{"external_bug_id_$id"})
        };
    }
  }

  my @external_urls = grep {/^external_bug_url_\d+$/} keys %$params;
  foreach my $url (@external_urls) {
    my $bug_url = delete($params->{$url});
    next unless ($bug_url && $bug_url ne '');
    my ($tracker_id, $ext_bug_id) = resolve_ext_tracker_url($bug_url);
    if (not Bugzilla->user->can_see_ext_bz($tracker_id)) {
      ThrowUserError('external_bug_id_not_valid');
    }

    push @external_bugs, {ext_bz_id => $tracker_id, ext_bz_bug_id => $ext_bug_id};
  }

  $deleted_params->{external_bugs} = \@external_bugs;

  return;
}

sub bug_end_of_create {
  my ($self, $args) = @_;

  my $bug    = $args->{bug};
  my $params = $args->{deleted_params};

  # Insert new external bugs. We know that the data has been verified
  # as part of the bug_before_create subroutine
  if (scalar @{$params->{external_bugs}}) {
    foreach my $external_bug (@{$params->{external_bugs}}) {
      my $paras = {
        ext_bz_id     => $external_bug->{ext_bz_id},
        ext_bz_bug_id => $external_bug->{ext_bz_bug_id},
        bug_id        => $bug->{bug_id},
      };

      _sync_internal_links($paras);

      my $new_ext_bug = Bugzilla::Extension::ExternalBugs::Bug->create($paras);

      push @{$bug->{external_bugs}}, $new_ext_bug;
    }
  }

  return;
}

sub _sync_internal_links {
  my ($args) = @_;

  my $int_ext_bz;
  if (Bugzilla->params->{'internal_link_name'}) {
    my $results = Bugzilla::Extension::ExternalBugs::Type->match(
      {description => Bugzilla->params->{'internal_link_name'}});
    $int_ext_bz = $results->[0];
  }

  if ($int_ext_bz && $int_ext_bz->id == $args->{ext_bz_id}) {
    my $int_bug = Bugzilla::Bug->check($args->{ext_bz_bug_id});
    $args->{ext_status}      = $int_bug->bug_status;
    $args->{ext_description} = $int_bug->short_desc;
    $args->{ext_priority}    = $int_bug->priority;
  }

  return;
}

sub _process_external_bug {
  my ($self, $args, $id) = @_;

  my $bug    = $args->{bug};
  my $params = $args->{params};
  my $input  = Bugzilla->input_params;

  my $external_id      = $input->{"external_id_$id"};
  my $external_bug_id  = $input->{"external_bug_id_$id"};
  my $external_bug_url = $input->{"external_bug_url_$id"};

  if ($external_bug_url) {
    push @{$params->{external_bug}{add}}, {ext_bz_bug_url => $external_bug_url,};
  }
  elsif ($external_id && $external_bug_id) {
    push @{$params->{external_bug}{add}},
      {ext_bz_id => $external_id, ext_bz_bug_id => $external_bug_id,};
  }

  return;
}

## REDHAT EXTENSION START 1087696
sub bug_before_set_all {
  my ($self, $args) = @_;

  my $bug    = $args->{bug};
  my $params = $args->{params};
  my $input  = Bugzilla->input_params;

  # This only is used for browsers
  if (Bugzilla->usage_mode == USAGE_MODE_BROWSER) {

    foreach my $key (keys %$input) {
      if ($key =~ m/^external_id_([0-9]+)$/) {
        $self->_process_external_bug($args, $1);
      }
    }

    foreach my $key (keys %{$input}) {
      if ($key =~ /^rem_ext_bz_bug_(\d+)$/ && $input->{$key}) {
        push @{$params->{external_bug}{remove}}, $1;
      }
    }
  }

  # This value can be set in other places (like WebService.pm)
  if (exists $params->{external_bug}) {
    $bug->_add_remove($params, 'external_bug');
  }

  return;
}

sub bug_start_of_update {
  my ($self, $args) = @_;

  my $bug       = $args->{bug};
  my $timestamp = $args->{timestamp};
  my $changes   = $args->{changes};

  # Bail out if no changes have been made
  return
    unless defined $bug->{_added_external_bugs}
    or defined $bug->{_removed_external_bugs};

  # Make a copy of the original state so we can generate the history
  my $original_ext_bugs = clone($bug->external_bugs);

  # Process any new bugs that have been added.
  foreach my $ext_bug (@{$bug->{_added_external_bugs}}) {
    my $params = {bug_id => $ext_bug->{bug_id}, %{$ext_bug->{ext_bz}},};

    _sync_internal_links($params);

    my $new_ext_bz = Bugzilla::Extension::ExternalBugs::Bug->create($params);

    push @{$bug->{external_bugs}}, $new_ext_bz;
  }

  # And process any bugs that have been removed.
  foreach my $ext_bug (@{$bug->{_removed_external_bugs}}) {
    my $ext_bz_obj = $ext_bug->{ext_bz};

    my $id = $ext_bz_obj->id;
    $ext_bz_obj->remove_from_db();

    # Update cache
    $bug->{external_bugs} = [grep { $_->id != $id } @{$bug->{external_bugs}}];

    # Add the bug to the (per request) cache, used for bug_send_changes
    Bugzilla->request_cache->{'external_bugs_removed'} ||= [0];
    push @{Bugzilla->request_cache->{'external_bugs_removed'}}, $ext_bz_obj;
  }

  # Calculate history.
  my ($removed_ext_bz, $added_ext_bz)
    = diff_arrays($original_ext_bugs, $bug->external_bugs, 'id');

  # If any changes were found, record it in the activity log
  if (scalar @$removed_ext_bz || scalar @$added_ext_bz) {
    $changes->{'ext_bz_bug_map.ext_bz_bug_id'} = [
      join(', ',
        map { $_->type->description . ' ' . $_->ext_bz_bug_id } @$removed_ext_bz),
      join(', ',
        map { $_->type->description . ' ' . $_->ext_bz_bug_id } @$added_ext_bz)
    ];
  }

  return;
}

sub bug_end_of_update {
  my ($self, $args) = @_;
  my $bug = $args->{'bug'};

  ## RED HAT EXTENSION 1390377 BEGIN
  if ($bug->resolution eq 'DUPLICATE') {
    my $dup_id = $bug->dup_id;
    my $dup_ext_bugs;

    # On scrubbed databases this can error, just ignore it. Bug 1600479
    eval {
      $dup_ext_bugs
        = Bugzilla::Extension::ExternalBugs::Bug->match({bug_id => $dup_id});
    };
    my @external_bugs;

    foreach my $ext_bug (@{$bug->external_bugs}) {
      my $ext_bz = new Bugzilla::Extension::ExternalBugs::Type($ext_bug->ext_bz_id);

      # ET name matches and link doesn't exist on original bug
      if (
        (
          $ext_bz->description =~ /^(?:Red Hat Customer Portal|Red Hat Knowledge Base)/
        )
        && (
          none {
            $_->ext_bz_id == $ext_bug->ext_bz_id
              and $_->ext_bz_bug_id eq $ext_bug->ext_bz_bug_id
          }
          @$dup_ext_bugs
        )
        )
      {
        my %fields = ();
        foreach
          my $key (qw/ext_status ext_description ext_priority ext_bz_bug_id ext_bz_id/)
        {
          $fields{$key} = $ext_bug->$key;
        }
        push @external_bugs, \%fields;
      }
    }

    if (@external_bugs) {

      # Add ETs to duped bug
      my $duped = Bugzilla::Bug->new($dup_id);
      $duped->set_all({external_bug => {add => \@external_bugs}});
      my $changes = $duped->update();

      # We don't need to do anything if nothing changed
      if (scalar keys %$changes) {
        $duped->send_changes();
      }
    }
  }
  ## RED HAT EXTENSION 1390377 END

  my $int_ext_bz;
  if (Bugzilla->params->{'internal_link_name'}) {
    $int_ext_bz = Bugzilla::Extension::ExternalBugs::Type->match(
      {description => Bugzilla->params->{'internal_link_name'}});
  }

  if (@$int_ext_bz) {
    my $dbh = Bugzilla->dbh;

    my $sql = <<EOS;
UPDATE ext_bz_bug_map SET
ext_status = ?,
ext_description = ?,
ext_priority = ?,
ext_last_updated = NOW()
WHERE ext_bz_id = ?
AND ext_bz_bug_id = ?
EOS

    $dbh->do($sql, undef, $bug->bug_status, $bug->short_desc, $bug->priority,
      $int_ext_bz->[0]->id, $bug->id);
  }

  return;
}

sub bug_send_changes {
  my ($self, $args) = @_;

  my $bug_id   = $args->{bug};
  my $diffs    = $args->{diffs};
  my $start    = $args->{start};
  my $end      = $args->{end};
  my $changer  = $args->{changer};
  my $sentlist = $args->{sent};
  my $excluded = $args->{excluded};
  my $comments = $args->{comments};

  my $input  = Bugzilla->input_params;
  my $logger = Bugzilla->logger;

  my $bug = Bugzilla::Bug->new($bug_id);

  # Check first if we can change this field
  my $privs    = 0;
  my $can_edit = $bug->check_can_change_field('external_bugs', 0, 1, \$privs);

  # Send any changes that can be sent to the external trackers
  my %already_sent      = ();
  my %already_sent_type = ();
  my @to_send;

  ## REDHAT EXTENSION BEGIN 1010254

  # Check to see if this bug has an external tracker, or has had it's last ones
  # removed.
  # If it does not, abort processing here.
  unless (_has_external_tracker($bug)
    || defined Bugzilla->request_cache->{'external_bugs_removed'})
  {
    return;
  }

  ## REDHAT EXTENSION BEGIN 1165516
  # If the bug is in a security group, but not in
  # 'redhat' or 'private', do not sync.
  if (any { ($_->{name} eq 'security') && ($_->{ison} == 1) } @{$bug->groups}) {
    my $sync_ok = 0;
    foreach my $group (@{$bug->groups}) {
      if (($group->{name} eq 'redhat') && ($group->{ison} == 1)) {
        $sync_ok = 1;
        last;
      }
    }

    unless ($sync_ok) {

      # Can't sync this bug as it is in the security group, but not redhat.
      return;
    }
  }
  ## REDHAT EXTENSION END 1165516

 # Locate all the bugs that are marked 'can_send' and the user has ticked the
 # box to sync the change.  Record the state if any have been checked by the user.
  my %can_send;
  foreach my $key (keys %{$input}) {
    if ($key =~ /send_ext_bz_bug_([0-9]+)/) {
      $can_send{int($1)} = 1;
    }
  }

  # - Send all changes to the job queue for sfdc synchronization.
  # If this bug is not related to SFDC it will be ignored when the job queue
  # is processed.
  #
  #
  # The 'should_send' argument should be ignored if must_send is true.
  # This will need to be handled in the code that processes 'send_sfdc'
  Bugzilla->job_queue->insert(
    'send_sfdc',
    {
      bug_id      => $bug->id,
      delta_ts    => $bug->delta_ts,
      start       => $start,
      should_send => \%can_send,
    }
  );
  ## REDHAT EXTENSION END 1010254

  foreach my $key (keys %{$input}) {
    if ($key =~ /^send_ext_bz_bug_(\d+)$/ && $input->{$key}) {
      $can_edit
        || ThrowUserError('illegal_change',
        {field => 'external_bugs', privs => $privs});
      my $external_bug_id = $1;
      my $ext_bug_obj = Bugzilla::Extension::ExternalBugs::Bug->new($external_bug_id);
      if ( $ext_bug_obj
        && $ext_bug_obj->type->can_send
        && !$already_sent{$ext_bug_obj->id})
      {
        unless ($ext_bug_obj->type->send_once
          and $already_sent_type{$ext_bug_obj->ext_bz_id})
        {
          push(@to_send, $ext_bug_obj);
          $already_sent{$ext_bug_obj->id}             = 1;
          $already_sent_type{$ext_bug_obj->ext_bz_id} = 1;
        }
      }

      # Delete the params in case this gets ran multiple times
      delete $input->{$key};
    }
  }

  # send any changes to external trackers that are required to have changes sent.
  foreach my $ext_bug_obj (@{$bug->external_bugs}) {
    if ($ext_bug_obj->type->must_send && !$already_sent{$ext_bug_obj->id}) {
      unless ($ext_bug_obj->type->send_once
        and $already_sent_type{$ext_bug_obj->ext_bz_id})
      {
        push(@to_send, $ext_bug_obj);
        $already_sent{$ext_bug_obj->id}             = 1;
        $already_sent_type{$ext_bug_obj->ext_bz_id} = 1;
      }
    }
  }

  # Also check for any external bugs removed, and if they are must_send,
  # we send those as well
  if (Bugzilla->request_cache->{'external_bugs_removed'}) {
    foreach my $ext_bug_obj (@{Bugzilla->request_cache->{'external_bugs_removed'}})
    {
      if ($ext_bug_obj && !$already_sent{$ext_bug_obj->id}) {
        push(@to_send, $ext_bug_obj);
        $already_sent{$ext_bug_obj->id} = 1;
      }
    }
  }

  # Send the changes
  foreach my $ext_bug_obj (@to_send) {
    if ($ext_bug_obj->type->send_changes({
      ext_bug  => $ext_bug_obj,
      bug      => $bug,
      diffs    => $diffs,
      start    => $start,
      end      => $end,
      changer  => $changer,
      comments => $comments,
    }))
    {
      $logger->debug("EXTERNAL BUG SENT: ("
          . $ext_bug_obj->type->description . " "
          . $ext_bug_obj->ext_bz_bug_id
          . ")");

      # push(@$sentlist, "(" . $ext_bug_obj->type->description . " " .
      #                  $ext_bug_obj->ext_bz_bug_id . ")");
    }
    else {
      $logger->debug("EXTERNAL BUG EXCLUDED: ("
          . $ext_bug_obj->type->description . " "
          . $ext_bug_obj->ext_bz_bug_id
          . ")");

      # push(@$excluded, "(" . $ext_bug_obj->type->description . " " .
      #                  $ext_bug_obj->ext_bz_bug_id . ")");
    }
  }

  return;
}

## REDHAT EXTENSION BEGIN 1010254

# Note: We are using SQL to determine if there is an external tracker.
#
# This is because $bug->external_bugs checks the users permissions to
# see if they are allowed to see the tracker.
# If they are not allowed to see the tracker, it will appear as if there
# are no external trackers.

sub _has_external_tracker {
  my $bug = shift;

  my $query = 'SELECT COUNT(*) FROM ext_bz_bug_map WHERE bug_id = ?';

  my ($bug_count) = Bugzilla->dbh->selectrow_array($query, undef, $bug->id);

  return $bug_count;
}
## REDHAT EXTENSION END 1010254

sub bug_fields {
  my ($self, $args) = @_;

  my $fields = $args->{fields};

  push(@$fields, 'external_bugs',         'external_choices');
  push(@$fields, 'external_bugzilla.url', 'external_bugzilla.description');

  return;
}

sub buglist_columns {
  my ($self, $args) = @_;

  my $columns = $args->{'columns'};

  $columns->{ext_bz_list} = {title => 'Link List', name => 'NULL'};
  $columns->{ext_bz_count}
    = {title => 'Link Count', name => 'COUNT(DISTINCT map_ext_bz_count.id)'};

  return;
}

sub buglist_column_joins {
  my ($self, $args) = @_;
  my $joins = $args->{'column_joins'};

  $joins->{'ext_bz_count'} = {table => 'ext_bz_bug_map', join => 'LEFT',};

  return;
}

sub bug_filter_change {
  my ($self, $args) = @_;

  my $user      = $args->{user};
  my $field     = $args->{field};
  my $object_id = $args->{object_id};
  my $visible   = $args->{visible};

  if ($field eq 'ext_bz_bug_map.ext_bz_bug_id' and $object_id) {
    if (not $user->can_see_ext_bz($object_id)) {
      $$visible = 0;
    }
  }

  return;
}

sub enter_bug_entrydefaultvars {
  my ($self, $args) = @_;

  my $vars = $args->{vars};

  $vars->{'external_choices'} = Bugzilla::Extension::ExternalBugs::Type->match();

  return;
}

sub post_bug_fields {
  my ($self, $args) = @_;

  my $bug_fields = $args->{bug_fields};
  my $input      = Bugzilla->input_params;

  push @$bug_fields, grep {/^external_(bug_)?id_\d+$/} keys %$input;
  push @$bug_fields, grep {/^external_bug_url_\d+$/} keys %$input;

  return;
}

sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{'ext_bz_bug_map.ext_bz_bug_id'} = {
    _non_changed => sub { _ext_bz_bug_nonchanged(@_, 'ext_bz_bug_id') }
  };
  $operators->{'ext_bz_bug_map.ext_status'} = {
    _non_changed => sub { _ext_bz_bug_nonchanged(@_, 'ext_status') }
  };
  $operators->{'external_bugzilla.description'} = {
    _non_changed => sub { _ext_bz_nonchanged(@_, 'description') }
  };
  $operators->{'external_bugzilla.url'} = {
    _non_changed => sub { _ext_bz_nonchanged(@_, 'url') }
  };
  ## REDHAT EXTENSION START 1216256
  $operators->{'ext_bz_count'} = {_default => sub { _ext_bz_count() },};
  ## REDHAT EXTENSION END 1216256

  return;
}

sub _ext_bz_bug_nonchanged {
  my ($invocant, $args, $column) = @_;

  ## REDHAT EXTENSION BEGIN 821293
  my $group_list = Bugzilla->user->groups_as_string;
  my $extra_sql  = "
AND ext_bz_bug_map.ext_bz_id IN (
    SELECT DISTINCT id
    FROM external_bugzilla b
        LEFT JOIN ext_bz_group_map m ON (m.ext_bz_id = b.id)
    WHERE m.group_id IS NULL OR m.group_id IN ($group_list)
)";
  ## REDHAT EXTENSION END 821293

  $args->{full_field} = "ext_bz_bug_map.$column";
  $invocant->_do_operator_function($args);
  $args->{term} = Bugzilla::Search::build_subselect(
    "bugs.bug_id",    "ext_bz_bug_map.bug_id",
    "ext_bz_bug_map", $args->{term} . $extra_sql
  );

  return;
}

sub _ext_bz_nonchanged {
  my ($invocant, $args, $column) = @_;

  ## REDHAT EXTENSION BEGIN 821293
  my $group_list = Bugzilla->user->groups_as_string;
  my $extra_sql  = "
AND external_bugzilla.id IN (
    SELECT DISTINCT id
    FROM external_bugzilla b
        LEFT JOIN ext_bz_group_map m ON (m.ext_bz_id = b.id)
    WHERE m.group_id IS NULL OR m.group_id IN ($group_list)
)";
  ## REDHAT EXTENSION END 821293

  $args->{full_field} = "external_bugzilla.$column";
  $invocant->_do_operator_function($args);
  $args->{term} = Bugzilla::Search::build_subselect(
    "bugs.bug_id",
    "ext_bz_bug_map.bug_id",
    "ext_bz_bug_map JOIN external_bugzilla ON ext_bz_bug_map.ext_bz_id = external_bugzilla.id",
    $args->{term} . $extra_sql
  );

  return;
}

## REDHAT EXTENSION BEGIN 693936
# Remove external trackers when removing bugs
sub bug_remove_from_db {
  my ($self, $args) = @_;
  my $bug_id = $args->{bug_id};

  my $query = "DELETE FROM ext_bz_bug_map WHERE bug_id = ?";
  Bugzilla->dbh->do($query, undef, $bug_id);

  return;
}
## REDHAT EXTENSION END 693936

## REDHAT EXTENSION START 1216256
sub _ext_bz_count {
  my ($invocant, $args) = @_;
  my $joins = $args->{joins};
  my $table = "ext_bz_count";
  my $join  = {
    table => "(SELECT bug_id, COUNT(*) AS num"
      . " FROM ext_bz_bug_map GROUP BY bug_id)",
    as => $table,
  };
  push(@$joins, $join);
  $args->{full_field} = "${table}.num";

  return;
}
## REDHAT EXTENSION END 1216256

##############
# WebService #
##############

sub webservice {
  my ($self, $args) = @_;
  my $dispatch = $args->{dispatch};
  $dispatch->{ExternalBugs} = "Bugzilla::Extension::ExternalBugs::WebService";

  return;
}

sub localconfig_symbols {
  my ($self, $args) = @_;
  my $read_symbols       = $args->{read_symbols};
  my @local_read_symbols = map($_->{name}, LOCALCONFIG_VARS);
  push(@$read_symbols, @local_read_symbols);
  return;
}

sub localconfig_vars {
  my ($self, $args) = @_;
  my $localconfig_vars = $args->{localconfig_vars};
  push(@$localconfig_vars, LOCALCONFIG_VARS);
  return;
}

#################
# Non-Bug Hooks #
################

sub template_before_process {
  my ($self, $args) = @_;

  my ($vars, $file, $context) = @$args{qw(vars file context)};

  if ($file eq 'list/edit-multiple.html.tmpl') {
    $vars->{external_choices} = Bugzilla::Extension::ExternalBugs::Type->match();
  }

  return;
}

## REDHAT EXTENSION START 1206002
sub template_after_create {
  my ($self, $args) = @_;
  my $context = $args->{template}->context;

  $context->define_vmethod(
    list => sort_ets => sub {
      my ($list) = @_;
      return [
        sort {
          $a->type->description cmp $b->type->description
            || (
            (
              looks_like_number($a->ext_bz_bug_id) && looks_like_number($b->ext_bz_bug_id)
            )
            ? $a->ext_bz_bug_id <=> $b->ext_bz_bug_id
            : $a->ext_bz_bug_id cmp $b->ext_bz_bug_id
            )
        } @$list
      ];
    },
  );

  return;
}
## REDHAT EXTENSION END 1206002

## REDHAT EXTENSION START 1538429
sub template_before_create {
  my ($self, $args) = @_;

  # This is how we add dynamic filters
  $args->{config}->{FILTERS}->{filter_ets} = [
    sub {
      my ($context, @args) = @_;
      return sub {
        my $txt = shift;
        my ($match, $format) = @args;

        # don't process strings that don't match
        if ($match && $format && $match ne '' && $format ne '' && $txt =~ m/$match/) {
          $txt = sprintf($format, $txt =~ m/$match/);
        }
        return ($txt);
      }
    },
    1
  ];

  return;
}
## REDHAT EXTENSION END 1538429


sub job_map {
  my ($self, $args) = @_;

  my $job_map = $args->{job_map};

  # This adds the named class (an instance of TheSchwartz::Worker) as a handler
  $job_map->{'send_sfdc'} = 'Bugzilla::Extension::ExternalBugs::Job::SFDC';
  $job_map->{'send_jira'} = 'Bugzilla::Extension::ExternalBugs::Job::JIRA';

  return;
}

## REDHAT EXTENSION BEGIN 821293
sub _user_can_see_ext_bz {
  my ($self, $ext_bz_id) = @_;

  return 0 unless ($ext_bz_id);

  return (1) if ($self->in_group('admin'));

  # Check that the external tracker exists
  my $exists;
  if (exists(Bugzilla->request_cache->{external_bugzilla_exists}->{$ext_bz_id})) {
    $exists = Bugzilla->request_cache->{external_bugzilla_exists}->{$ext_bz_id};
  }
  else {
    my $query = "SELECT id FROM external_bugzilla WHERE id = ?";
    $exists = Bugzilla->dbh->selectcol_arrayref($query, undef, $ext_bz_id);
    Bugzilla->request_cache->{external_bugzilla_exists}->{$ext_bz_id} = $exists;
  }

  return 0 unless $exists;

  # What groups does this external tracker have

  my $group_ids;
  if (exists(
    Bugzilla->request_cache->{external_bugzilla_group_map}->{$ext_bz_id}))
  {
    $group_ids
      = Bugzilla->request_cache->{external_bugzilla_group_map}->{$ext_bz_id};
  }
  else {
    my $query = "SELECT group_id FROM ext_bz_group_map WHERE ext_bz_id = ?";
    $group_ids = Bugzilla->dbh->selectcol_arrayref($query, undef, $ext_bz_id);
    Bugzilla->request_cache->{external_bugzilla_group_map}->{$ext_bz_id}
      = $group_ids;
  }

  # If none, anyone can see it
  if (scalar(@$group_ids) == 0) {
    return 1;
  }

  # Lets check that the user is in at least one group
  foreach my $group_id (@$group_ids) {
    if ($self->in_group_id($group_id)) {

      # They are!
      return 1;
    }
  }

  # They're not :(
  return 0;
}
## REDHAT EXTENSION END 821293

__PACKAGE__->NAME;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs

=head1 DESCRIPTION

External Bugs, or External Trackers, or Links, provides a way to link a Bugzilla
bug to an external issue source. The user can provide extra data, such as a
subject, the state of the external bug, etc. The idea is to allow a user viewing a
bug in Bugzilla to have an understanding of externally related issues and an
overview of the state of those issues.

A Tracker is a specific service such as bugzilla.mozilla.org. Each service can be
individually configured and has a L<Type|Bugzilla::Extension::ExternalBugs::Type>.
Some Types, such as Bugzilla and Jira, can pull data from Trackers when a user
loads the bug they are on in the web UI. This syncing must be enabled on the
individual Tracker. If a Tracker has pull enabled the sync icon on the bug edit
page will allow users to manually sync a link's data if required. 

Some Types, such as Jira and SFDC, can push comments and bug fields to a Tracker
when the bug is updated.

A Link can be updated or removed via the L<API|Bugzilla::Extension::ExternalBugs::WebService>.

There are several parameters an admin can set to control what the system is doing.
See L<Bugzilla::Extension::ExternalBugs::Config>.

=head1 Methods

=head2  C<Bugzilla::Bug::add_external_bug>

Link an ET to a bug.

=head2  C<Bugzilla::Bug::external_bugs>

Returns a list of ETs on a bug.

=head2  C<Bugzilla::Bug::external_choices>

Returns a list if Trackers the current user can see on this bug. 

=head2  C<Bugzilla::Bug::internal_links>

Returns a list of ETs on a bug that are in the same Bugzilla instance.

=head2  C<Bugzilla::Bug::remove_external_bug>

Remove an ET from the bug.

=head2  C<Bugzilla::User::can_see_ext_bz>

Determines if the user can see the specified Tracker.

=cut
