# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Job::JIRA;
use strict;
use warnings;
use 5.10.1;

BEGIN {
  ## no critic (ProhibitStringyEval)
# This is safe and avoids having to conditionally use this module if TheSchwartz is not installed
  eval "use parent qw(TheSchwartz::Worker Bugzilla::Extension)";
  ## use critic
}


# The longest we expect a job to possibly take, in seconds.
use constant grab_for => 600;

# We don't want JIRA to fail permanently very easily. Retry for 24 hours.
use constant max_retries => 29;

# The first few retries happen quickly, but after that we wait an hour for
# each retry.
sub retry_delay {
  my ($class, $num_retries) = @_;
  if ($num_retries < 5) {
    return (10, 30, 60, 300, 600)[$num_retries];
  }

  # One hour
  return 60 * 60;
}

sub work {
  my ($class, $job) = @_;
  my $action        = $job->arg->{action};
  my $bug_id        = $job->arg->{bug_id};
  my $ext_bz_bug_id = $job->arg->{ext_bz_bug_id};
  my $ext_bz_id     = $job->arg->{ext_bz_id};
  my $comment       = $job->arg->{comment};
  my $is_private    = $job->arg->{is_private};
  my $group         = $job->arg->{group};

  my $success = 0;

  if (not $ext_bz_bug_id) {

    # This could happen because the external bug has been removed
    # between when the job was created and now. If that is an issue,
    # we may want to mute this.
    $job->permanent_failure("Ext_bz_bug_id not specified");
    return;
  }

  # We need to be a fake super user to see restricted external trackers
  Bugzilla->set_user(Bugzilla::User->super_user);

  my $tracker = Bugzilla::Extension::ExternalBugs::Type->new($ext_bz_id);
  if (not $tracker) {
    $job->failed("Cannot find external tracker '$ext_bz_id'");
    return;
  }
  elsif ($tracker->type ne 'JIRA') {
    $job->failed("Tracker $ext_bz_id is not of JIRA type");
    return;
  }

  ## RED HAT EXTENSION START 1175565 1262660
  # if we aren't using external trackers, or not in mode 1,
  # and not in mode 4 and the tracker has ' stage' in it's name,
  # mark as done
  if (
    (!defined(Bugzilla->params->{'enable_external_trackers'}))
    || (
      Bugzilla->params->{'enable_external_trackers'} != 1
      && (!(
        Bugzilla->params->{'enable_external_trackers'} == 3 && $tracker->name =~ m/[\s-]Stage/i
      ))
    )
    )
  {
    $job->completed;
    return 1;
  }
  ## RED HAT EXTENSION END 1175565 1262660

  if ($action eq 'send_as_comment') {
    $success
      = eval { $tracker->_send_as_comment($ext_bz_bug_id, $comment, $is_private); };
  }
  elsif ($action eq 'set_jira_flags') {
    $success = eval { $tracker->_set_jira_flags($ext_bz_bug_id); };
  }
  else {
    $job->failed("Unknown action '$action'");
    return;
  }

  if (!$success) {
    $job->failed($@);
    undef $@;
  }
  else {
    $job->completed;
  }

  return;
}

1;
