# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Job::SFDC;
use strict;
use warnings;
use 5.10.1;

BEGIN {
  ## no critic (ProhibitStringyEval)
# This is safe and avoids having to conditionally use this module if TheSchwartz is not installed
  eval "use parent qw(TheSchwartz::Worker Bugzilla::Extension)";
  ## use critic
}

# The longest we expect a job to possibly take, in seconds.
use constant grab_for => 600;

# We don't want SFDC to fail permanently very easily. Retry for 48 hours.
use constant max_retries => 53;

use Bugzilla::WebService::Server::XMLRPC;
use Bugzilla::WebService::Bug;
use Bugzilla::Extension::ExternalBugs::Type;

# The first few retries happen quickly, but after that we wait an hour for
# each retry.
sub retry_delay {
  my ($class, $num_retries) = @_;
  if ($num_retries < 5) {
    return (10, 30, 60, 300, 600)[$num_retries];
  }

  # One hour
  return 60 * 60;
}

sub _fetch_bug_data {
  my $bug_id = shift;
  my $start  = shift;

  my $serializer = Bugzilla::XMLRPC::Serializer->new();

  # Send a full copy of the bug report using Bug.get
  my $bug_data = Bugzilla::WebService::Bug->get({
    ids            => [$bug_id],
    extra_fields   => ['external_bugs', 'description', 'flags'],
    exclude_fields => ['attachments'],
  });
  my $bug_xml_data = $serializer->envelope('response', 'Bug.get', $bug_data);

  # Get the comments for this bug
  my $bug_comments = Bugzilla::WebService::Bug->comments(
    {ids => [$bug_id], new_since => $start});

  # We only send comments if there are some
  my $bug_xml_comments;
  my $comments = $bug_comments->{bugs}{$bug_id}{comments};
  if (defined($comments) && scalar @$comments) {
    $bug_xml_comments
      = $serializer->envelope('response', 'Bug.comments', $bug_comments);
  }

  return ($bug_xml_data, $bug_xml_comments);
}

sub work {
  my ($class, $job) = @_;
  my $bug_id      = $job->arg->{bug_id};
  my $delta_ts    = $job->arg->{delta_ts};
  my $should_send = $job->arg->{should_send};
  my $start       = $job->arg->{start};

  ## RED HAT EXTENSION START 1175565
  # if we aren't using external trackers mark as done
  unless (Bugzilla->params->{'enable_external_trackers'}) {
    $job->completed;
    return 1;
  }
  ## RED HAT EXTENSION END 1175565

  # We need to be a fake super user to see restricted external trackers
  Bugzilla->set_user(Bugzilla::User->super_user);

  my $bug = Bugzilla::Bug->new($bug_id);
  ## RED HAT EXTENSION START 1175565 1262660
  # Ignore production SFDC in mode 2
  my $re = qr/^SFDC/;
  if (Bugzilla->params->{'enable_external_trackers'} >= 2) {
    $re = qr/^SFDC./;
  }

  my @trackers = grep { $_->type->type =~ /$re/ } @{$bug->external_bugs};
  ## RED HAT EXTENSION END 1175565 1262660

  unless (scalar @trackers) {
    $job->completed;
    return 1;
  }

  ## REDHAT EXTENSION START 1132415
  # To prevent SFDC from trying to update a bug before it is committed, we
  # will check it against the recorded delta_ts. If the bug value is earlier,
  # it's a good sign that the bug has not been commited
  if ($bug->delta_ts lt $delta_ts) {
    $job->failed("Bug $bug_id does not exist, or delta_ts is too early");
    return;
  }
  ## REDHAT EXTENSION END 1132415

  my ($bug_xml_data, $bug_xml_comments) = _fetch_bug_data($bug_id, $start);

  foreach my $tracker (@trackers) {
    my $type = Bugzilla::Extension::ExternalBugs::Type->new($tracker->ext_bz_id);

# If the tracker is set to 'must_send' OR the tracker is a 'can_send' AND the sync box has been
# checked, send the bug. if not, skip this tracker.
    if ($type->must_send || ($type->can_send && $should_send->{$tracker->id})) {
      my $success
        = eval { $type->send_to_sfdc($bug_id, $bug_xml_data, $bug_xml_comments,); };

      if (!$success) {
        $job->failed($@);
        undef $@;
        return;
      }
    }
  }

  $job->completed;

  return;
}

1;
