# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases::ACKList;

use warnings;
use strict;
use 5.10.1;

use base qw(Bugzilla::Object);
our $VERSION = '0.2';

use Bugzilla;
use Bugzilla::Extension::Releases::Util;
use Bugzilla::Extension::Releases::ACKCustomSearch;
use Bugzilla::Extension::ExternalBugs::Bug;

use YAML::Syck;
use JSON::Any;
use List::MoreUtils qw(any none);

#################################################

# input parameters:
# action
#           descr:
#           values: undef|Edit|Submit
# combined
#           descr:
#           values:
# Bugzilla_login
#           descr:
#           values:
# Bugzilla_password
#           descr:
#           values:
# view-version
#           descr:
#           values:
# show-sql
#           descr:
#           values:
# apply_update
#           descr:
#           values:
# update
#           descr:
#           values:
# list
#           descr:
#           values:
# type
#           descr:
#           values:
# apply_listtype
#           descr:
#           values:
# listtype
#           descr:
#           values:
# resolved
#           descr:
#           values:
# resolved-type
#           descr:
#           values:
# column
#           descr:
#           values:
# apply_group
#           descr:
#           values:
# group
#           descr:
#           values:
# apply_member
#           descr:
#           values:
# member
#           descr:
#           values:
# apply_component
#           descr:
#           values:
# component
#           descr:
#           values:
# priorities
#           descr:
#           values:
# product
#           descr:
#           values:
# version
#           descr:
#           values:
# mode
#           descr:
#           values:
# bug_status
#           descr:
#           values:
# apply_pm_ack
#           descr:
#           values:
# pm_ack
#           descr:
#           values:
# apply_devel_ack
#           descr:
#           values:
# devel_ack
#           descr:
#           values:
# apply_qa_ack
#           descr:
#           values:
# qa_ack
#           descr:
#           values:
# blocker_flag
#           descr:
#           values:
# blocker_beta_flag
#           descr:
#           values:
# exception_flag
#           descr:
#           values:
# hot_fix_requested_flag
#           descr:
#           values:
# bugzilla-buglist
#           descr:
#           values:
# ctype
#           descr:
#           values:
# apply_sort
#           descr:
#           values:
# sort
#           descr:
#           values:
# partner
#           descr:
#           values:
# subtype
#           descr:
#           values:

sub new {
  my $invocant = shift;
  my $class    = ref($invocant) || $invocant;
  my $params   = shift;
  my $self     = {};
  bless($self, $class);

  $self->{config} = ack_config();

  $self->{input} = Bugzilla->input_params;

  return $self;
}

sub title {
  my ($self) = @_;

  my $view_version = $self->qpu_int('view-version', 0);
  my $update       = $self->qpu_str('update');
  my $list         = $self->qpu_str('list');
  my $type         = $self->qpu_str('type');
  my $listtype     = $self->qpu_str('listtype');

  if (!$view_version && $listtype) {
    if ($listtype =~ /^(\w+)-(\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^(\w+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  elsif ($listtype) {
    if ($listtype =~ /^([\w-]+) (\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^([\w-]+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  if (!$list) {
    $list = $self->{config}->{TYPE_ACK};
  }

  my $member       = $self->qpu_str('member');
  my @group_params = Bugzilla->cgi->param('group');
  my $selected_groups = []; ## BUGBUG agile groups $self->orgchart_group_id( \@group_params );

  my $title = '';

  # print query release
  if ($update) {
    $title = $update . ' ' . $list . ($type ? ' ' . $type : '');
  }

  # print query user email
  if ($member) {
    $title .= ' -- ' . $member;
  }

  # print query groups
  if (!is_empty_array($selected_groups)) {
    my $group_names = []; ## BUGBUG agile groups $self->orgchart_group_name($selected_groups);
    $title .= ' -- ' . join(', ', @$group_names);
  }

  return $title || 'Bugs';
}

sub get_milestones {
  my ($self) = @_;

  my $release = $self->qpu_str('release', '');

  my $j = JSON::Any->new;

  my $sql = <<SQL;
        SELECT DISTINCT m.value
          FROM milestones m
            LEFT JOIN products p ON m.product_id=p.id
            LEFT JOIN flaginclusions i ON i.product_id=m.product_id
            LEFT JOIN flagtypes as f ON f.id=i.type_id
          WHERE f.name=
SQL

  my $dbh = Bugzilla->dbh();
  $sql .= $dbh->quote($release);
  my @rows = $dbh->selectcol_arrayref($sql);

  print Bugzilla->cgi->header(-type => "application/json", -charset => "utf-8",);
  print $j->to_json(@rows);
  return;
}

sub show_flag_bugs {
  my ($self, $vars) = @_;

  $vars->{title} = Bugzilla->cgi->param('bug_id') ? 'custom' : $self->title;
  $self->flag_bugs($vars);

  return;
}

sub bug_define_form {
  my ($self, $vars) = @_;

  my $dbh          = Bugzilla->dbh();
  my @param_name   = ();
  my @param_value  = ();
  my $update       = $self->qpu_str('update');
  my $view_version = $self->qpu_int('view-version', 0);
  if ($view_version) {
    push @param_name,  'view-version';
    push @param_value, $view_version;
  }
  my $combined = $self->qpu_bool('combine', 0);
  if ($combined) {
    push @param_name,  'combine';
    push @param_value, $combined;
  }
  my $list     = $self->qpu_str('list');
  my $type     = $self->qpu_str('type');
  my $listtype = $self->qpu_str('listtype');
  if (!$view_version && $listtype) {
    if ($listtype =~ /^(\w+)-(\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^(\w+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  elsif ($listtype) {
    if ($listtype =~ /^([\w-]+) (\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^([\w-]+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  my $raw_list         = $list;
  my $milestone_filter = undef;
  if ($view_version > 2) {

    # check whether we have a composite blocker
    if (defined $list && $list =~ /^BLOCKER-(.+)$/) {
      $milestone_filter = $1;
      $list             = 'BLOCKER';
    }
  }
  my $release_type = undef;
  if ($combined
    && ($type =~ /^(PROPOSED|APPROVED)_(PROPOSED|APPROVED|REJECTED)$/))
  {
    $type =~ s/_((PROPOSED|APPROVED|REJECTED))$//;
    $release_type = $1;
  }
  if ($list) {
    push @param_name,  'list';
    push @param_value, $list;
  }
  if ($type) {
    push @param_name,  'type';
    push @param_value, $type;
  }
  my $column_set = $self->qpu_str('column-set');
  if ($column_set) {
    push @param_name,  'column-set';
    push @param_value, $column_set;
  }
  my $column = $self->qpu_str('column');

  # add hidden parameters to form
  my @hidden = ();
  for (my $i = 0; $i < @param_name; $i++) {
    my %param_data;
    $param_data{'name'}  = $param_name[$i];
    $param_data{'value'} = $param_value[$i];
    push(@hidden, \%param_data);
  }
  $vars->{hidden_params} = \@hidden;

  # update, if supplied
  $vars->{release} = $update if $update;

  # retrieve approved releases
  my $ack_releases_ref = get_releases();

  # fill releases selection box
  my @updates = ();

  # requires blank element
  push(@updates, {name => '', value => '',});

  for my $r (@$ack_releases_ref) {
    push(@updates,
      {name => $r, value => $r, selected => ($update && $r eq $update) ? 1 : 0,});
  }

  # if not a configured update, add as custom
  if ($update
    && (!@$ack_releases_ref || (none { $_ eq $update } @$ack_releases_ref)))
  {
    push(@updates, {name => 'custom', value => $update, selected => 1,});
  }

  $vars->{updates} = \@updates;

  # retrieve all milestones
  my $sth = $dbh->prepare(<<RETRIEVE_ALL_MILESTONES);
        SELECT DISTINCT bugs.target_milestone
          FROM bugs,
               flags,
               flagtypes
         WHERE bugs.bug_id=flags.bug_id
               AND flagtypes.id=flags.type_id
               AND flagtypes.name='blocker'
               AND NOT (bugs.target_milestone IS NULL)
               AND NOT (bugs.target_milestone='')
               AND NOT (bugs.target_milestone='---')
      ORDER BY bugs.target_milestone
RETRIEVE_ALL_MILESTONES
  $sth->execute();

  my @milestones = ("no-target");

  while (my ($milestone) = $sth->fetchrow_array()) {
    push @milestones, $milestone;
  }

  $sth->finish();

  # fill list selection box
  my @listtypes = ();
  if (!$view_version) {
    my @acktypes  = ('PROPOSED', 'APPROVED', 'BLOCKER');
    my @fasttypes = ('PROPOSED', 'APPROVED');
    my %ack_data;
    $ack_data{'name'}  = 'ACK';
    $ack_data{'value'} = 'ACK';
    if ((defined $list && $list eq 'ACK') && !$type) {
      $ack_data{'selected'} = 1;
    }
    push(@listtypes, \%ack_data);
    for (my $i = 0; $i < @acktypes; $i++) {
      my %type_data;
      $type_data{'value'} = 'ACK' . '-' . $acktypes[$i];
      $type_data{'name'}  = 'ACK-' . $acktypes[$i];
      if ((defined $list && $list eq 'ACK') && ($type && ($type eq $acktypes[$i]))) {
        $type_data{'selected'} = 1;
      }
      push(@listtypes, \%type_data);
    }
    my %fast_data;
    $fast_data{'name'}  = 'FAST';
    $fast_data{'value'} = 'FAST';
    if ((defined $list && $list eq 'FAST') && !$type) {
      $fast_data{'selected'} = 1;
    }
    push(@listtypes, \%fast_data);
    for (my $i = 0; $i < @fasttypes; $i++) {
      my %type_data;
      $type_data{'value'} = 'FAST' . '-' . $fasttypes[$i];
      $type_data{'name'}  = 'FAST-' . $fasttypes[$i];
      if ((defined $list && $list eq 'FAST') && $type && ($type eq $fasttypes[$i])) {
        $type_data{'selected'} = 1;
      }
      push(@listtypes, \%type_data);
    }
  }
  else {
    my @flag_list_types = ('TOTAL', 'ACK');
    unshift @flag_list_types, 'RAW' if (defined $list && $list eq 'RAW');
    push @flag_list_types, @{$self->{config}->{FLAG_LIST}};
    my @acktypes = ('PROPOSED', 'APPROVED');
    foreach my $flag_list_type (@flag_list_types) {
      if (($view_version > 2) && ($flag_list_type =~ /^BLOCKER/)) {
        next;
      }
      my %ack_data;
      $ack_data{'name'}  = $flag_list_type;
      $ack_data{'value'} = $flag_list_type;
      if ((defined $list && $list eq $flag_list_type) && !$type) {
        $ack_data{'selected'} = 1;
      }
      push(@listtypes, \%ack_data);
      next if $flag_list_type eq 'TOTAL' || $flag_list_type eq 'RAW';
      for (my $i = 0; $i < @acktypes; $i++) {
        my %type_data;
        $type_data{'value'} = $flag_list_type . ' ' . $acktypes[$i];
        $type_data{'name'}  = $flag_list_type . ' ' . $acktypes[$i];
        if ( (defined $list && $list eq $flag_list_type)
          && ($type && ($type eq $acktypes[$i])))
        {
          $type_data{'selected'} = 1;
        }
        push(@listtypes, \%type_data);
      }
    }

    if ($view_version > 2) {
      foreach my $milestone (@milestones) {
        my $flag_list_type = 'BLOCKER-' . $milestone;
        my %ack_data;
        $ack_data{'name'}  = $flag_list_type;
        $ack_data{'value'} = $flag_list_type;
        if ((defined $raw_list && $raw_list eq $flag_list_type) && !$type) {
          $ack_data{'selected'} = 1;
        }
        push(@listtypes, \%ack_data);
        for (my $i = 0; $i < @acktypes; $i++) {
          my %type_data;
          $type_data{'value'} = $flag_list_type . ' ' . $acktypes[$i];
          $type_data{'name'}  = $flag_list_type . ' ' . $acktypes[$i];
          if ( (defined $raw_list && $raw_list eq $flag_list_type)
            && ($type && ($type eq $acktypes[$i])))
          {
            $type_data{'selected'} = 1;
          }
          push(@listtypes, \%type_data);
        }
      }
    }
  }
  $vars->{listtypes} = \@listtypes;

  # generate selection box for columns
  my $ccolumns = 'BLOCKER_COLUMNS';
  if ($column_set) {
    $ccolumns = 'BLOCKER_' . $column_set . '_COLUMNS';
  }
  my @ack_columns    = @{$self->{config}->{$ccolumns}};
  my @columns_select = ();
  my %column_data;
  $column_data{'name'}  = 'All';
  $column_data{'value'} = '';
  push(@columns_select, \%column_data);
  foreach my $ack_column (@ack_columns) {
    my %param_data;
    $param_data{'name'}  = $ack_column;
    $param_data{'value'} = $ack_column;
    if ($column && $column eq $ack_column) {
      $param_data{'selected'} = 1;
    }
    push(@columns_select, \%param_data);
  }
  $vars->{columns} = \@columns_select;

  return;
}

sub bug_filter_form {
  my ($self, $vars, $param) = @_;

  my $dbh                   = Bugzilla->dbh();
  my $keyword_filter        = $self->qpu_str('keyword', '');
  my $keyword_logic         = $self->qpu_str('keyword_logic', 'any');
  my $blocks_filter         = $self->qpu_str('blocks', '');
  my $blocks_depth          = $self->qpu_int('blocks_depth', 1);
  my $selected_groups       = [Bugzilla->cgi->param('group')];
  my $group_negate          = $self->qpu_int('group_negate', 0);
  my $member                = $self->qpu_str('member');
  my $component             = $self->qpu_str('component');
  my $component_not         = $self->qpu_bool('component_exclude');
  my $approved_components   = $self->qpu_bool('approved_components');
  my $capacity_components   = $self->qpu_bool('capacity_components');
  my $unassigned_components = $self->qpu_bool('unassigned_components');
  my $external_tracker      = $self->qpu_str('external_tracker');
  my $external_tracker_exclusive
    = $self->qpu_bool('external_tracker_exclusive', 0);
  my $customer_mode    = $self->qpu_bool('mode', 0);
  my $sort_by          = $self->qpu_int('sort', 0);
  my $partner_mode     = $self->qpu_bool('partner', 0);
  my $bug_status       = $param->{'bug_status'};
  my @verified         = Bugzilla->cgi->param('verified');
  my $hide_z_proposed  = $self->qpu_bool('hide_z_proposed');
  my $hide_z_approved  = $self->qpu_bool('hide_z_approved');
  my $profiles_join    = $self->qpu_str('user_type', 'assigned_to');
  my @products         = Bugzilla->cgi->param('product');
  my @layered_products = Bugzilla->cgi->param('layered_products');
  my @priorities       = Bugzilla->cgi->param('priorities');
  my $update           = $self->qpu_str('update');
  my $user_logic       = $self->qpu_str('user_logic', 'any');
  my @cf_internal_target_release
    = Bugzilla->cgi->param('cf_internal_target_release');

  # generate selection box for products

  my @classifications;
  my $pc_str = $self->{config}->{'PRODUCT_CLASSIFICATIONS'} || '';

  if ($pc_str) {
    @classifications = split(/,/, $pc_str);
  }

  push @classifications, 'Red Hat' if !@classifications;

  my @show_products = ();
  my $select_products
    = "SELECT products.name FROM classifications, products WHERE classifications.id=products.classification_id AND products.isactive=1 AND ("
    . join(' OR ',
    map { "classifications.name=" . $dbh->quote($_) } @classifications)
    . ") ORDER BY classifications.sortkey, products.name";

  my $sth = $dbh->prepare($select_products);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref()) {
    my %param_data;
    $param_data{name}  = $row->{name};
    $param_data{value} = $row->{name};

    foreach my $selected (@products) {
      if ($selected eq $row->{name}) {
        $param_data{selected} = 1;
        last;
      }
    }

    push(@show_products, \%param_data);
  }

  $vars->{products} = \@show_products if @show_products;

  # generate selection box for priorities

  my %all_priorities = %{$self->get_priorities()};
  $vars->{sort_hash_by} = $self->sort_hash_by();

  if (%all_priorities) {
    foreach my $priority (@priorities) {
      if (defined $all_priorities{$priority}) {
        $all_priorities{$priority}{selected} = 1;
      }
    }
    $vars->{all_priorities} = \%all_priorities;
  }

  # generate selection box for groups
  my $groups = groups_selection($selected_groups);
  $vars->{groups}       = $groups;
  $vars->{group_negate} = $group_negate if $group_negate;

  # member section

  my @profiles_joins = ();
  foreach my $profiles_type (qw(assigned_to qa_contact reporter)) {
    push(
      @profiles_joins,
      {
        name    => $profiles_type,
        value   => $profiles_type,
        checked => $profiles_join eq $profiles_type
      }
    );
  }
  $vars->{user_types} = \@profiles_joins;

  if ($member && !($member =~ /\@/)) {
    $member .= '@redhat.com';
  }
  $vars->{member} = $member if $member;

  my @partners = Bugzilla->cgi->param('cf_partner');
  my $p_sth    = $dbh->prepare(
    "SELECT value FROM cf_partner WHERE isactive=1 ORDER BY sortkey");
  $p_sth->execute();

  my @partner_list = ();

  # default select all partners
  my $select_all_partners = (!@partners || has_member('All', @partners)) ? 1 : 0;

  while (my ($partner_name) = $p_sth->fetchrow_array) {
    my %param_data;
    $param_data{'name'}  = $partner_name;
    $param_data{'value'} = $partner_name;
    if (!$select_all_partners && has_member($partner_name, @partners)) {
      $param_data{'selected'} = 1;
    }
    push(@partner_list, \%param_data);
  }
  $p_sth->finish();

  $vars->{'partners'} = \@partner_list;

  # get all valid bug states
  my @valid_states = @{
    $dbh->selectcol_arrayref(
      "SELECT value FROM bug_status WHERE isactive=1 ORDER BY sortkey")
      || []
  };

  # fill bug status list box
  my @list_states = ();
  for (my $i = 0; $i < @valid_states; $i++) {
    my %param_data;
    $param_data{'name'}     = $valid_states[$i];
    $param_data{'value'}    = $valid_states[$i];
    $param_data{'selected'} = 1 if has_member($valid_states[$i], @$bug_status);
    push(@list_states, \%param_data);
  }
  $vars->{valid_states} = \@list_states;

  # get all verified states
  my $verified_select = <<VERIFIED;
        SELECT value
          FROM cf_verified
         WHERE isactive=1
         ORDER BY sortkey
VERIFIED

  my @cf_verified = ("None");
  push @cf_verified, @{$dbh->selectcol_arrayref($verified_select) || []};

  # fill verified list box
  my @verified_states = ();
  for (my $i = 0; $i < @cf_verified; $i++) {
    my %param_data;
    $param_data{'name'}     = $cf_verified[$i];
    $param_data{'value'}    = $cf_verified[$i];
    $param_data{'selected'} = 1 if has_member($cf_verified[$i], @verified);
    push(@verified_states, \%param_data);
  }
  $vars->{verified_states} = \@verified_states;

  # get all layered products
  my $layered_products_select = <<LAYERED;
        SELECT value
          FROM cf_layered_products
         WHERE isactive=1
         ORDER BY sortkey
LAYERED

  my @cf_layered_products = ("None");
## BUGBUG dependent prods
##    push @cf_layered_products, @{ $dbh->selectcol_arrayref($layered_products_select) || [] };

  # fill layered products list box
  my @layered_products_list = ();
  for (my $i = 0; $i < @cf_layered_products; $i++) {
    my %param_data;
    $param_data{'name'}     = $cf_layered_products[$i];
    $param_data{'value'}    = $cf_layered_products[$i];
    $param_data{'selected'} = 1
      if has_member($cf_layered_products[$i], @layered_products);
    push(@layered_products_list, \%param_data);
  }
  $vars->{layered_products_list} = \@layered_products_list;

  # Get all external bugzilla description
  my $ex_bz_select = <<EXTERNAL;
        SELECT description
          FROM external_bugzilla
         ORDER BY description
EXTERNAL

  my $external_bz = $dbh->selectcol_arrayref($ex_bz_select) || [];
  my @external_bz_list;
  foreach my $name (@{$external_bz}) {
    my %param_data;
    $param_data{'name'}     = $name;
    $param_data{'value'}    = $name;
    $param_data{'selected'} = 1 if has_member($name, $external_tracker);
    push(@external_bz_list, \%param_data);
  }
  $vars->{external_bz_list} = \@external_bz_list;

  my $itr_select = <<ITR;
        SELECT value
          FROM cf_internal_target_release
         ORDER BY sortkey
ITR

  my $itr = $dbh->selectcol_arrayref($itr_select) || [];
  my @cf_internal_target_release_list;
  foreach my $value (@{$itr}) {
    my %param_data;
    $param_data{'name'}     = $value;
    $param_data{'value'}    = $value;
    $param_data{'selected'} = 1 if has_member($value, @cf_internal_target_release);
    push(@cf_internal_target_release_list, \%param_data);
  }
  $vars->{cf_internal_target_release_list} = \@cf_internal_target_release_list;


  # get list of configured conditional naks
  my $nak_select = <<CONDITIONAL_NAKS;
        SELECT value
          FROM cf_conditional_nak
         WHERE isactive=1
         ORDER BY sortkey
CONDITIONAL_NAKS

  my @conditional_naks = ("None");
  push @conditional_naks, @{$dbh->selectcol_arrayref($nak_select) || []};

  # fill conditional naks list box
  my @bug_naks  = Bugzilla->cgi->param('conditional_nak');
  my @list_naks = ();
  for (my $i = 0; $i < @conditional_naks; $i++) {
    my %param_data;
    $param_data{'name'}  = $conditional_naks[$i];
    $param_data{'value'} = $conditional_naks[$i];
    my $j = 0;
    while ($j < @bug_naks) {
      if ($conditional_naks[$i] eq $bug_naks[$j]) {
        $param_data{'selected'} = 1;
        last;
      }
      $j++;
    }
    push(@list_naks, \%param_data);
  }
  $vars->{conditional_naks} = \@list_naks;

  # get list of configured qe conditional naks
  my $qe_nak_select = <<QE_CONDITIONAL_NAKS;
        SELECT value
          FROM cf_qe_conditional_nak
         WHERE isactive=1
         ORDER BY sortkey
QE_CONDITIONAL_NAKS

  my @qe_conditional_naks = ("None");
  push @qe_conditional_naks, @{$dbh->selectcol_arrayref($qe_nak_select) || []};

  # fill qe conditional naks list box
  my @bug_qe_naks  = Bugzilla->cgi->param('qe_conditional_nak');
  my @list_qe_naks = ();
  for (my $i = 0; $i < @qe_conditional_naks; $i++) {
    my %param_data;
    $param_data{'name'}  = $qe_conditional_naks[$i];
    $param_data{'value'} = $qe_conditional_naks[$i];
    my $j = 0;
    while ($j < @bug_qe_naks) {
      if ($qe_conditional_naks[$i] eq $bug_qe_naks[$j]) {
        $param_data{'selected'} = 1;
        last;
      }
      $j++;
    }
    push(@list_qe_naks, \%param_data);
  }
  $vars->{qe_conditional_naks} = \@list_qe_naks;

  # fill sorting selection box
  my $sort_by_assigned = 0;
  my %sort_bys         = ();
  for (my $i = 0; $i < @{$self->{config}->{SORT_BY}}; $i++) {
    my $config_name = @{$self->{config}->{SORT_BY}}[$i];
    my %param_data;
    $param_data{'name'}  = $config_name;
    $param_data{'value'} = $i;
    if ($sort_by && ($i eq $sort_by)) {
      $param_data{'selected'} = 1;
      if ($config_name =~ /assigned_to/) {

        # user requested a sort-by assigned_to
        $sort_by_assigned = 1;
      }
    }
    $sort_bys{$config_name} = \%param_data;
  }
  $vars->{sort_bys} = \%sort_bys;

  # flag filter values
  my %acks;
  $self->determine_flag_filter(\%acks);

  foreach my $filter (keys %acks) {
    if ($filter !~ /_ack$/) {
      my $fid = $filter . '_flag';

      if ($acks{$filter}) {
        $vars->{$fid} = $acks{$filter};
      }
    }
  }

  my @ack_flag_filter = ();
  my @ack_flags       = @{$self->{config}->{ACK_FLAGS}};
  foreach my $ack_flag (@ack_flags) {
    push(@ack_flag_filter,
      {name => $ack_flag, value => ($acks{$ack_flag} ? $acks{$ack_flag} : undef),});
  }

  $vars->{ack_flags} = \@ack_flag_filter;

  # component values
  $vars->{bug_component}         = $component;
  $vars->{bug_component_exclude} = 1 if $component_not;
  $vars->{approved_components}   = 1 if $approved_components;
  $vars->{capacity_components}   = 1 if $capacity_components;
  $vars->{unassigned_components} = 1 if $unassigned_components;

  $vars->{keyword}       = $keyword_filter if $keyword_filter;
  $vars->{keyword_logic} = $keyword_logic  if $keyword_logic;

  $vars->{blocks}       = $blocks_filter if $blocks_filter;
  $vars->{blocks_depth} = $blocks_depth  if $blocks_depth;

  $vars->{external_tracker_exclusive} = $external_tracker_exclusive
    if $external_tracker_exclusive && $external_tracker;

  $vars->{mode}                = $customer_mode;
  $vars->{partner_mode}        = $partner_mode;
  $vars->{select_all_partners} = $select_all_partners;

  $vars->{hide_z_proposed} = $hide_z_proposed;
  $vars->{hide_z_approved} = $hide_z_approved;

  $vars->{user_logic} = $user_logic if $user_logic;

  return;
}

sub _header {
  my ($self, $value) = @_;

  my $header;

  # neither acks, nor specific flags are in fielddefs
  if ($value =~ /^acks$/) {
    $header = 'Acks';
  }
  elsif ($value =~ /^flag_/) {
    $header = $value;
    $header =~ s/^flag//;
    while ($header =~ /_(\w)/) {
      my $first    = $1;
      my $uc_first = uc($first);
      $header =~ s/_$first/ $uc_first/ if !($header =~ s/^_$first/$uc_first/);
    }
  }
  elsif ($value =~ /^agile_group/) {
    $header = 'Agile Group';
  }
  elsif ($value =~ /^short_short_desc$/) {
    $header = 'Short Summary';
  }
  elsif ($value =~ /^short_desc$/) {
    $header = 'Full Summary';
  }
  elsif ($value =~ /^external$/) {
    $header = 'Links';
  }
  elsif ($value =~ /^description$/) {
    $header = 'Description';
  }
  else {
    $header = $self->get_description($value) || $value;
  }

  return $header;
}

sub bug_view_form {
  my ($self, $vars) = @_;

##  my $update = $self->qpu_str('update');

  # retrieve all fields with display information
  my %view;
  my @columns = $self->my_columns(\%view);

  # headers
  my @headers = ();
  my %h_data;
  $h_data{'desc'}
    = $self->{config}->{'c_bug_id'}
    || $self->get_description('bug_id')
    || 'bug_id';
  $h_data{'name'} = 'bug_id';
  $h_data{'col'}  = 'bug_id';
  push(@headers, \%h_data);

  # show checkboxes
  my @cols = ();
  foreach my $view_column (@columns) {
    my %col_data;
    $col_data{'value'}   = 'c_' . $view_column;
    $col_data{'name'}    = $self->_header($view_column);
    $col_data{'checked'} = 1 if $view{$view_column};

    push(@cols, \%col_data);

    # table headers
    if ($view{$view_column}) {
      if ($view_column =~ /^acks$/) {
        my @ack_flags = @{$self->{config}->{ACK_FLAGS}};
        foreach my $flag (@ack_flags) {
          my %header_data;
          $header_data{'desc'} = $self->{config}->{'c_' . $flag};
          $header_data{'desc'} = uc($1) . $2
            if !$header_data{'desc'} && $flag =~ /([a-z])([a-z])[^_]*_ack$/;
          $header_data{'desc'} = $flag if !$header_data{'desc'};

          $header_data{'name'} = $flag;
          push(@headers, \%header_data);
        }

        my $update = $self->qpu_str('update');
        if ($update) {
          my %header_data;
          $update =~ s/^\w+-//;
          $header_data{'desc'} = $self->{config}->{'c_release'};
          $header_data{'desc'} = $update if !$header_data{'desc'};
          $header_data{'name'} = 'release';
          push(@headers, \%header_data);
        }
      }
      elsif ($view_column =~ /^flag_/) {
        my %header_data;
        $header_data{'desc'} = $self->{config}->{$col_data{'value'}};
        $header_data{'desc'} = $1
          if !$header_data{'desc'} && $col_data{'name'} =~ /^(\w\w)/;
        $header_data{'desc'} = $col_data{'name'} if !$header_data{'desc'};
        $header_data{'name'} = $view_column;
        push(@headers, \%header_data);
      }
      elsif ($view_column =~ /^(agile_group)$/) {
        my %header_data;
        $header_data{'desc'}
          = $self->{config}->{$col_data{'value'}} || $col_data{'name'};
        $header_data{'name'} = $view_column;
        push(@headers, \%header_data);
      }
      else {
        my %header_data;
        $header_data{'desc'}
          = $self->{config}->{$col_data{'value'}} || $col_data{'name'};
        $header_data{'name'} = $view_column;
        $header_data{'col'}  = $view_column;
        push(@headers, \%header_data);
      }
    }
  }

  $vars->{cols}    = \@cols;
  $vars->{headers} = \@headers;

  return;
}

sub flag_invert {
  my $val = shift;
  return '' if !$val;
  my @valay     = split(//, $val);
  my @inv_valay = ();
  foreach my $value (('+', '?', '-', 'X')) {
    push @inv_valay, $value if !has_member($value, @valay);
  }
  return join('', @inv_valay);
}

sub flag_union {
  my @vals = grep($_, @_);
  my %valays;
  for (my $i = 0; $i <= $#vals; $i++) {
    $vals[$i] = flag_invert($vals[$i]) if $vals[$i] =~ s/^!//;
    @{$valays{$i}} = split(//, $vals[$i]);
  }
  my @unionay = ();
  foreach my $value (('+', '?', '-', 'X')) {
    my $i = 0;
    while ($i < @vals) {
      last if !has_member($value, @{$valays{$i}});
      $i++;
    }
    push @unionay, $value if $i == @vals;
  }
  return join('', @unionay);
}

sub flag_canonic {
  my $val = shift;
  return $val if $val !~ /X/;
  return flag_invert($val) if $val =~ s/^!//;
  return '!' . flag_invert($val);
}

sub nak_invert {
  my @naks       = @_;
  my $dbh        = Bugzilla->dbh();
  my $nak_select = "SELECT value FROM cf_conditional_nak WHERE isactive=1";
  $nak_select
    .= " AND " . join(' AND ', map { "value!=" . $dbh->quote($_) } @naks)
    if @naks;

  return @{$dbh->selectcol_arrayref($nak_select) || []};
}

sub qe_nak_invert {
  my @naks       = @_;
  my $dbh        = Bugzilla->dbh();
  my $nak_select = "SELECT value FROM cf_qe_conditional_nak WHERE isactive=1";
  $nak_select
    .= " AND " . join(' AND ', map { "value!=" . $dbh->quote($_) } @naks)
    if @naks;

  return @{$dbh->selectcol_arrayref($nak_select) || []};
}

sub verified_invert {
  my @verified        = @_;
  my $dbh             = Bugzilla->dbh();
  my $verified_select = "SELECT value FROM cf_verified WHERE isactive=1";
  $verified_select
    .= " AND " . join(' AND ', map { "value!=" . $dbh->quote($_) } @verified)
    if @verified;

  return @{$dbh->selectcol_arrayref($verified_select) || []};
}

sub layered_products_invert {
  my @layered_products = @_;
  my $dbh              = Bugzilla->dbh();
  my $products_select  = "SELECT value FROM cf_layered_products WHERE isactive=1";
  $products_select
    .= " AND "
    . join(' AND ', map { "value!=" . $dbh->quote($_) } @layered_products)
    if @layered_products;

  return @{$dbh->selectcol_arrayref($products_select) || []};
}

sub list_query {
  my ($self, $param) = @_;

  my $dbh         = Bugzilla->dbh();
  my %flag_filter = ();
  %flag_filter = %{$param->{flag_filter}} if $param->{flag_filter};
  my $bz_search = $param->{bz_search};
  my $view      = undef;
  $view = $param->{view} if $param->{view};
  my $login_join = $param->{login_join} || 'assigned_to';
  my @columns    = ();

  # determine relevant columns of table bugs
  my $sth_columns = $dbh->column_info(undef, undef, 'bugs', '%');
  my $ref         = $dbh->selectall_hashref($sth_columns, 'COLUMN_NAME');

  for my $f (keys %{$ref}) {
    next if $f eq 'bug_id';
    my $fn = $f;

    # adjust field name for component_id and product_id
    $fn =~ s/_id$// if $fn eq 'component_id' || $fn eq 'product_id';
    $f = 'bugs.product_id AS product_id' if $fn eq 'product';

    push @columns, $f if $view->{$fn};
  }

  # extensions
  push @columns, 'resolution' if $view->{'bug_status'} && !$view->{'resolution'};
  push @columns, 'short_desc'
    if $view->{'short_short_desc'} && !$view->{'short_desc'};
  push @columns, 'target_milestone' if !$view->{'target_milestone'};

  my $select     = "SELECT DISTINCT bugs.bug_id";
  my $from       = "FROM bugs";
  my $where      = "";
  my $order_by   = "ORDER BY bugs.bug_id";
  my $profiles   = 0;
  my $components = 0;
  my $boolean    = 0;

  # append relevant columns
  $select .= ", " . join(', ', @columns) if @columns;

  # sort by

  if ( $param->{sort_by}
    && ($param->{sort_by} >= 0)
    && ($param->{sort_by} < @{$self->{config}->{SORT_BY}}))
  {
    my $sort_by = '';

    # Only add fields to sort order if they are displayed.
    foreach my $f (split(',', @{$self->{config}->{SORT_BY}}[$param->{sort_by}])) {
      if ($select =~ /$f/) {
        if ($sort_by ne '') {
          $sort_by .= ",$f";
        }
        else {
          $sort_by = "$f";
        }
      }
    }

    # component and login_name types are allowed and require table join

    if ($sort_by =~ /component$/ || $sort_by =~ /component[^_]/) {
      $sort_by =~ s/component$/components.name/;
      $sort_by =~ s/component([^_])/components.name$1/;
      $from   .= " JOIN components ON bugs.component_id=components.id";
      $select .= ", components.name" unless ($select =~ /components\.name/);

      $components = 1;
    }

    if ($sort_by =~ /assigned_to/) {
      $from .= " JOIN profiles ON $login_join=profiles.userid";
      $profiles = 1;
    }

    # special sorting considerations

    if ($sort_by =~ /pm_score/) {

      # force conversion of cf_pm_score to int, as it is varchar
      $sort_by =~ s/cf_pm_score/CONVERT(cf_pm_score,UNSIGNED) DESC/
        || $sort_by =~ s/pm_score/CONVERT(cf_pm_score,UNSIGNED) DESC/;
    }

    # bug_id may be ambiguous, prepend with table bugs

    if ($sort_by =~ /bug_id/) {
      $sort_by =~ s/bug_id/bugs.bug_id/;
    }

    # sort by bug severity

    if ($sort_by =~ /bug_severity/) {
      $from .= " JOIN bug_severity ON bugs.bug_severity=bug_severity.value";
      $sort_by =~ s/bug_severity/bug_severity.sortkey/;
    }

    $order_by = "ORDER BY " . $sort_by;
  }

  # flags

  if ($param->{update}) {
    my $update = $param->{update};
    my $list   = $param->{list};
    my $type   = $param->{type};

    # consider update flag, if not already considered
    $flag_filter{$update} = '' if !$flag_filter{$update};

    if (!$param->{view_version} || $list eq 'TOTAL' || $list eq 'RAW') {

      # transfer update/list/type settings to flag filter
      if ($list eq 'ACK' || ($param->{view_version} && $list ne 'RAW')) {
        if (!$type) {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+?';
          }
        }
        elsif ($type eq 'PROPOSED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '?';
          }
        }
        elsif ($type eq 'APPROVED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+';
          }
          if ($flag_filter{blocker}) {
            $flag_filter{blocker} = flag_union('!+', $flag_filter{blocker});
            return () if !$flag_filter{blocker};    # zero bugs
          }
          else {
            $flag_filter{blocker} = '!+';
          }
        }
        elsif ($type eq 'BLOCKER') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+';
          }
          if ($flag_filter{blocker}) {
            $flag_filter{blocker} = flag_union('+', $flag_filter{blocker});
            return () if !$flag_filter{blocker};    # zero bugs
          }
          else {
            $flag_filter{blocker} = '+';
          }
        }
        if ($flag_filter{fast}) {
          $flag_filter{fast} = flag_union('!+?', $flag_filter{fast});
          return () if !$flag_filter{fast};         # zero bugs
        }
        else {
          $flag_filter{fast} = '!+?';
        }
      }
      elsif ($list eq 'FAST') {
        if (!$type) {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+?';
          }
        }
        elsif ($type eq 'PROPOSED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '?';
          }
        }
        elsif ($type eq 'APPROVED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+';
          }
        }
        if ($flag_filter{fast}) {
          $flag_filter{fast} = flag_union('+?', $flag_filter{fast});
          return () if !$flag_filter{fast};         # zero bugs
        }
        else {
          $flag_filter{fast} = '+?';
        }
      }
      elsif ($list eq 'NAK') {
        if ($flag_filter{$update}) {
          $flag_filter{$update} = flag_union('-', $flag_filter{$update});
          return () if !$flag_filter{$update};      # zero bugs
        }
        else {
          $flag_filter{$update} = '-';
        }
      }
    }
    elsif ($list ne 'ACK') {                        # extended flag list
      if ($param->{combined}
        && ($type =~ /^(PROPOSED|APPROVED)_(PROPOSED|APPROVED|REJECTED)$/))
      {
        $type =~ s/_((PROPOSED|APPROVED|REJECTED))$//;
        my $release_type = $1;

        if ($release_type eq 'PROPOSED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '?';
          }
        }
        elsif ($release_type eq 'APPROVED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+';
          }
        }
        elsif ($release_type eq 'REJECTED') {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('-', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '-';
          }
        }
      }

      # set flag values for the given list
      my $flag = lc($list);

      if ($flag =~ /^blocker-(.+)$/) {              # blocker references milestone
        my $milestone_filter = $1;

        if ($milestone_filter ne "no-target") {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "target_milestone=" . $dbh->quote($milestone_filter);

          # advanced query
          my %clause = (
            field    => 'target_milestone',
            operator => 'equals',
            value    => $milestone_filter
          );
          $bz_search->add(\%clause);
        }
        else {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "target_milestone='---'";

          # advanced query
          my %clause
            = (field => 'target_milestone', operator => 'equals', value => '---');
          $bz_search->add(\%clause);
        }

        $flag = 'blocker';
      }

      if (!$type) {
        if ($flag_filter{$flag}) {
          $flag_filter{$flag} = flag_union('+?', $flag_filter{$flag});
          return () if !$flag_filter{$flag};    # zero bugs
        }
        else {
          $flag_filter{$flag} = '+?';
        }
        if (!$param->{combined}) {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+?';
          }
        }
      }
      elsif ($type eq 'PROPOSED') {
        if ($flag_filter{$flag}) {
          $flag_filter{$flag} = flag_union('?', $flag_filter{$flag});
          return () if !$flag_filter{$flag};        # zero bugs
        }
        else {
          $flag_filter{$flag} = '?';
        }
        if (!$param->{combined}) {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+?', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+?';
          }
        }
      }
      elsif ($type eq 'APPROVED') {
        if ($flag_filter{$flag}) {
          $flag_filter{$flag} = flag_union('+', $flag_filter{$flag});
          return () if !$flag_filter{$flag};        # zero bugs
        }
        else {
          $flag_filter{$flag} = '+';
        }
        if (!$param->{combined}) {
          if ($flag_filter{$update}) {
            $flag_filter{$update} = flag_union('+', $flag_filter{$update});
            return () if !$flag_filter{$update};    # zero bugs
          }
          else {
            $flag_filter{$update} = '+';
          }
        }
      }
    }
    else {                                          # extended no-flag list
      if ($param->{combined}
        && ($type =~ /^(PROPOSED|APPROVED)_(PROPOSED|APPROVED|REJECTED)$/))
      {
        $type =~ s/_(PROPOSED|APPROVED|REJECTED)$//;
      }

      if (!$type) {
        if ($flag_filter{$update}) {
          $flag_filter{$update} = flag_union('+?', $flag_filter{$update});
          return () if !$flag_filter{$update};      # zero bugs
        }
        else {
          $flag_filter{$update} = '+?';
        }
      }
      elsif ($type eq 'PROPOSED') {
        if ($flag_filter{$update}) {
          $flag_filter{$update} = flag_union('?', $flag_filter{$update});
          return () if !$flag_filter{$update};      # zero bugs
        }
        else {
          $flag_filter{$update} = '?';
        }
      }
      elsif ($type eq 'APPROVED') {
        if ($flag_filter{$update}) {
          $flag_filter{$update} = flag_union('+', $flag_filter{$update});
          return () if !$flag_filter{$update};      # zero bugs
        }
        else {
          $flag_filter{$update} = '+';
        }
      }

      foreach my $flag (@{$self->{config}->{FLAG_LIST}}) {
        $flag = lc($flag);
        next if $flag =~ /^blocker-/;               # ignore composite blockers
        if ($flag_filter{$flag}) {
          $flag_filter{$flag} = flag_union('!+?', $flag_filter{$flag});
          return () if !$flag_filter{$flag};        # zero bugs
        }
        else {
          $flag_filter{$flag} = '!+?';
        }
      }
    }
  }

  # check whether to hide z stream bugs

  if ($param->{hide_z_proposed} && $param->{hide_z_approved}) {
    $flag_filter{'%.z'} = '!+?';
  }
  elsif ($param->{hide_z_proposed}) {
    $flag_filter{'%.z'} = '!?';
  }
  elsif ($param->{hide_z_approved}) {
    $flag_filter{'%.z'} = '!+';
  }

  my @flag_or  = ();
  my @flag_not = ();
  foreach my $flag (keys %flag_filter) {
    if ($param->{ack_or} && $flag =~ /_ack$/) {
      push @flag_or, $flag;
      next;
    }

    my $flag_query
      = "SELECT flags.bug_id FROM flags, flagtypes WHERE flags.type_id=flagtypes.id";
    if ($flag !~ /%/) {
      $flag_query .= "  AND flagtypes.name=" . $dbh->quote($flag);
    }
    else {
      $flag_query .= "  AND flagtypes.name like " . $dbh->quote($flag);
    }
    my $negate = 0;
    my $value  = '';
    if ($flag_filter{$flag}) {
      $value  = flag_canonic($flag_filter{$flag});
      $negate = ($value =~ s/^!//);
      $flag_query
        .= " AND ("
        . join(' OR ', map { "flags.status=" . $dbh->quote($_) } split(//, $value))
        . ")";
    }
    $where .= $where ? " AND " : "WHERE ";
    $where .= "bugs.bug_id" . ($negate ? " NOT " : " ") . "IN ($flag_query)";

    # advanced query

    if ($negate) {
      if ($flag !~ /%/) {
        if ($value) {
          push @flag_not, map { $flag . $_ } split(//, $value);
        }
        else {
          push @flag_not, $flag;
        }
      }
      else {

        # requires construction of regular expression

        my $regexp = '^' . $flag;
        $regexp =~ s/\./\\\./g;
        $regexp =~ s/%/\.\*/g;
        $regexp .= '[' . $value . ']$';

        my %clause
          = (field => 'flagtypes.name', operator => 'notregexp', value => $regexp);
        $bz_search->add(\%clause);
      }

      next;
    }

    $value = '+-?' if !$value;

    my $fo = '';
    my $fv = '';

    if ($flag !~ /%/) {
      $fo = length($value) == 1 ? 'equals' : 'anyexact';
      $fv = join(',', map { $flag . $_ } split(//, $value));
    }
    else {

      # requires construction of regular expression

      my $regexp = '^' . $flag;
      $regexp =~ s/\./\\\./g;
      $regexp =~ s/%/\.\*/g;
      $regexp .= '[' . $value . ']$';

      $fo = 'regexp';
      $fv = $regexp;
    }

    my %clause = (field => 'flagtypes.name', operator => $fo, value => $fv);
    $bz_search->add(\%clause);
  }

  if (@flag_not) {
    my %clause = (
      field    => 'flagtypes.name',
      operator => 'anyexact',
      value    => join(',', @flag_not),
      negate   => 1,
    );

    $bz_search->add(\%clause);
  }

  if (@flag_or) {
    my @flag_string   = ();
    my $flag_or_query = "";
    my $or_clause
      = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
    foreach my $flag (@flag_or) {
      $flag_or_query .= " OR " if $flag_or_query;

      my $flag_query
        = "SELECT flags.bug_id FROM flags, flagtypes WHERE flags.type_id=flagtypes.id AND flagtypes.name="
        . $dbh->quote($flag);
      my $negate = 0;
      my $value  = '';
      if ($flag_filter{$flag}) {
        $value  = flag_canonic($flag_filter{$flag});
        $negate = ($value =~ s/^!//);
        $flag_query
          .= " AND ("
          . join(' OR ', map { "flags.status=" . $dbh->quote($_) } split(//, $value))
          . ")";
      }
      $flag_or_query
        .= "(bugs.bug_id" . ($negate ? " NOT " : " ") . "IN ($flag_query))";

      # advanced query

      $value = '+-?' if !$value;
      my $fo         = ($negate) ? "notregexp" : "regexp";
      my $fv         = '^' . $flag . '[' . $value . ']$';
      my $sub_clause = {field => 'flagtypes.name', operator => $fo, value => $fv};
      $or_clause->add($sub_clause);
    }
    $where .= $where ? " AND " : "WHERE ";
    $where .= "($flag_or_query)";

    # advanced query

    $bz_search->add($or_clause);
  }

  if ($param->{customer_mode}) {
    $where .= $where ? " AND " : "WHERE ";
    $where
      .= "(bugs.bug_id IN (SELECT ext_bz_bug_map.bug_id FROM ext_bz_bug_map WHERE ext_bz_bug_map.ext_bz_id=60))";

    # advanced query
    my %clause = (
      field    => 'external_bugzilla.description',
      operator => 'equals',
      value    => 'Red Hat Customer Portal'
    );
    $bz_search->add(\%clause);
  }

  if ($param->{external_tracker}) {
    $where .= $where ? " AND " : "WHERE ";
    $where
      .= "(bugs.bug_id IN (SELECT ext_bz_bug_map.bug_id FROM ext_bz_bug_map, external_bugzilla WHERE ext_bz_bug_map.ext_bz_id=external_bugzilla.id AND external_bugzilla.description="
      . $dbh->quote($param->{external_tracker}) . "))";

    # advanced query
    my %clause = (
      field    => 'external_bugzilla.description',
      operator => 'equals',
      value    => $param->{external_tracker}
    );
    $bz_search->add(\%clause);
  }

  if ($param->{partner_mode}) {
    $from .= " JOIN bug_cf_partner ON bugs.bug_id=bug_cf_partner.bug_id";

    # advanced query
    my $and_clause  = Bugzilla::Extension::Releases::ACKCustomSearch->new();
    my $sub_clause1 = {field => 'cf_partner', operator => 'isnotnull', value => ''};
    $and_clause->add($sub_clause1);

    # specific partners

    my $partners = $param->{partners} || [];

    if (@$partners && !has_member('All', @$partners)) {
      $where
        .= " AND ("
        . join(' OR ', map { "bug_cf_partner.value=" . $dbh->quote($_) } @$partners)
        . ")";

      # advanced query
      my $fo          = (@$partners == 1) ? 'equals' : 'anyexact';
      my $fv          = join(',', @$partners);
      my $sub_clause2 = {field => 'cf_partner', operator => $fo, value => $fv};
      $and_clause->add($sub_clause2);
    }

    $bz_search->add($and_clause);
  }

  if (!is_empty_array($param->{priorities})) {
    $where .= $where ? " AND " : "WHERE ";
    $where .= "("
      . join(' OR ',
      map { "bugs.priority=" . $dbh->quote($_) } @{$param->{priorities}})
      . ")";

    # advanced query
    my %clause = (
      field    => 'priority',
      operator => 'anyexact',
      value    => join(',', @{$param->{priorities}})
    );
    $bz_search->add(\%clause);
  }

  # member or group restriction

  if ($param->{member} || !is_empty_array($param->{group})) {
    my $profile_type = $login_join;

    if (!$profiles) {
      $from .= " JOIN profiles ON $login_join=profiles.userid";
      $profiles = 1;
    }

    my @conditions;
    my @clauses;

    if ($param->{member}) {
      push @conditions, "profiles.login_name=" . $dbh->quote($param->{member});

      push @clauses,
        {field => $profile_type, operator => 'equals', value => $param->{member}};
    }

    if (!is_empty_array($param->{group})) {
      my $emails = group_user_email($param->{group});

      if (@$emails) {
        my $condition = "";
        my $operator  = "IN";
        my $clause    = {
          field    => $profile_type,
          operator => 'anyexact',
          value    => join(',', @$emails)
        };

        if ($param->{group_negate}) {
          $operator = "NOT IN";
          $clause->{negate} = 1;
        }

        $condition .= "profiles.login_name $operator ("
          . join(',', map { $dbh->quote($_) } @$emails) . ")";

        push @conditions, $condition;
        push @clauses,    $clause;
      }
    }

    if (@conditions) {
      my $joiner = "";
      my $join_clause;
      if ($param->{user_logic} eq "all") {
        $joiner      = "AND";
        $join_clause = Bugzilla::Extension::Releases::ACKCustomSearch->new();
      }
      else {
        $joiner = "OR";
        $join_clause
          = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => $joiner});
      }

      $where .= $where ? " AND " : "WHERE ";
      $where .= "( " . join(" $joiner ", @conditions) . " )";

      map { $join_clause->add($_) } @clauses;
      $bz_search->add($join_clause);
    }

    # return all the bugs if empty users and negate is true
    elsif (!$param->{group_negate}) {

      # zero bugs
      return ();
    }
  }

  if ($param->{bug_status}) {
    my @bug_status = @{$param->{bug_status}};
    if (@bug_status) {
      $where .= $where ? " AND " : "WHERE ";
      $where
        .= "("
        . join(' OR ', map { "bugs.bug_status=" . $dbh->quote($_) } @bug_status)
        . ")";

      # advanced query
      my %clause = (
        field    => 'bug_status',
        operator => 'anyexact',
        value    => join(',', @bug_status)
      );
      $bz_search->add(\%clause);
    }
  }

  if ($param->{component}) {
    my @component_strings = ();
    my $component_string  = $param->{component};

    # extract double quoted strings first
    while ($component_string =~ s/"([^"]+)"//) {
      push @component_strings, $1;
    }
    push @component_strings, split(/[, ]+/, $component_string);
    my @like  = ();
    my @equal = ();
    foreach my $c_string (@component_strings) {
      if ($c_string =~ s/\*/%/g) {
        push @like, $c_string;
      }
      else {
        push @equal, $c_string;
      }
    }
    if (@like || @equal) {
      if (!$components) {
        $from .= " JOIN components ON bugs.component_id=components.id";
        $components = 1;
      }

      $where .= $where ? " AND " : "WHERE ";
      $where .= "(NOT " if $param->{component_not};
      $where .= "(";

      if (@equal) {
        $where .= join(' OR ', map { "components.name=" . $dbh->quote($_) } @equal);
        $where .= ' OR ' if @like;
      }

      if (@like) {
        $where .= join(' OR ', map { "components.name like " . $dbh->quote($_) } @like);
      }

      $where .= ")";
      $where .= ")" if $param->{component_not};

      # advanced query

      if ($param->{component_not}) {
        my $and_clause = Bugzilla::Extension::Releases::ACKCustomSearch->new();

        if (@equal) {
          my $sub_clause1 = {
            field    => 'component',
            operator => 'anyexact',
            value    => join(',', @equal),
            negate   => 1
          };
          $and_clause->add($sub_clause1);
        }

        my $c = 0;
        foreach my $l (@like) {
          $l =~ s/%/.*/g;

          my $sub_clause2
            = {field => 'component', operator => 'notregexp', value => '^' . $l . '$'};
          $and_clause->add($sub_clause2);
        }

        $bz_search->add($and_clause);
      }
      else {
        my $or_clause
          = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});

        my $c = 0;
        if (@equal) {
          my $sub_clause1
            = {field => 'component', operator => 'anyexact', value => join(',', @equal)};
          $or_clause->add($sub_clause1);
        }

        foreach my $l (@like) {
          $l =~ s/%/.*/g;

          my $sub_clause2
            = {field => 'component', operator => 'regexp', value => '^' . $l . '$'};
          $or_clause->add($sub_clause2);
        }

        $bz_search->add($or_clause);
      }

      #$boolean++;
    }
  }

  # restrict by approved/capacity components if selected
  if (
    (
         $param->{approved_components}
      || $param->{capacity_components}
      || $param->{unassigned_components}
    )
    && (!(
         $param->{approved_components}
      && $param->{capacity_components}
      && $param->{unassigned_components}
    ))
    )
  {
    my $approved_components   = $param->{approved_components};
    my $capacity_components   = $param->{capacity_components};
    my $unassigned_components = $param->{unassigned_components};
    my %approved_component    = ();
    %approved_component = %{$param->{approved_component_list}}
      if $param->{approved_component_list};
    my %capacity_component = ();
    %capacity_component = %{$param->{capacity_component_list}}
      if $param->{capacity_component_list};
    my @ac_components = ();
    push @ac_components, keys %approved_component if $approved_components;
    push @ac_components, keys %capacity_component if $capacity_components;

    if ($unassigned_components) {
      my @uc_components = ();
      push @uc_components, keys %approved_component if !$approved_components;
      push @uc_components, keys %capacity_component if !$capacity_components;

      if (@uc_components) {
        $where .= $where ? " AND " : "WHERE ";
        $where .= "(" . join(' AND ', map {"component_id != $_"} @uc_components) . ")";

        # advanced query

        my $not_component_sql = "(" . join(' OR ', map {"id=$_"} @uc_components) . ")";
        my @advanced_not_components = @{
          $dbh->selectcol_arrayref(
            "SELECT name FROM components WHERE $not_component_sql")
            || []
        };

        if (@advanced_not_components) {
          my %clause = (
            field    => 'component',
            operator => 'anyexact',
            value    => join(',', @advanced_not_components),
            negate   => 1
          );
          $bz_search->add(\%clause);
        }
      }
    }
    elsif (@ac_components) {
      $where .= $where ? " AND " : "WHERE ";
      $where .= "(" . join(' OR ', map {"component_id = $_"} @ac_components) . ")";

      # advanced query

      my $component_sql = "(" . join(' OR ', map {"id=$_"} @ac_components) . ")";
      my @advanced_components
        = @{$dbh->selectcol_arrayref(
          "SELECT name FROM components WHERE $component_sql")
          || []};

      return () if !@advanced_components;    # zero bugs

      my %clause = (
        field    => 'component',
        operator => 'anyexact',
        value    => join(',', @advanced_components)
      );
      $bz_search->add(\%clause);

    }
    else {
      return ();                             # zero bugs
    }
  }

  # keyword section

  my @use_keywords     = ();
  my @not_use_keywords = ();

  # retrieve configured sql for the column and extract keywords, if exist
  my $sql_column = '';
  if ($param->{column}) {
    my $column     = $param->{column};
    my $id         = 'BLOCKER_COLUMN_' . $column;
    my $column_sql = $self->{config}->{$id};
    if (!defined $column_sql) {
      die 'Sql clause for column ' . $column . ' not defined in configuration file.';
    }
    if ($column_sql =~ /\s+OR\s+/) {
      die 'Sql clause for column ' . $column . ' is not supported: ' . $column_sql;
    }
    my @sql_conjunction = split(/\s+AND\s+/, $column_sql);
    foreach my $item (@sql_conjunction) {
      if ($item !~ /^[\w.]+\s*(=|!=)\s*'?[^']+'?$/) {
        die 'Sql clause within column ' . $column . ' is not supported: ' . $item;
      }

      if ($item =~ /^keyworddefs.name\s*=\s*'([^']+)'$/) {
        push @use_keywords, $1;
      }
      elsif ($item =~ /^keyworddefs.name\s*!=\s*'([^']+)'$/) {
        push @not_use_keywords, $1;
      }
      else {
        $sql_column .= " AND " if $sql_column;
        $sql_column .= $item;
      }
    }
  }

  if ($param->{keyword_filter}) {
    my @additional_keywords = split(/[, ]+/, $param->{keyword_filter});
    my $keyword_logic       = $param->{keyword_logic};

    if ($keyword_logic eq "all") {
      foreach my $additional_keyword (@additional_keywords) {
        push @use_keywords, $additional_keyword;
      }
    }
    elsif ($keyword_logic eq "none") {
      foreach my $additional_keyword (@additional_keywords) {
        push @not_use_keywords, $additional_keyword;
      }
    }
    else {    # ( $keyword_logic eq "any" )
      my $keyword_query
        = "SELECT keywords.bug_id FROM keywords, keyworddefs WHERE keywords.keywordid=keyworddefs.id";
      $keyword_query .= " AND (";
      $keyword_query .= join(' OR ',
        map { "keyworddefs.name=" . $dbh->quote($_) } @additional_keywords);
      $keyword_query .= ")";

      $where .= $where ? " AND " : "WHERE ";
      $where .= "bugs.bug_id IN ($keyword_query)";

      # advanced query
      my %clause = (
        field    => 'keywords',
        operator => 'anyexact',
        value    => join(',', @additional_keywords)
      );
      $bz_search->add(\%clause);
    }
  }

  foreach my $keyword (@use_keywords) {
    return () if has_member($keyword, @not_use_keywords);    # zero bugs

    my $keyword_query
      = "SELECT keywords.bug_id FROM keywords, keyworddefs WHERE keywords.keywordid=keyworddefs.id AND (keyworddefs.name="
      . $dbh->quote($keyword) . ")";

    $where .= $where ? " AND " : "WHERE ";
    $where .= "bugs.bug_id IN ($keyword_query)";

    # advanced query
    my %clause = (field => 'keywords', operator => 'equals', value => $keyword);
    $bz_search->add(\%clause);
  }

  foreach my $keyword (@not_use_keywords) {
    my $keyword_query
      = "SELECT keywords.bug_id FROM keywords, keyworddefs WHERE keywords.keywordid=keyworddefs.id AND (keyworddefs.name="
      . $dbh->quote($keyword) . ")";

    $where .= $where ? " AND " : "WHERE ";
    $where .= "bugs.bug_id NOT IN ($keyword_query)";

    # advanced query
    my %clause = (field => 'keywords', operator => 'notequals', value => $keyword);
    $bz_search->add(\%clause);
  }

  if ($sql_column) {
    $where .= $where ? " AND " : "WHERE ";
    $where .= '(' . $sql_column . ')';

    # advanced query
    my $and_clause = Bugzilla::Extension::Releases::ACKCustomSearch->new();

    my $and             = 0;
    my @sql_conjunction = split(/\s+AND\s+/, $sql_column);
    foreach my $item (@sql_conjunction) {
      if ($item =~ /^([\w]+)\s*([=!]+)\s*'?([^']+)'?$/) {
        my $ff     = $1;
        my $fo     = ($2 eq '=') ? 'equals' : 'notequals';
        my $fv     = $3;
        my %clause = (field => $ff, operator => $fo, value => $fv);
        $and_clause->add(\%clause);
      }
      else {
        die 'Sql clause is not supported: ' . $item;
      }
    }
    $bz_search->add($and_clause);
  }

  # conditional naks

  if ($param->{conditional_naks}) {
    my @bug_naks   = @{$param->{conditional_naks}};
    my @values     = grep($_ ne 'None', @bug_naks);
    my $naks_query = "";
    $naks_query
      = "SELECT bug_id FROM bug_cf_conditional_nak WHERE ("
      . join(' OR ', map { "value=" . $dbh->quote($_) } @values) . ")"
      if @values;

    if (@bug_naks > @values) {    # None
      my @inv_values = nak_invert(@values);

      if (@inv_values) {
        my $inv_naks_query = "SELECT bug_id FROM bug_cf_conditional_nak WHERE ("
          . join(' OR ', map { "value=" . $dbh->quote($_) } @inv_values) . ")";

        if (@values) {
          $where .= $where ? " AND " : "WHERE ";
          $where
            .= "(bugs.bug_id IN ($naks_query) OR bugs.bug_id NOT IN ($inv_naks_query))";

          # advanced query
          my $or_clause
            = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
          my $sub_clause1 = {
            field    => 'cf_conditional_nak',
            operator => (@values == 1) ? 'equals' : 'anyexact',
            value    => join(',', @values)
          };
          my $sub_clause2 = {
            field    => 'cf_conditional_nak',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          };
          $or_clause->add($sub_clause1);
          $or_clause->add($sub_clause2);
          $bz_search->add($or_clause);
        }
        else {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "bugs.bug_id NOT IN ($inv_naks_query)";

          # advanced query
          my %clause = (
            field    => 'cf_conditional_nak',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          );
          $bz_search->add(\%clause);
        }
      }
    }
    elsif (@values) {
      $where .= $where ? " AND " : "WHERE ";
      $where .= "bugs.bug_id IN ($naks_query)";

      # advanced query
      my $fo     = (@values == 1) ? 'equals' : 'anyexact';
      my $fv     = join(',', @values);
      my %clause = (field => 'cf_conditional_nak', operator => $fo, value => $fv);
      $bz_search->add(\%clause);
    }
  }

  # qe conditional naks

  if ($param->{qe_conditional_naks}) {
    my @bug_naks   = @{$param->{qe_conditional_naks}};
    my @values     = grep($_ ne 'None', @bug_naks);
    my $naks_query = "";
    $naks_query
      = "SELECT bug_id FROM bug_cf_qe_conditional_nak WHERE ("
      . join(' OR ', map { "value=" . $dbh->quote($_) } @values) . ")"
      if @values;

    if (@bug_naks > @values) {    # None
      my @inv_values = qe_nak_invert(@values);

      if (@inv_values) {
        my $inv_naks_query = "SELECT bug_id FROM bug_cf_qe_conditional_nak WHERE ("
          . join(' OR ', map { "value=" . $dbh->quote($_) } @inv_values) . ")";

        if (@values) {
          $where .= $where ? " AND " : "WHERE ";
          $where
            .= "(bugs.bug_id IN ($naks_query) OR bugs.bug_id NOT IN ($inv_naks_query))";

          # advanced query
          my $or_clause
            = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
          my $sub_clause1 = {
            field    => 'cf_qe_conditional_nak',
            operator => (@values == 1) ? 'equals' : 'anyexact',
            value    => join(',', @values)
          };
          my $sub_clause2 = {
            field    => 'cf_qe_conditional_nak',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          };
          $or_clause->add($sub_clause1);
          $or_clause->add($sub_clause2);
          $bz_search->add($or_clause);
        }
        else {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "bugs.bug_id NOT IN ($inv_naks_query)";

          # advanced query
          my %clause = (
            field    => 'cf_qe_conditional_nak',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          );
          $bz_search->add(\%clause);
        }
      }
    }
    elsif (@values) {
      $where .= $where ? " AND " : "WHERE ";
      $where .= "bugs.bug_id IN ($naks_query)";

      # advanced query
      my $fo     = (@values == 1) ? 'equals' : 'anyexact';
      my $fv     = join(',', @values);
      my %clause = (field => 'cf_qe_conditional_nak', operator => $fo, value => $fv);
      $bz_search->add(\%clause);
    }
  }

  # cf_layered_products
  ## BUGBUG switch to dependent products
  if (0 && $param->{layered_products}) {
    my @layered_products       = @{$param->{layered_products}};
    my @values                 = grep($_ ne 'None', @layered_products);
    my $layered_products_query = "";
    $layered_products_query
      = "SELECT bug_id FROM bug_cf_layered_products WHERE ("
      . join(' OR ', map { "value=" . $dbh->quote($_) } @values) . ")"
      if (@values);

    if (@layered_products > @values) {    # None
      my @inv_values = layered_products_invert(@values);

      if (@inv_values) {
        my $inv_layered_products_query
          = "SELECT bug_id FROM bug_cf_layered_products WHERE ("
          . join(' OR ', map { "value=" . $dbh->quote($_) } @inv_values) . ")";

        if (@values) {
          $where .= $where ? " AND " : "WHERE ";
          $where
            .= "(bugs.bug_id IN ($layered_products_query) OR bugs.bug_id NOT IN ($inv_layered_products_query))";

          # advanced query
          my $or_clause
            = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
          my $sub_clause1 = {
            field    => 'cf_layered_products',
            operator => (@values == 1) ? 'equals' : 'anyexact',
            value    => join(',', @values)
          };
          my $sub_clause2 = {
            field    => 'cf_layered_products',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          };
          $or_clause->add($sub_clause1);
          $or_clause->add($sub_clause2);
          $bz_search->add($or_clause);

        }
        else {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "bugs.bug_id NOT IN ($inv_layered_products_query)";

          # advanced query
          my %clause = (
            field    => 'cf_layered_products',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          );
          $bz_search->add(\%clause);
        }
      }
    }
    elsif (@values) {
      $where .= $where ? " AND " : "WHERE ";
      $where .= "bugs.bug_id IN ($layered_products_query)";

      # advanced query
      my $fo     = (@values == 1) ? 'equals' : 'anyexact';
      my %clause = (
        field    => 'cf_layered_products',
        operator => $fo,
        value    => join(',', @values)
      );
      $bz_search->add(\%clause);
    }
  }

  # cf_verified

  if ($param->{verified}) {
    my @verified = @{$param->{verified}};
    my @values   = grep($_ ne 'None', @verified);

    if (@verified > @values) {    # None
      my @inv_values = verified_invert(@values);

      if (@inv_values) {
        my $inv_verified_query = "SELECT bug_id FROM bug_cf_verified WHERE ("
          . join(' OR ', map { "value=" . $dbh->quote($_) } @inv_values) . ")";

        if (@values) {
          my $verified_query = "SELECT bug_id FROM bug_cf_verified WHERE ("
            . join(' OR ', map { "value=" . $dbh->quote($_) } @values) . ")";

          $where .= $where ? " AND " : "WHERE ";
          $where
            .= "(bugs.bug_id IN ($verified_query) OR bugs.bug_id NOT IN ($inv_verified_query))";

          # advanced query
          my $or_clause
            = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
          my $sub_clause1 = {
            field    => 'cf_verified',
            operator => (@values == 1) ? 'equals' : 'anyexact',
            value    => join(',', @values)
          };
          my $sub_clause2 = {
            field    => 'cf_verified',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          };
          $or_clause->add($sub_clause1);
          $or_clause->add($sub_clause2);
          $bz_search->add($or_clause);
        }
        else {
          $where .= $where ? " AND " : "WHERE ";
          $where .= "bugs.bug_id NOT IN ($inv_verified_query)";

          # advanced query
          my %clause = (
            field    => 'cf_verified',
            operator => 'nowordssubstr',
            value    => join(',', @inv_values)
          );
          $bz_search->add(\%clause);
        }
      }
    }
    elsif (@verified) {
      $from .= " JOIN bug_cf_verified ON bugs.bug_id=bug_cf_verified.bug_id";

      $where
        .= " AND ("
        . join(' OR ', map { "bug_cf_verified.value=" . $dbh->quote($_) } @verified)
        . ")";

      # advanced query
      my %clause = (
        field    => 'cf_verified',
        operator => (@verified == 1) ? 'equals' : 'anyexact',
        value    => join(',', @verified)
      );
      $bz_search->add(\%clause);
    }
  }

  # target milestones

  if ($param->{target_milestone}) {
    my @target_milestones = @{$param->{target_milestone}};

    if (@target_milestones) {
      $where .= $where ? " AND " : "WHERE ";
      $where
        .= '('
        . join(' OR ',
        map { "target_milestone=" . $dbh->quote($_) } @target_milestones)
        . ')';

      # advanced query
      my %clause = (
        field    => 'target_milestone',
        operator => (@target_milestones == 1) ? 'equals' : 'anyexact',
        value    => join(',', @target_milestones)
      );
      $bz_search->add(\%clause);
    }
  }

  # products

  if ($param->{products}) {
    my @products = @{$param->{products}};

    if (@products) {
      $where .= $where ? " AND " : "WHERE ";
      $where .= '(bugs.product_id IN (SELECT id FROM products WHERE ('
        . join(' OR ', map { "name=" . $dbh->quote($_) } @products) . ')))';

      # advanced query
      my %clause = (
        field    => 'product',
        operator => (@products == 1) ? 'equals' : 'anyexact',
        value    => join(',', @products)
      );
      $bz_search->add(\%clause);
    }
  }

  # blocks

  if ($param->{blocks_filter}) {

    # given aliases are allowed, retrieve bug_id list first

    my %blocks_bugs;
    my @blocks_filter = split(/[,\s]+/, $param->{blocks_filter});
    foreach my $bug (@blocks_filter) {
      if ($bug =~ /^\d+$/) {    # bug_id
        $blocks_bugs{$bug} = 1;
      }
      else {                    # alias
        my @bug_ids = @{
          $dbh->selectcol_arrayref(
            "SELECT bug_id FROM bugs_aliases WHERE alias=" . $dbh->quote($bug)
            )
            || []
        };
        foreach my $bug_id (@bug_ids) {
          $blocks_bugs{$bug_id} = 1;
        }
      }
    }

    return () if !(keys %blocks_bugs);    # zero bugs

    # generate entire dependency tree up to max depth
    # take care to avoid potential loops

    my $blocks_depth   = $param->{blocks_depth} ? $param->{blocks_depth} : 1;
    my @blocks_current = keys %blocks_bugs;
    while (@blocks_current && $blocks_depth > 0) {
      my %blocks_new;
      my @bug_ids = @{
        $dbh->selectcol_arrayref(
          "SELECT dependencies.dependson FROM dependencies WHERE dependencies.blocked IN ("
            . join(',', @blocks_current) . ")"
          )
          || []
      };
      foreach my $bug_id (@bug_ids) {
        if (!$blocks_bugs{$bug_id}) {
          $blocks_bugs{$bug_id} = 1;
          $blocks_new{$bug_id}  = 1;
        }
      }
      @blocks_current = keys %blocks_new;
      $blocks_depth--;
    }

    # query

    my $blocks = join(',', keys %blocks_bugs);

    $where .= $where ? " AND " : "WHERE ";
    $where .= "bugs.bug_id IN ($blocks)";

    # advanced query
    my %clause = (field => 'blocked', operator => 'anyexact', value => $blocks);
    $bz_search->add(\%clause);
  }

  # arbitrary bug list

  if ($param->{bug_id}) {
    my @bug_id       = @{$param->{bug_id}};
    my $dependencies = $param->{dependencies};

    if (@bug_id) {
      $where .= $where ? " AND " : "WHERE ";
      my $bug_query = "(" . join(' OR ', map { "bugs.bug_id=" . $_ } @bug_id) . ")";

      # advanced query
      my $or_clause
        = Bugzilla::Extension::Releases::ACKCustomSearch->new({joiner => 'OR'});
      my $sub_clause1
        = {field => 'bug_id', operator => 'anyexact', value => join(',', @bug_id)};
      $or_clause->add($sub_clause1);

      if ($dependencies) {
        my $dependency_query = "SELECT dependson FROM dependencies WHERE ("
          . join(' OR ', map { "blocked=" . $_ } @bug_id) . ")";

        $where .= "($bug_query OR bugs.bug_id IN ($dependency_query))";

        # advanced query
        my $sub_clause2
          = {field => 'blocked', operator => 'anyexact', value => join(',', @bug_id)};
        $or_clause->add($sub_clause2);
      }
      else {
        $where .= $bug_query;
      }

      #$boolean++;

      $bz_search->add($or_clause);
    }
  }

  if (@{$param->{cf_internal_target_release}}) {
    my @itrs = @{$param->{cf_internal_target_release}};
    my $itrs = join(', ', map { $dbh->quote($_) } @itrs);
    $where .= $where ? " AND " : "WHERE ";
    $where .= "(bugs.cf_internal_target_release IN ($itrs))";
  }

  # BUG 1121484: Add bug visibility restrictions
  # Ideally, I'd like to pass this through RPC, but that would require
  # to much of a change
  my $user_id   = Bugzilla->user->id;
  my $group_ids = Bugzilla->user->groups_as_string();

  $from .= qq{
        LEFT JOIN bug_group_map AS security_map ON bugs.bug_id = security_map.bug_id
        LEFT JOIN cc AS security_cc
            ON bugs.bug_id = security_cc.bug_id AND security_cc.who = $user_id
    };
  $where .= qq{
        AND ( (security_map.group_id IS NULL OR security_map.group_id IN ($group_ids))
            OR (bugs.reporter_accessible = 1 AND bugs.reporter = $user_id)
            OR (bugs.cclist_accessible = 1 AND security_cc.who IS NOT NULL)
            OR bugs.assigned_to = $user_id
            OR bugs.qa_contact = $user_id
            OR bugs.docs_contact = $user_id
        )
    };

  return "$select $from $where $order_by";
}

sub prepare_flags {
  return Bugzilla->dbh()->prepare(<<QUERY_FLAGS);
        SELECT flagtypes.name,
               flags.status
          FROM flags,
               flagtypes
         WHERE flags.type_id=flagtypes.id
           AND flags.bug_id=?
QUERY_FLAGS
}

sub prepare_aliases {
  return Bugzilla->dbh()->prepare(<<QUERY_ALIAS);
        SELECT alias
          FROM bugs_aliases
         WHERE bug_id=?
QUERY_ALIAS
}

sub prepare_blocked {
  return Bugzilla->dbh()->prepare(<<QUERY_BLOCKED);
        SELECT blocked
          FROM dependencies
         WHERE dependson=?
      ORDER BY blocked
QUERY_BLOCKED
}

sub prepare_dependson {
  return Bugzilla->dbh()->prepare(<<QUERY_DEPENDSON);
        SELECT dependson
          FROM dependencies
         WHERE blocked=?
      ORDER BY dependson
QUERY_DEPENDSON
}

sub prepare_keywords {
  return Bugzilla->dbh()->prepare(<<QUERY_KEYWORDS);
        SELECT keyworddefs.name
          FROM keywords,
               keyworddefs
         WHERE keywords.bug_id=?
           AND keywords.keywordid=keyworddefs.id
QUERY_KEYWORDS
}

sub prepare_releases {
  return Bugzilla->dbh()->prepare(<<QUERY_RELEASES);
        SELECT value
          FROM bugs_release
         WHERE bug_id=?
QUERY_RELEASES
}

sub prepare_profiles {
  return Bugzilla->dbh()->prepare(<<FIND_PROFILES_MEMBERS);
            SELECT login_name,
                   realname,
                   userid,
                   disabledtext
              FROM profiles
             WHERE userid=?
FIND_PROFILES_MEMBERS
}

sub prepare_components {
  return Bugzilla->dbh()->prepare(<<QUERY_COMPONENT);
        SELECT name
          FROM components
         WHERE id=?
QUERY_COMPONENT
}

sub prepare_products {
  return Bugzilla->dbh()->prepare(<<QUERY_PRODUCT);
        SELECT name
          FROM products
         WHERE id=?
QUERY_PRODUCT
}

sub prepare_external {
  return Bugzilla->dbh()->prepare(<<QUERY_EXTERNAL);
        SELECT ext_bz_bug_map.ext_bz_bug_id, external_bugzilla.full_url, external_bugzilla.description, ext_bz_bug_map.ext_description
          FROM external_bugzilla, ext_bz_bug_map
         WHERE external_bugzilla.id=ext_bz_bug_map.ext_bz_id
           AND bug_id=?
      ORDER BY ext_bz_bug_map.ext_bz_bug_id
QUERY_EXTERNAL
}

sub prepare_description {
  return Bugzilla->dbh()->prepare(<<QUERY_COMMENTS);
        SELECT longdescs.thetext
          FROM longdescs
         WHERE longdescs.bug_id=?
      ORDER BY longdescs.comment_id
         LIMIT 1
QUERY_COMMENTS
}

sub db_list_tables {
  my ($dbh, $table) = @_;
  my $schema  = undef;
  my $catalog = undef;

  # get a list of table names from the database
  if ($dbh->{Driver}{Name} eq 'Pg') {
    $catalog = 'pg_catalog';
    $schema  = 'public';
  }

  my $sth        = $dbh->table_info($catalog, $schema, $table);
  my $table_info = $sth->fetchall_hashref('TABLE_NAME');
  my @table_list = keys(%{$table_info});

  return (@table_list);
}

sub prepare_custom {
  my ($self, $table_id) = @_;
  my $table_name = 'bug_' . $table_id;

# BUGBUG Need to keep this as it appears to be trying to use params as tables, and then later as columns ...
  my @tables = db_list_tables(Bugzilla->dbh(), $table_name);

  return () if !@tables;

  return Bugzilla->dbh()->prepare(<<QUERY_CUSTOM);
        SELECT value
          FROM $table_name
         WHERE bug_id=?
      ORDER BY value
QUERY_CUSTOM
}

sub approved_component_list {
  my ($self, $update, $approved_component, $capacity_component) = @_;

  my $dbh                     = Bugzilla->dbh();
  my $sth_approved_components = $dbh->prepare(<<APPROVED_COMPONENTS);
            SELECT component_id,
                   component_type
              FROM rh_release_components, rh_release, flagtypes
             WHERE flagtypes.id=rh_release.type_id
               AND rh_release.id=rh_release_components.release_id
               AND flagtypes.name='$update'
APPROVED_COMPONENTS

  $sth_approved_components->execute();

  while (my ($component_id, $component_type)
    = $sth_approved_components->fetchrow_array())
  {
    if ($component_type eq 'approved') {
      $approved_component->{$component_id} = 1;
    }
    elsif ($component_type eq 'capacity') {
      $capacity_component->{$component_id} = 1;
    }
  }

  return;
}

sub is_closed {
  my ($self, $bug_id) = @_;

  return @{Bugzilla->dbh()->selectcol_arrayref(
      "SELECT bug_id FROM bugs WHERE bug_id=$bug_id AND bug_status='CLOSED'")
      || []
  };
}

sub process_row {
  my ($self, $row, $param) = @_;

  my $all_rows                   = $param->{all_rows};
  my $all_assignees              = $param->{all_assignees};
  my $sth                        = $param->{sth};
  my $fields                     = $param->{fields};
  my $view                       = $param->{view};
  my $update                     = $param->{update};
  my $view_version               = $param->{view_version};
  my %acks                       = %{$param->{acks}};
  my $edit_mode                  = $param->{edit_mode};
  my $can_grant                  = $param->{can_grant};
  my $list                       = $param->{list};
  my %approved_component         = %{$param->{approved_component_list}};
  my %capacity_component         = %{$param->{capacity_component_list}};
  my $partner_mode               = $param->{partner_mode};
  my $type                       = $param->{type};
  my $can_edit                   = $param->{can_edit};
  my $external_tracker           = $param->{external_tracker};
  my $external_tracker_exclusive = $param->{external_tracker_exclusive};
  my $bug_id                     = $row->{bug_id};
  my $item_vars;
  my %row_data;
  $row_data{bug_id} = $bug_id;

  # retrieve flags
  my %flags;
  $sth->{flags}->execute($bug_id);
  my $flagsref = $sth->{flags}->fetchall_arrayref;
  for my $f (@$flagsref) {
    $flags{$f->[0]} = $f->[1];
  }

  # retrieve keywords
  $sth->{keywords}->execute($bug_id);
  my $kws = $sth->{keywords}->fetchall_arrayref;
  my %keyword;
  for my $k (@$kws) {
    $keyword{$k->[0]} = 1;
  }

  # css class
  my $class = '';
  $class .= ' bz_' . $row->{bug_severity} if ($row->{bug_severity});
  $class .= ' bz_' . $row->{priority}     if ($row->{priority});
  $class .= ' bz_' . $row->{bug_status}   if ($row->{bug_status});
  if ($keyword{Regression}) {
    $class .= ' bz_regression';
  }
  if ($keyword{FutureFeature}) {
    $class .= ' bz_enhancement';
  }
  $row_data{class}  = $class;
  $row_data{parity} = $param->{parity};

  # form switches
  if ($view_version || ($list eq $self->{config}->{TYPE_ACK})) {
    $row_data{fix_tick} = 1;
  }
  if ($view_version) {
    $row_data{no_blocker} = 1;
  }

  # columns
  my $may_edit = 0;
  my @cols     = ();
  my %column_data;
  $column_data{name}   = 'bug_id';
  $column_data{value}  = $bug_id;
  $column_data{strike} = $row->{bug_status} eq 'CLOSED';
  push(@cols, \%column_data);

  my @ack_flags = @{$self->{config}->{ACK_FLAGS}};

  foreach my $column (grep($view->{$_}, @$fields)) {
    my %col_data;
    $col_data{name} = $column;

    if ($column eq 'acks') {
      foreach my $ack (@ack_flags) {
        my %ack_col_data;
        $ack_col_data{name} = $ack;

        if ($flags{$ack}) {
          $ack_col_data{value} = $flags{$ack};
          $ack_col_data{title} = $ack . $flags{$ack};

          if ($flags{$ack} eq '?') {
            $may_edit++ if $can_grant->{$ack};
            $ack_col_data{edit} = $bug_id . '_' . $ack if ($can_grant->{$ack} && $edit_mode);
          }
        }
        else {
          $ack_col_data{value} = ' ';
        }

        push(@cols, \%ack_col_data);
      }

      if ($update) {
        my %update_col_data;
        $update_col_data{name}  = 'release';
        $update_col_data{title} = $update . $flags{$update} if $flags{$update};
        $update_col_data{value} = $flags{$update} if $flags{$update};

        push(@cols, \%update_col_data);
      }

      $col_data{name} = '';
    }
    elsif ($column eq 'cf_conditional_nak' || $column eq 'cf_qe_conditional_nak') {
      $sth->{$column}->execute($bug_id);
      my $naks = $sth->{$column}->fetchall_arrayref;
      if (@$naks) {
        $col_data{array}     = [map { $_->[0] } @$naks];
        $col_data{delimiter} = "\n";
      }
    }
    elsif ($column eq 'flag_blocker') {
      if ($flags{blocker}) {
        $col_data{value} = $flags{blocker};
        $col_data{title} = 'blocker' . $flags{blocker};
        if ($row->{target_milestone} && $row->{target_milestone} ne "---") {
          $col_data{title} .= " [" . $row->{target_milestone} . "]";
        }
      }
      else {
        $col_data{value} = ' ';
      }
    }
    elsif ($column =~ /^flag_/) {
      my $id = $column;
      $id =~ s/^flag_//;
      $col_data{value} = $flags{$id} || ' ';
      $col_data{title} = $id . $flags{$id} if $flags{$id};
    }
    elsif ($column eq 'external') {
      $sth->{$column}->execute($bug_id);
      my $ref = $sth->{$column}->fetchall_arrayref;
      $col_data{array} = [];
      for my $r (@$ref) {
        next
          if $external_tracker_exclusive
          && $external_tracker
          && $external_tracker ne $r->[2];
        my $et    = $r->[2];
        my $class = undef;
        if ($et eq 'Red Hat Errata Tool') {
          $class = "et_et";
        }
        elsif ($et eq 'Red Hat Product Errata') {
          $class = "et_access";
        }
        elsif ($et =~ /Red Hat Customer Portal/) {
          $class = "et_portal";
        }
        elsif ($et =~ /TCMS/) {
          $class = "et_tcms";
        }
        elsif ($et =~ /Polarion/) {
          $class = "et_polarion";
        }
        elsif ($et eq 'Red Hat Bugzilla') {
          $class = "et_rhbz";
        }
        elsif ($et eq 'Mozilla Foundation') {
          $class = "et_bmo";
        }
        elsif ($et =~ /Red Hat Knowledge Base/) {
          $class = "et_rhkb";
        }
        elsif ($et =~ /CPAN/) {
          $class = "et_cpan";
        }
        elsif ($et =~ /^IBM/) {
          $class = "et_ibm";
        }

        my $title = "[$et] ";
        if ($r->[3] && $r->[3] ne '') {
          $title .= $r->[3];
        }
        else {
          $title .= "No Summary";
        }
        my $url     = $r->[1];
        my $id      = $r->[0];
        my $tracker = Bugzilla::Extension::ExternalBugs::Bug->match(
          {bug_id => $bug_id, ext_bz_bug_id => $id});
        if ($tracker) {
          $url = $tracker->[0]->full_url();
        }
        push @{$col_data{array}},
          {id => $id, url => $url, title => $title, class => $class,};
      }
      $col_data{link} = 'external';
    }
    elsif ($column eq 'blocked') {
      $sth->{$column}->execute($bug_id);
      my $blocks = $sth->{$column}->fetchall_arrayref;
      $col_data{array} = [];
      for my $r (@$blocks) {
        my $closed = $self->is_closed($r->[0]) == 1 ? 1 : 0;
        $sth->{alias}->execute($r->[0]);
        my $aliases = $sth->{alias}->fetchall_arrayref;
        for my $rr (@$aliases) {
          push @{$col_data{array}}, {id => $r->[0], alias => $rr->[0], strike => $closed};
        }
        if (!@$aliases) {
          push @{$col_data{array}}, {id => $r->[0], strike => $closed};
        }
      }
    }
    elsif ($column eq 'dependson') {
      $sth->{$column}->execute($bug_id);
      my $ref = $sth->{$column}->fetchall_arrayref;
      $col_data{link}       = 'bugzilla';
      $col_data{count_link} = "?page=flag-bugs";
      $col_data{count_link} .= "&update=$update" if $update;
      $col_data{count_link} .= "&bug_id=" . join(',', map { $_->[0] } @$ref);
      $col_data{array} = [];
      for my $r (@$ref) {
        my $closed = $self->is_closed($r->[0]) == 1 ? 1 : 0;
        push @{$col_data{array}}, {id => $r->[0], strike => $closed};
      }
    }
    elsif ($column eq 'keywords') {
      $col_data{align} = 'left';
      if (@$kws) {
        $col_data{array} = [map { $_->[0] } @$kws];
      }
    }
    elsif ($column eq 'flagtypes.name') {
      $col_data{align} = 'left';
      if (@$flagsref) {
        $col_data{array} = [map { $_->[0] . $_->[1] } @$flagsref];
      }
    }
    elsif ($column eq 'cf_partner') {
      $sth->{$column}->execute($bug_id);
      my $pids = $sth->{$column}->fetchall_arrayref;
      if (@$pids) {
        $col_data{array}     = [map { $_->[0] } @$pids];
        $col_data{delimiter} = "\n";
      }
    }
    elsif ($column eq 'component') {
      my $component_id   = $row->{component_id};
      my $component_name = '';
      $sth->{$column}->execute($component_id);
      ($component_name) = $sth->{$column}->fetchrow_array;

      $col_data{value} = $component_name;
      if ($approved_component{$component_id}) {
        $col_data{status} = 'approved';
      }
      elsif ($capacity_component{$component_id}) {
        $col_data{status} = 'capacity';
      }
      else {
        $col_data{status} = 'undef';
      }
    }
    elsif ($column eq 'product') {
      my $product_id   = $row->{product_id};
      my $product_name = '';
      $sth->{$column}->execute($product_id);
      ($product_name) = $sth->{$column}->fetchrow_array;
      $col_data{value} = $product_name;
    }
    elsif ($column eq 'cf_pm_score') {
      my $is_pm = is_bugzilla_group('product_management');
      my $value = (defined $row->{$column}) ? $row->{$column} : 0;

      if ($is_pm && $edit_mode) {
        $col_data{edit} = $bug_id . '_' . $column;
      }

      $col_data{value} = $value;
    }
    elsif ($column eq 'assigned_to'
      || $column eq 'reporter'
      || $column eq 'qa_contact')
    {
      my $RX_LOGIN = qr/<([^<]{1,8})[^<]*\@[^>]+>/x;
      $sth->{profiles}->execute($row->{$column});
      my ($login, $name, $id, $disabledtext) = $sth->{profiles}->fetchrow_array;
      my $string = "";
      $string = "$name <$login>"            if $login;
      $string = "DISABLED ACCOUNT: $string" if $disabledtext;
      $col_data{title} = $string;
      $col_data{value} = !$disabledtext && $string =~ $RX_LOGIN ? $1 : "---";
    }
    elsif ($column eq 'agile_group') {
      my $RX_LOGIN = qr/<([^<]{1,8})[^<]*\@[^>]+>/x;
      $sth->{profiles}->execute($row->{'assigned_to'});
      my ($login, $name, $id, $disabledtext) = $sth->{profiles}->fetchrow_array;

      my $string = "";
      $string = "$name <$login>"            if $login;
      $string = "DISABLED ACCOUNT: $string" if $disabledtext;

      if (!$disabledtext && $string =~ $RX_LOGIN) {
        $col_data{value} = $login;
        $all_assignees->{$login} ||= 1;
      }
      else {
        $col_data{value} = "---";
      }
    }
    elsif ($column eq 'bug_status') {
      $col_data{value} = $row->{$column};
      if (($row->{$column} eq 'CLOSED') && $row->{resolution}) {
        $col_data{value} .= ': ' . $row->{resolution};
      }
      if ($row->{$column} =~ /^(\w\w\w\w)/) {
        $col_data{title} = $col_data{value};
        $col_data{value} = $1;
      }
    }
    elsif ($column eq 'bug_severity' || $column eq 'priority') {
      $col_data{value} = $row->{$column};
      if ($row->{$column} =~ /^(\w\w\w)/) {
        $col_data{sortkey} = $self->get_sort_key($col_data{value}, $column);
        $col_data{title}   = $col_data{value};
        $col_data{value}   = $1;
      }
    }
    elsif ($column eq 'short_short_desc') {
      $col_data{align} = 'left';
      if ($row->{short_desc} && $row->{short_desc} =~ /^(.{60})./) {
        $col_data{value} = $1;
        $col_data{title} = $row->{short_desc};
      }
      elsif ($row->{short_desc}) {
        $col_data{value} = $row->{short_desc};
        $col_data{title} = $row->{short_desc};
      }
    }
    elsif ($column =~ /_whiteboard$/ || $column =~ /_desc$/) {
      $col_data{align} = 'left';
      $col_data{value} = $row->{$column};
    }
    elsif ($column =~ /^cf_/ && defined $sth->{$column}) {
      $sth->{$column}->execute($bug_id);
      my $ref = $sth->{$column}->fetchall_arrayref;
      $col_data{array} = [map { $_->[0] } @$ref] if @$ref;
    }
    elsif ($column eq 'description') {
      $col_data{align} = 'left';
      $sth->{$column}->execute($bug_id);
      ($col_data{value}) = $sth->{$column}->fetchrow_array;
    }
    elsif ($column eq 'target_release') {
      $sth->{releases}->execute($bug_id);
      my $rel = $sth->{releases}->fetchall_arrayref;
      $col_data{align} = 'left';
      if (@$rel) {
        $col_data{array} = [map { $_->[0] } @$rel];
      }
    }
    else {
      $col_data{value} = $row->{$column};
    }

    push(@cols, \%col_data) if $col_data{name};
  }

  if ($may_edit) {
    $can_edit->{count} = 0 if !defined $can_edit->{count};
    $can_edit->{count} += 1;
  }

  $row_data{fields} = \@cols;
  $item_vars->{row} = \%row_data;

  push @{$all_rows}, $item_vars;

  return $bug_id;
}

sub my_default_columns {
  my ($self, $column) = @_;

  # globally configured default

  my @default_columns = @{$self->{config}->{VIEW} || []};
  my @config_columns  = ();
  for (my $c = 0; $c < @default_columns; $c++) {
    my $view_column = $default_columns[$c];
    my $show        = 1;
    $show = $1 if $view_column =~ s/=(\d)$//;

    $column->{$view_column} = $show;
    push @config_columns, $view_column;
  }

  my @user_columns = @{$self->user_columns() || []};

  return @config_columns if !@user_columns;

  # user configured default

  my @columns = ();
  for (my $c = 0; $c < @user_columns; $c++) {
    my $view_column = $user_columns[$c];
    my $show        = 1;
    $show = $1 if $view_column =~ s/=(\d)$//;

    # user configured column must be a globally configured column
    next if !has_member($view_column, @config_columns);

    $column->{$view_column} = $show;
    push @columns, $view_column;
  }

  # append globally configured columns as selectable columns
  foreach my $view_column (grep(!has_member($_, @columns), @config_columns)) {
    $column->{$view_column} = 0;
    push @columns, $view_column;
  }

  return @columns;
}

sub my_columns {
  my ($self, $column) = @_;

  my $q = Bugzilla->cgi();

  # legacy support for externally existing links
  my %legacy_map = (
    'blocks'              => 'blocked',
    'pm_score'            => 'cf_pm_score',
    'conditional_naks'    => 'cf_conditional_nak',
    'devel_whiteboard'    => 'cf_devel_whiteboard',
    'internal_whiteboard' => 'cf_internal_whiteboard',
    'partner'             => 'cf_partner'
  );
  my %legacy_invert_map;
  foreach my $key (keys %legacy_map) {
    $legacy_invert_map{$legacy_map{$key}} = $key;
  }

  # retrieve default columns
  my @columns = $self->my_default_columns($column);

  # determine whether any show column parameters are given in the url
  my $c = 0;
  while ($c < @columns) {
    last if ($q->param('c_' . $columns[$c]));
    last
      if $legacy_invert_map{$columns[$c]}
      && $q->param('c_' . $legacy_invert_map{$columns[$c]});    # legacy support
    $c++;
  }

  if ($c < @columns) {

    # show column parameters are given, overwrite default values
    for ($c = 0; $c < @columns; $c++) {
      if ($q->param('c_' . $columns[$c])) {
        $column->{$columns[$c]} = 1;
      }
      elsif ($legacy_invert_map{$columns[$c]}
        && $q->param('c_' . $legacy_invert_map{$columns[$c]}))
      {    # legacy support
        $column->{$columns[$c]} = 1;
      }
      else {
        $column->{$columns[$c]} = 0;
      }
    }
  }

  return @columns;
}

sub determine_flag_filter {
  my ($self, $flag_filter, $proper) = @_;
  my $update    = $self->qpu_str('update');
  my @ack_flags = @{$self->{config}->{ACK_FLAGS}};

  foreach my $flag (@ack_flags) {
    $flag_filter->{$flag} = $self->qpu_flag($flag) if $self->qpu_flag($flag);
  }

  my @available_flags = @{$self->{config}->{'AVAILABLE_FLAGS'}};
  foreach my $flag (@available_flags) {
    my $fid = $flag;
    $fid .= '_flag';
    $flag_filter->{$flag} = $self->qpu_flag($fid) if $self->qpu_flag($fid);
  }

  my $release = 'release';
  $release = $self->qpu_str('update') if $self->qpu_str('update') && $proper;
  $flag_filter->{$release} = $self->qpu_flag('release_flag')
    if $self->qpu_flag('release_flag');

  return;
}

sub bug_statuses {
  my ($self, $params) = @_;

  my $resolved_types = delete $params->{resolved_types} || ();
  my $first_state_in = delete $params->{first_state_in} || undef;
  my $resolved
    = (defined $params->{resolved}) ? delete $params->{resolved} : undef;

  my $dbh          = Bugzilla->dbh();
  my @valid_states = @{
    $dbh->selectcol_arrayref(
      "SELECT value FROM bug_status WHERE isactive=1 ORDER BY sortkey")
      || []
  };

  return @valid_states if !defined $resolved;

  my @ack_resolved = ();

  # backward numeric status compatible
  if (defined $first_state_in) {
    my $first_state = (@$resolved_types) ? $resolved_types->[0] : undef;

    $first_state = map_old_state($first_state)
      if (defined $first_state && $first_state =~ /^\d$/);

    if (!defined $first_state
      || (defined $first_state && !has_member($first_state, @valid_states)))
    {
      $first_state = $self->{config}->{'DEFAULT_ACK_RESOLVED_TYPE'} || 'CLOSED';
    }

    @ack_resolved = get_later_types($first_state, @valid_states);
  }
  elsif (@$resolved_types) {
    foreach my $type (@$resolved_types) {
      push @ack_resolved, $type if (has_member($type, @valid_states));
    }
  }

  return @ack_resolved if $resolved;

  return grep(!has_member($_, @ack_resolved), @valid_states);
}

sub flag_bugs {
  my ($self, $vars) = @_;

  my $q          = Bugzilla->cgi();
  my $action     = $self->qpu_str('action');
  my $combined   = $self->qpu_bool('combine', 0);
  my $update     = $self->qpu_str('update');
  my @ack_flags  = @{$self->{config}->{ACK_FLAGS}};
  my $can_grant  = $self->can_grant_acks($update, \@ack_flags);
  my $allow_edit = (%$can_grant) ? 1 : 0;

  my $edit_mode = ($allow_edit && defined $action && ($action eq 'Edit')) ? 1 : 0;
  my %can_edit;

  my $view_version = $self->qpu_int('view-version', 0);
  my $list         = $self->qpu_str('list');
  my $type         = $self->qpu_str('type');
  my $listtype     = $self->qpu_str('listtype');

  if (!$view_version && $listtype) {
    if ($listtype =~ /^(\w+)-(\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^(\w+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  elsif ($listtype) {
    if ($listtype =~ /^([\w-]+) (\w+)$/) {
      $list = $1;
      $type = $2;
    }
    elsif ($listtype =~ /^([\w-]+)$/) {
      $list = $1;
      $type = undef;
    }
    else {
      $list = undef;
      $type = undef;
    }
  }
  if (!$list) {
    $list = $self->{config}->{TYPE_ACK};
  }
  my $resolved
    = defined $q->param('resolved') ? $self->qpu_int('resolved') : undef;

  my @bug_status = $q->param('bug_status');

  my @resolved_types
    = defined $q->param('s-resolved-type') ? $q->param('s-resolved-type') : ();

  # backward compatible
  my $first_state_in = 0;
  if (!@resolved_types && defined $self->qpu_str('resolved-type')) {
    $first_state_in = 1;
    push @resolved_types, $q->param('resolved-type');
  }

  if (!@bug_status && defined $resolved && @resolved_types) {
    @bug_status = $self->bug_statuses({
      resolved_types => \@resolved_types,
      resolved       => $resolved,
      first_state_in => $first_state_in
    });
  }

  my $column_set     = $self->qpu_str('column-set');
  my $column         = $self->qpu_str('column');
  my $keyword_filter = $self->qpu_str('keyword', '');
  my $keyword_logic  = $self->qpu_str('keyword_logic', 'any');
  my $blocks_filter  = $self->qpu_str('blocks', '');
  my $blocks_depth   = $self->qpu_int('blocks_depth', 0);
  my @group_params   = $q->param('group');

  my $selected_groups = [$q->param('group')];

  my $group_negate = $self->qpu_int('group_negate', 0);
  my $member       = $self->qpu_str('member');
  $member .= '@redhat.com' if $member && $member !~ /\@/;
  my $component             = $self->qpu_str('component');
  my $component_not         = $self->qpu_bool('component_exclude');
  my $approved_components   = $self->qpu_bool('approved_components');
  my $capacity_components   = $self->qpu_bool('capacity_components');
  my $unassigned_components = $self->qpu_bool('unassigned_components');
  my $external_tracker      = $self->qpu_str('external_tracker');
  my $external_tracker_exclusive
    = $self->qpu_bool('external_tracker_exclusive', 0);
  my $customer_mode = $self->qpu_bool('mode',   0);
  my $ack_or        = $self->qpu_bool('ack_or', 0);
  my $hide_z_proposed            = $self->qpu_bool('hide_z_proposed');
  my $hide_z_approved            = $self->qpu_bool('hide_z_approved');
  my @cf_internal_target_release = $q->param('cf_internal_target_release');

  my %acks;
  $self->determine_flag_filter(\%acks, 1);

  my $sort_by          = $self->qpu_int('sort', 0);
  my $partner_mode     = $self->qpu_bool('partner', 0);
  my @partners         = $q->param('cf_partner');
  my @bug_naks         = $q->param('conditional_nak');
  my @bug_qe_naks      = $q->param('qe_conditional_nak');
  my @bug_targets      = $q->param('target_milestone');
  my @products         = $q->param('product');
  my @layered_products = $q->param('layered_products');
  my @priorities       = $q->param('priorities');
  my $user_logic       = $self->qpu_str('user_logic', 'any');

  # determine what columns to display
  my %view;
  my @columns    = $self->my_columns(\%view);
  my @verified   = $q->param('verified');
  my $login_join = $self->qpu_str('user_type', 'assigned_to');

  # allow the passing of arbitrary bug lists

  my @bug_id = $q->param('bug_id') ? $q->param('bug_id') : ();
  @bug_id = split(/[, ]+/, $bug_id[0]) if @bug_id == 1 && $bug_id[0] =~ /[, ]/;
  my $depends = $self->qpu_int('dependencies', 0);

  # fill template forms

  $self->bug_define_form($vars) if !@bug_id;
  $self->bug_filter_form($vars, {'bug_status' => \@bug_status});
  $self->bug_view_form($vars);

  # link for custom view configuration

  my %param = (page => 'edit-columns');
  for my $k ($q->param) {
    next if ($k eq 'page' || $k eq 'clear_default');
    my @v = $q->param($k);
    $param{$k} = \@v if @v;
  }

  $param{id} = 'releases/ack_edit_columns.html';
  my $uri = URI->new;
  $uri->query_form(%param);
  $vars->{column_edit} = $uri;

  # set template variables

  $vars->{hide_update}  = 1 if @bug_id;
  $vars->{bug_id}       = join(',', @bug_id);
  $vars->{dependencies} = $depends;
  $vars->{image_path}   = $self->{config}->{IMAGE_PATH};
  $vars->{contact}      = $self->{config}->{CONTACT};
  $vars->{edit_mode}    = 1 if $edit_mode;
  $vars->{allow_edit}   = 1 if $allow_edit;
  $vars->{ack_or}       = 1 if $ack_or;
  $vars->{fix_tick}     = 1
    if $view_version || ($list eq $self->{config}->{TYPE_ACK});
  $vars->{no_blocker} = 1 if $view_version;

  # if no update supplied, only show the options header
  if (!$update) {
    $vars->{bugs}        = 0;
    $vars->{header_only} = 1;
    ## BUGBUG WTF is this
    #   print $template->{SERVICE}->{CONTEXT}
    #   ->process( 'bug_list_header.tmpl', $vars );
    return;
  }

  # retrieve approved components
  my %approved_component;
  my %capacity_component;
  $self->approved_component_list($update, \%approved_component,
    \%capacity_component);

  my $bz_search = Bugzilla::Extension::Releases::ACKCustomSearch->new();

  # generate sql query
  my $sql_query = $self->list_query({
    'update' => !@bug_id ? $update : undef,
    'column' => !@bug_id ? $column : undef,
    'view_version'               => $view_version,
    'list'                       => $list,
    'type'                       => $type,
    'combined'                   => $combined,
    'sort_by'                    => $sort_by,
    'view'                       => \%view,
    'member'                     => $member,
    'group'                      => $selected_groups,
    'group_negate'               => $group_negate,
    'external_tracker'           => $external_tracker,
    'customer_mode'              => $customer_mode,
    'partner_mode'               => $partner_mode,
    'partners'                   => \@partners,
    'component'                  => $component,
    'component_not'              => $component_not,
    'priorities'                 => \@priorities,
    'bug_status'                 => \@bug_status,
    'verified'                   => \@verified,
    'layered_products'           => \@layered_products,
    'approved_components'        => $approved_components,
    'capacity_components'        => $capacity_components,
    'unassigned_components'      => $unassigned_components,
    'approved_component_list'    => \%approved_component,
    'capacity_component_list'    => \%capacity_component,
    'flag_filter'                => \%acks,
    'keyword_filter'             => $keyword_filter,
    'keyword_logic'              => $keyword_logic,
    'user_logic'                 => $user_logic,
    'conditional_naks'           => \@bug_naks,
    'qe_conditional_naks'        => \@bug_qe_naks,
    'target_milestone'           => \@bug_targets,
    'products'                   => \@products,
    'blocks_filter'              => $blocks_filter,
    'blocks_depth'               => $blocks_depth,
    'bug_id'                     => \@bug_id,
    'dependencies'               => $depends,
    'ack_or'                     => $ack_or,
    'hide_z_proposed'            => $hide_z_proposed,
    'hide_z_approved'            => $hide_z_approved,
    'login_join'                 => $login_join,
    'bz_search'                  => $bz_search,
    'cf_internal_target_release' => \@cf_internal_target_release,
  });

  if (!$sql_query) {
    $vars->{bugs} = 0;
    return;
  }

  # bugzilla advanced query form
  $bz_search->set();
  my $advanced_query = $bz_search->get_query_params();
  my @query_form     = ();
  foreach my $key (keys %{$advanced_query}) {
    my %param_data;
    $param_data{'name'}  = $key;
    $param_data{'value'} = $advanced_query->{$key};
    push(@query_form, \%param_data);
  }
  $vars->{form_params} = \@query_form;

  # link makers
  my $bzlink = $self->{config}->{BZ_LINK};
  $vars->{bugzilla_link} = sub { my $bugid = shift; sprintf $bzlink, $bugid };

  my $buglist_base = $bzlink;
  $buglist_base =~ s/[^\/]+\.cgi\?.*$/buglist.cgi/;
  my $query_base = $bzlink;
  $query_base =~ s/[^\/]+\.cgi\?.*$/query.cgi/;

  $vars->{buglist_link} = sub {
    my $bug_list = shift;
    return $buglist_base . '?' . 'bug_id=' . $bug_list;
  };
  $vars->{query_link}
    = sub { my $query = shift; return $buglist_base . '?' . $query };

  # query data

  my $dbh = Bugzilla->dbh();
  Bugzilla->logger->info($sql_query);
  my $sth_bugs = $dbh->prepare($sql_query);
  $sth_bugs->execute();

  # generate advanced query link and get url length and max length info

  my $max_uri_length = $self->{config}->{'MAX_URL_LENGTH'} || 8000;
  my $uri_length     = length($q->self_url());
  my $query_text     = $bz_search->get_query_text();
  my $query_length   = length($buglist_base . '?' . $query_text);

  $vars->{bugquery} = $query_text
    if $uri_length <= $max_uri_length && $query_length <= $max_uri_length;
  $vars->{bugs} = $sth_bugs->rows();

  # prepare sub-queries
  my %sth;
  $sth{blocked}     = $view{blocked}     ? $self->prepare_blocked()     : undef;
  $sth{alias}       = $view{blocked}     ? $self->prepare_aliases()     : undef;
  $sth{dependson}   = $view{dependson}   ? $self->prepare_dependson()   : undef;
  $sth{component}   = $view{component}   ? $self->prepare_components()  : undef;
  $sth{product}     = $view{product}     ? $self->prepare_products()    : undef;
  $sth{external}    = $view{external}    ? $self->prepare_external()    : undef;
  $sth{description} = $view{description} ? $self->prepare_description() : undef;
  $sth{keywords}    = $self->prepare_keywords();
  $sth{profiles}    = $self->prepare_profiles();
  $sth{flags}       = $self->prepare_flags();
  $sth{releases}    = $self->prepare_releases();

  foreach my $custom_field (grep($_ =~ /^cf_/ && $view{$_}, @columns)) {
    $sth{$custom_field} = $self->prepare_custom($custom_field);
  }

  my @all_rows;
  my %all_assignees;
  my @buglist   = ();
  my $row_count = 0;

  while (my $row = $sth_bugs->fetchrow_hashref) {

    my $parity = (++$row_count % 2 == 0) ? 'bz_even' : 'bz_odd';

    my $bug_id = $self->process_row(
      $row,
      {
        'update'                     => $update,
        'view_version'               => $view_version,
        'list'                       => $list,
        'type'                       => $type,
        'partner_mode'               => $partner_mode,
        'approved_component_list'    => \%approved_component,
        'capacity_component_list'    => \%capacity_component,
        'sth'                        => \%sth,
        'fields'                     => \@columns,
        'view'                       => \%view,
        'acks'                       => \%acks,
        'edit_mode'                  => $edit_mode,
        'can_grant'                  => $can_grant,
        'can_edit'                   => \%can_edit,
        'external_tracker'           => $external_tracker,
        'external_tracker_exclusive' => $external_tracker_exclusive,
        'parity'                     => $parity,
        'all_rows'                   => \@all_rows,
        'all_assignees'              => \%all_assignees,
      }
    );

    push @buglist, $bug_id if $bug_id;
  }

  # get assignees group
  my $group_and_member_mappings = {};
  if (%all_assignees) {
    my @emails = (keys %all_assignees);
    $group_and_member_mappings = []; ## BUGBUG agile groups $self->orgchart()->get_group( { 'email' => \@emails } );
  }

  # head

  if ($edit_mode) {
    $vars->{blank} = %can_edit ? 0 : 1;
  }

  # body
  foreach my $row_item (@all_rows) {
    my $bug_id = $row_item->{row}{bug_id};
    $row_item->{group_and_member_mappings} = $group_and_member_mappings;
  }

  $vars->{rows} = \@all_rows;

  # footer

  $vars->{buglist}  = join(',', @buglist);
  $vars->{bugs}     = $#buglist + 1;
  $vars->{bugquery} = $query_text
    if $uri_length <= $max_uri_length && $query_length <= $max_uri_length;

  if ($uri_length > $max_uri_length) {

    # require redirect
    $vars->{action_buglist} = 'bugzilla-metrics.cgi';
    $vars->{action_query}   = 'bugzilla-metrics.cgi';

    my @buglist_params = (
      {'name' => 'page',         'value' => 'redirect'},
      {'name' => 'redirect_url', 'value' => $buglist_base}
    );
    my @query_params = (
      {'name' => 'page',         'value' => 'redirect'},
      {'name' => 'redirect_url', 'value' => $query_base},
      {'name' => 'format',       'value' => 'advanced'}
    );

    $vars->{buglist_params} = \@buglist_params;
    $vars->{query_params}   = \@query_params;
  }
  else {
    $vars->{action_buglist} = $buglist_base;
    $vars->{action_query}   = $query_base;
  }

  return;
}

sub flag_bugs_submit {
  my $self   = shift;
  my $vars   = shift;
  my $q      = Bugzilla->cgi();
  my $action = untaint($q->param('action'));
  my $update = untaint($q->param('update'));
  my $logger = Bugzilla->logger;

  # check what flags this user may grant
  my @ack_flags = @{$self->{config}->{ACK_FLAGS}};
  my $can_grant = $self->can_grant_acks($update, \@ack_flags);

  $can_grant || ThrowUserError("ackview_cant_grant");

  $vars->{title} = $self->title;

  my %seen;      # to record which bugs have been seen from cgi args
  my %report;    # to record which bugs have been processed by xmlrpc
  my $bugs_updated = 0;
  my $bugs_failed  = 0;

  # loop thru all cgi params pulling out flag updates
  foreach my $key ($q->param) {
    my $error  = 0;
    my $bug_id = undef;
    my %bug_report;

    if ($key =~ /^(\d+)_(\w+)$/) {
      $bug_id = $1 if (grep($2, @ack_flags) || $2 eq 'edit' || $2 eq "cf_pm_score");
    }

    # is it a valid parameter, has this bug already been processed?
    next if !$bug_id || $seen{$bug_id};

    # nope, record bug and start looking for flags to update
    $seen{$bug_id} = {};
    my $bug = Bugzilla::Bug->check($bug_id);

    # determine which flags are to be updated
    my @flags_to_upd;
    foreach my $ack_flag (@ack_flags) {
      next if !$can_grant->{$ack_flag};

      my $varname = $bug_id . '_' . $ack_flag;    # cgi form name for flag
      my $flag    = $self->qpu_str($varname);
      if ($flag and $flag ne '?') {               # only want + or -
        my $id;
      FLAG_LOOP:
        foreach my $type (@{$bug->flag_types}) {
          foreach my $flag (@{$type->{flags}}) {
            if ($flag->id && $flag->name eq $ack_flag) {
              $id = $flag->id;
              last FLAG_LOOP;
            }
          }
        }
        push @flags_to_upd, {id => $id, status => $flag};
      }
    }

    # record which flags (if any) are about to be updated
    if (@flags_to_upd) {

      # if not [Submit], do not update flags
      if (!$action || $action ne 'Submit') {
        $report{$bug_id}{flags_updated} = \@flags_to_upd;
      }
      else {
        $bug_report{$bug_id}{flags_updated} = \@flags_to_upd;

        eval {
          $bug->set_flags(\@flags_to_upd, []);
          $bug_report{$bug_id}{flags}{result} = 'OK';    ##$ret;    # success
        };
        if (my $e = $@) {
          $bug_report{$bug_id}{flags}{error} = $e;       # fail
          $error++;
        }
      }
    }

    # Update pm score
    my $is_pm            = is_bugzilla_group('product_management');
    my $pm_score         = $self->qpu_str($bug_id . '_cf_pm_score', 0);
    my $pm_score_default = $self->qpu_str($bug_id . '_cf_pm_score_default', 0);
    if ($is_pm && $pm_score =~ /\d+/ && $pm_score ne $pm_score_default) {
      my $values_to_upd = {cf_pm_score => $pm_score};

      if (!$action || $action ne 'Submit') {
        $report{$bug_id}{values_updated} = $values_to_upd;
      }
      else {
        $bug_report{$bug_id}{values_updated} = $values_to_upd;
        eval {
          my $field = new Bugzilla::Field({name => 'cf_pm_score'});
          $bug->set_custom_field($field, $pm_score);
          $bug_report{$bug_id}{values}{result} = 'OK';    # success
        };
        if (my $e = $@) {
          $bug_report{$bug_id}{values}{error} = $e;       # fail
          $bug_report{$bug_id}{values}{error}
            .= "\n" . Data::Dumper::Dumper($values_to_upd);
          $error++;
        }
      }
    }

    eval { $bug->update(); };

    if (my $e = $@) {
      $bug_report{$bug_id}{values}{error} = $e;           # fail
      $error++;
    }

    if ($key =~ /edit$/ && !defined $report{$bug_id}) {
      $report{$bug_id} = ();
    }

    if (%bug_report) {
      $bugs_failed++  if ($error);
      $bugs_updated++ if (!$error);
      $report{$bug_id} = $bug_report{$bug_id};
    }

  }

  if ($action && $action eq 'Submit') {
    if (!$bugs_updated && !$bugs_failed) {
      $vars->{submit_message} = "releases_no_changes";
    }
    else {
      $vars->{submit_message} = '';
      $vars->{submit_message}
        .= "$bugs_updated bug"
        . ($bugs_updated > 1 ? "s" : "")
        . " updated successfully.<br />"
        if $bugs_updated;
      $vars->{submit_message}
        .= "Failed to update $bugs_failed bug"
        . ($bugs_failed > 1 ? "s" : "")
        . ".<br />"
        if $bugs_failed;
    }

    $self->flag_bugs($vars);

    return;
  }

  return $self->add_additional_data($vars, \%report);
}

sub print_report {
  my $self       = shift;
  my $template   = shift;
  my $bug_report = shift;

  my $item_vars;

  # convert report to an array
  my @item_report = map { {bug_id => $_, %{$bug_report->{$_}}} }
    sort { $a <=> $b } keys %{$bug_report};

  $item_vars->{submit_report} = \@item_report;

  my $bzlink = $self->{config}->{BZ_LINK};

  $item_vars->{bugzilla_link}
    = sub { my $bugid = shift; sprintf $bzlink, $bugid };

  print $template->{SERVICE}->{CONTEXT}
    ->process('bug_update_item.tmpl', $item_vars);

  return;
}

sub add_additional_data {
  my $self    = shift;
  my $vars    = shift;
  my $ack_ref = shift;
  my $q       = Bugzilla->cgi();
  my $release = $self->qpu_str('update');
  my $dbh     = Bugzilla->dbh();

  return $self->show_flag_bugs() if !$ack_ref || !keys %{$ack_ref};

  # retrieve list of flags compatible with given update release

  my $sql_inclusions = <<INCLUSION_QUERY;
        SELECT  DISTINCT n.name
          FROM  flagtypes c,
                flaginclusions i,
                flaginclusions m,
                flagtypes n
          WHERE c.name = '$release'
            AND c.id = i.type_id
            AND i.product_id = m.product_id
            AND m.type_id = n.id
            AND n.is_active = 1
INCLUSION_QUERY

  my $sth = $dbh->prepare($sql_inclusions);
  $sth->execute();

  my %compatible;

  while (my ($flag_name) = $sth->fetchrow_array) {
    $compatible{$flag_name} = 1;
  }

  $sth->finish();

  # retrieve all relevant release flags

  my %release_flags;
  my @releases = @{get_releases() || []};
  foreach my $release_flag (@releases) {
    $release_flags{$release_flag} = 1
      if $release_flag ne $release && $compatible{$release_flag};
  }

  # define other relevant flags

  my @ack_flags = @{$self->{config}->{ACK_FLAGS}};
  my @flags     = @ack_flags;
  push @flags, ('blocker', 'exception', 'fast', $release);
  my @all_flags = keys %release_flags;
  push @all_flags, @flags;
  my $flags_where = join(' OR ', map {"flagtypes.name='$_'"} @all_flags);
  my $bugs_where  = join(' OR ', map {"bug_id=$_"} keys %{$ack_ref});

  $vars->{title}          = 'Edit ' . $release;
  $vars->{target_runmode} = 'flag-bugs-submit-complex';
  $vars->{update_release} = $release;
  $vars->{bugzilla_link}
    = sub { my $id = shift; sprintf $self->{config}->{BZ_LINK}, $id };

  # set parameter context

  my @all_params = $q->param();
  my @hidden     = ();
  foreach my $param (@all_params) {
    next
      if $param eq 'page'
      || $param eq 'action'
      || $param eq 'bug_status'
      || $param =~ /^\d+_/;
    my %param_data;
    $param_data{'name'}  = $param;
    $param_data{'value'} = $self->qpu_str($param);
    push(@hidden, \%param_data);
  }
  my @selected = $q->param('bug_status');
  foreach my $select (@selected) {
    my %param_data;
    $param_data{'name'}  = 'bug_status';
    $param_data{'value'} = $select;
    push(@hidden, \%param_data);
  }
  $vars->{'hidden_params'} = \@hidden if @hidden;

  # set edit permissions

  my $sql_grant = <<GRANT_QUERY;
        SELECT flagtypes.name,
               groups.name
          FROM flagtypes,
               groups
         WHERE flagtypes.grant_group_id=groups.id
           AND ( $flags_where )
GRANT_QUERY

  $sth = $dbh->prepare($sql_grant);
  $sth->execute();

  my $grant = $self->can_grant_acks($release, \@ack_flags);

  while (my ($flag_name, $grant_group) = $sth->fetchrow_array) {

    # set flag permission

    if ($flag_name eq $release) {    # current release flag
      $vars->{'allow_release'} = 1 if is_bugzilla_group($grant_group);
    }
    elsif ($release_flags{$flag_name}) {    # potential new release flag
      delete $release_flags{$flag_name} if !is_bugzilla_group($grant_group);
    }
    elsif ($flag_name !~ /_ack$/) {         # other flag, except ack flag
      $vars->{'allow_' . $flag_name} = 1 if is_bugzilla_group($grant_group);
    }
  }
  $sth->finish();

  my $is_devel = is_bugzilla_group('devel');
  my $is_pm    = is_bugzilla_group('product_management');

  $vars->{'allow_edit'} = 1 if is_bugzilla_group('redhat');

  $vars->{'allow_devel_whiteboard'} = 1 if $is_devel;

  $vars->{'allow_edit_comment'} = 1 if Bugzilla->user->can_edit_comments();

  $vars->{'allow_private_comment'} = 1 if is_bugzilla_group('private_comment');

  # devel group allow to view pm_score
  # product_management allow to view and edit pm_score

  my $allow_pm_score
    = ($is_devel && !$is_pm) ? "view" : ($is_pm) ? "edit" : undef;
  $vars->{'allow_pm_score'} = $allow_pm_score;

  my @ack_flag_filter = ();
  foreach my $ack_flag (@ack_flags) {
    push(@ack_flag_filter, {name => $ack_flag, allow => $grant->{$ack_flag},});
  }
  $vars->{'ack_flags'} = \@ack_flag_filter;

  # set new releases

  my @new_releases = ();
  push @new_releases, {'name' => ""};
  foreach my $new_release (sort keys %release_flags) {
    push(@new_releases, {'name' => $new_release});
  }
  $vars->{'release_flags'} = \@new_releases if @new_releases > 1;

  # generate flags cache

  my $sql_flags = <<FLAG_QUERY;
        SELECT flags.bug_id,
               flagtypes.name,
               flags.status
          FROM flags,
               flagtypes
         WHERE flags.type_id=flagtypes.id
           AND ( $bugs_where )
           AND ( $flags_where )
FLAG_QUERY

  $sth = $dbh->prepare($sql_flags);
  $sth->execute();

  my %flags;
  while (my ($bug_id, $name, $status) = $sth->fetchrow_array) {
    $flags{$bug_id}{$name} = $status;
  }
  $sth->finish();

  # retrieve approved components

  $sth = $dbh->prepare(<<APPROVED_COMPONENTS);
            SELECT component_id,
                   component_type
              FROM rh_release_components, rh_release, flagtypes
             WHERE flagtypes.id=rh_release.type_id
               AND rh_release.id=rh_release_components.release_id
               AND flagtypes.name='$release'
APPROVED_COMPONENTS

  $sth->execute();

  my %release_component;

  while (my ($component_id, $component_type) = $sth->fetchrow_array) {
    $release_component{$component_id} = $component_type;
  }
  $sth->finish();

  # get open bug states

  my @bug_states = ();
  $sth
    = $dbh->prepare(
    "SELECT value FROM bug_status WHERE isactive=1 AND value!='CLOSED' ORDER BY sortkey"
    );
  $sth->execute();

  while (my ($value) = $sth->fetchrow_array) {
    push @bug_states, $value;
  }
  $sth->finish();

  # process bug list

  my $sql_bugs = <<BUG_QUERY;
        SELECT bug_id,
               bug_status,
               cf_devel_whiteboard,
               cf_pm_score,
               components.id,
               components.name
          FROM bugs, components
         WHERE bugs.component_id=components.id
           AND ( $bugs_where )
BUG_QUERY

  $sth = $dbh->prepare($sql_bugs);
  $sth->execute();

  my @rows = ();
  while (
    my (
      $bug_id,   $bug_status,   $devel_whiteboard,
      $pm_score, $component_id, $component_name
    )
    = $sth->fetchrow_array
    )
  {
    my %row_data;

    $row_data{bug_id}     = $bug_id;
    $row_data{bug_closed} = $bug_status eq 'CLOSED';

    my @row_ack_flags;
    foreach my $ack_flag (@ack_flags) {
      my %ack_data;
      $ack_data{'name'}    = $ack_flag;
      $ack_data{'allow'}   = $grant->{$ack_flag};
      $ack_data{'default'} = $flags{$bug_id}{$ack_flag};

      foreach my $up_flag (@{$ack_ref->{$bug_id}{flags_updated}}) {
        if ($up_flag->{name} eq $ack_flag) {
          $ack_data{'values'} = $up_flag->{status};
          last;
        }
      }
      $ack_data{'values'} = $flags{$bug_id}{$ack_flag} if (!$ack_data{'values'});

      push(@row_ack_flags, \%ack_data);
    }

    $row_data{row_ack_flags} = \@row_ack_flags;

    $row_data{release_default} = $flags{$bug_id}{$release};

    $row_data{blocker_default} = $flags{$bug_id}{blocker};

    $row_data{exception_default} = $flags{$bug_id}{exception};

    $row_data{fast_default} = $flags{$bug_id}{fast};

    $row_data{devel_whiteboard_default} = $devel_whiteboard if $devel_whiteboard;

    if ($allow_pm_score) {
      my %pm_score;
      my $updated_vals = $ack_ref->{$bug_id}{values_updated} || undef;

      $pm_score{value}   = $pm_score || 0;
      $pm_score{default} = $pm_score || 0;

      if ( $allow_pm_score eq 'edit'
        && defined $updated_vals
        && defined $updated_vals->{cf_pm_score})
      {
        $pm_score{value} = $updated_vals->{cf_pm_score};
      }

      $row_data{pm_score} = \%pm_score;
    }

    $row_data{bug_status_default} = $bug_status;

    if ($bug_status ne 'CLOSED') {
      my @states = ();
      foreach my $status (@bug_states) {
        my %param_data;
        $param_data{'name'}     = $status;
        $param_data{'value'}    = $status;
        $param_data{'selected'} = 1 if $status eq $bug_status;
        push(@states, \%param_data);
      }

      $row_data{bug_states} = \@states;
    }

    $row_data{component_name} = $component_name;
    $row_data{component_status}
      = $release_component{$component_id}
      ? $release_component{$component_id}
      : "undef";

    push(@rows, \%row_data);
  }
  $sth->finish();

  $vars->{rows} = \@rows;

  return;    ## BUGBUG $self->template_process( 'add_data.tmpl', $template_vars );
}

sub flag_bugs_submit_complex {
  my $self      = shift;
  my $vars      = shift;
  my $q         = Bugzilla->cgi();
  my $release   = $self->qpu_str('update');
  my @ack_flags = @{$self->{config}->{ACK_FLAGS}};
  my @flags     = @ack_flags;
  push @flags, ('blocker', 'exception', 'fast');
##    my $template = BugzillaMetrics->template;

##    $self->template_pre_process($vars);
  $vars->{title} = $self->title;
  print Bugzilla->cgi->header(-content_type => 'text/html');
##    print $template->{SERVICE}->{CONTEXT}        ->process( 'bugs_frame_header.tmpl', $vars );
##    print $template->{SERVICE}->{CONTEXT}        ->process( 'bug_update_header.tmpl', {} );

  my @updates;
  my %seen;    # to record which bugs have been seen from cgi args
  my $bugs_updated = 0;
  my $bugs_failed  = 0;

  # loop thru all cgi params pulling out flag and bug updates
  foreach my $key ($q->param) {
    my $bug_id = undef;
    if ($key =~ /^(\d+)_(\w+)$/) {
      $bug_id = $1 if grep($2, @ack_flags);
    }

    # is it a valid parameter, has this bug already been processed?
    next if !$bug_id || $seen{$bug_id};

    # nope, record bug and start looking for flags to update
    $seen{$bug_id} = {};
    my %report;    # to record results for bug processed by xmlrpc
    $report{$bug_id} = {};
    my $bug_ok = 1;

    # determine which flags are to be updated

    my @flags_to_upd;
    foreach my $flag (@flags) {
      my $varname    = $bug_id . '_' . $flag;
      my $flag_value = $self->qpu_str($varname, 'X');
      $varname .= '_default';
      my $flag_default = $self->qpu_str($varname, 'X');

      if ($flag_value ne $flag_default) {
        push @flags_to_upd, {name => $flag, status => $flag_value};
      }
    }

    # check release flag

    my $varname    = $bug_id . '_release';
    my $flag_value = $self->qpu_str($varname, 'X');
    $varname .= '_default';
    my $flag_default = $self->qpu_str($varname, 'X');

    if ($flag_value && (!$flag_default || $flag_value ne $flag_default)) {
      my $update_release = $self->qpu_str('update');
      push @flags_to_upd, {name => $update_release, status => $flag_value}
        if $update_release;
    }

    # check new release flag

    my $new_release_name  = $self->qpu_str($bug_id . '_new_release_name',  '');
    my $new_release_value = $self->qpu_str($bug_id . '_new_release_value', 'X');

    if ($new_release_name ne '' && $new_release_value ne 'X') {
      push @flags_to_upd, {name => $new_release_name, status => $new_release_value};
    }

    # require flag update?

    if (@flags_to_upd) {
      $report{$bug_id}->{flags_updated} = \@flags_to_upd;
      $report{$bug_id}->{flags}         = {};

      eval {
        my $ret = xmlrpc_update_flags($bug_id, \@flags_to_upd);
        $report{$bug_id}->{flags}->{result} = $ret;    # success
      };
      if (my $e = $@) {
        $report{$bug_id}->{flags}->{error} = $e;       # fail
        $report{$bug_id}->{flags}->{error}
          .= "\n" . Data::Dumper::Dumper(\@flags_to_upd);
        $bug_ok = 0;
      }
    }

    # check bug values

    my %values_to_upd;

    # devel whiteboard updated?

    my $default_value = $self->qpu_str($bug_id . '_devel_whiteboard_default', '');
    my $new_value     = $self->qpu_str($bug_id . '_devel_whiteboard',         '');

    if ($new_value ne $default_value) {
      $values_to_upd{cf_devel_whiteboard} = $new_value;
    }

    # pm score updated?
    my $is_pm                  = is_bugzilla_group('product_management');
    my $pm_score_default_value = $self->qpu_str($bug_id . '_pm_score_default', '0');
    my $pm_score_new_value     = $self->qpu_str($bug_id . '_pm_score', '0');

    if ($is_pm && $pm_score_new_value ne $pm_score_default_value) {
      $values_to_upd{cf_pm_score} = $pm_score_new_value;
    }

    # check bug status

    $varname = $bug_id . '_bug_status';
    my $status_value = $self->qpu_str($varname);
    $varname .= '_default';
    my $status_default = $self->qpu_str($varname);

    if ($status_value && (!$status_default || $status_value ne $status_default)) {
      $values_to_upd{status} = $status_value;
    }

    # comment added?

    my $comment = $q->param($bug_id . '_comment');

    if ($comment) {
      my %comment_hash;
      $comment_hash{body}       = $comment;
      $comment_hash{is_private} = 1
        if $self->qpu_int($bug_id . '_comment_private', 0);

      $values_to_upd{comment} = \%comment_hash;
    }

    if (%values_to_upd) {
      $report{$bug_id}->{values_updated} = \%values_to_upd;
      $report{$bug_id}->{values}         = {};

      eval {
        my $ret = xmlrpc_update_values($bug_id, \%values_to_upd);
        $report{$bug_id}->{values}->{result} = $ret;    # success
      };
      if (my $e = $@) {
        $report{$bug_id}->{values}->{error} = $e;       # fail
        $report{$bug_id}->{values}->{error}
          .= "\n" . Data::Dumper::Dumper(\%values_to_upd);
        $bug_ok = 0;
      }
    }

    my $item_vars;

    # convert report to an array
    my @item_report
      = map { {bug_id => $_, %{$report{$_}}} } sort { $a <=> $b } keys %report;
    $item_vars->{submit_report} = \@item_report;
    my $bzlink = $self->{config}->{BZ_LINK};
    $item_vars->{bugzilla_link}
      = sub { my $bugid = shift; sprintf $bzlink, $bugid };
##        print $template->{SERVICE}->{CONTEXT}            ->process( 'bug_update_item.tmpl', $item_vars );

    if (@flags_to_upd || %values_to_upd) {
      if ($bug_ok) {
        $bugs_updated++;
      }
      else {
        $bugs_failed++;
      }
    }
  }

  if (!$bugs_updated && !$bugs_failed) {
    $vars->{submit_message} = "No changes made.";
  }
  else {
    $vars->{submit_message} = '';
    $vars->{submit_message}
      .= "$bugs_updated bug"
      . ($bugs_updated > 1 ? "s" : "")
      . " updated successfully.<br />"
      if $bugs_updated;
    $vars->{submit_message}
      .= "Failed to update $bugs_failed bug"
      . ($bugs_failed > 1 ? "s" : "")
      . ".<br />"
      if $bugs_failed;
  }
##    print $template->{SERVICE}->{CONTEXT}        ->process( 'bug_update_footer.tmpl', $vars );

  $self->flag_bugs();

##    print $template->{SERVICE}->{CONTEXT}->process( 'footer.html.tmpl', {} );

  return;
}

sub user_columns {
  my ($self) = @_;

  return () if !Bugzilla->user || !Bugzilla->user->id;

  my $dbh = Bugzilla->dbh();
  my $sth = $dbh->prepare(
    "SELECT value FROM rh_release_ackview_settings WHERE userid=?");
  $sth->execute(Bugzilla->user->id);
  my ($view) = $sth->fetchrow_array();

  return () if !$view;

  return Load($view);
}

sub set_user_columns {
  my ($self, $view, $clear_default) = @_;

  return 0 if !Bugzilla->user || !Bugzilla->user->id;

  my $dbh = Bugzilla->dbh;

  my $sth_value = $dbh->prepare(
    "SELECT value FROM rh_release_ackview_settings WHERE userid=?");
  $sth_value->execute(Bugzilla->user->id);

  my ($value) = $sth_value->fetchrow_array();

  if (($value && (!$view || !@$view)) || $clear_default) {
    my $sth
      = $dbh->prepare("DELETE FROM rh_release_ackview_settings WHERE userid=?");
    $sth->execute(Bugzilla->user->id);

    $sth_value->execute(Bugzilla->user->id);

    ($value) = $sth_value->fetchrow_array();

    return $value ? 0 : 1;
  }

  return 0 if (!$view || !@$view);

  my @newview = map { untaint($_) } @$view;

  my $new_value = Dump(\@newview);

  return -1 if $value && $value eq $new_value;

  if (!$value) {
    my $sth = $dbh->prepare(
      "INSERT INTO rh_release_ackview_settings (userid,value) VALUES (?,?)");
    $sth->execute(Bugzilla->user->id, $new_value);
  }
  else {
    my $sth = $dbh->prepare(
      "UPDATE rh_release_ackview_settings SET value=? WHERE userid=?");
    $sth->execute($new_value, Bugzilla->user->id);
  }

  $sth_value->execute(Bugzilla->user->id);

  ($value) = $sth_value->fetchrow_array();

  return $value eq $new_value ? 1 : 0;
}

sub edit_columns {
  my ($self, $vars) = @_;

  my $q      = Bugzilla->cgi();
  my @hidden = (
    {name => 'page', value => 'commit-columns'},
    {name => 'id',   value => 'releases/ack_list.html'}
  );
  for my $k ($q->param) {
    next
      if $k eq 'page' || $k eq 'id' || $k eq 'show' || $k eq 'hide' || $k =~ /^c_/;
    my @v = $q->param($k);
    foreach my $i (@v) {
      push(@hidden, {name => $k, value => $i});
    }
  }

  my %show;
  my @columns = $self->my_default_columns(\%show);

  my @view_columns;
  foreach my $column (@columns) {
    push(@view_columns,
      {name => $self->_header($column), value => $column, show => $show{$column}});
  }

  $vars->{columns} = \@view_columns;
  $vars->{hidden}  = \@hidden;
  return Bugzilla->template->process('edit_columns.tmpl', $vars);
}

sub commit_columns {
  my ($self, $vars) = @_;

  my $q    = Bugzilla->cgi();
  my @show = $q->param('show');

  my $clear_default = $q->param('clear_default') || '';
  my $result        = $self->set_user_columns(\@show, $clear_default);

  $vars->{title} = $q->param('bug_id') ? 'custom' : $self->title;
  $vars->{message}
    = !$result    ? "releases_columns_update_failed"
    : $result < 0 ? "releases_no_changes"
    :               "releases_columns_updated";

  $self->flag_bugs($vars);

  return;
}

# qpu - Bugzilla->cgi->param($name) and then untaint
# shorthand help routines
# string
sub qpu_str {
  my ($self, $name, $default) = @_;

  my $value = $self->{input}->{$name};

  if (defined $value) {
    if ($value =~ /^([_\/\w\d\. @()&*%+-?"]+)$/) {
      return $1;
    }
  }

  return $default if defined $default;
  return;
}

# date
sub qpu_date {
  my ($self, $name, $default) = @_;

  my $day   = $self->qpu_int("$name\_day");
  my $month = $self->qpu_int("$name\_month");
  my $year  = $self->qpu_int("$name\_year");

  if (  ($day and $day >= 1 and $day <= 31)
    and ($month and $month >= 1  and $month <= 12)
    and ($year  and $year > 1990 and $year < 2020))
  {
    return sprintf "%4d%02d%02d", $year, $month, $day;

  }

  return $default if defined $default;
  return;
}

#  integer
sub qpu_int {
  my ($self, $name, $default) = @_;

  my $value = $self->{input}->{$name};

  if (defined $value) {
    if ($value =~ /^(\d+)$/) {
      return $1;
    }
  }

  return $default if defined $default;
  return;
}

# boolean
sub qpu_bool {
  my ($self, $name, $default) = @_;

  my $int = $self->qpu_int($name, $default);

  return $int ? 1 : 0;
}

# flag values
sub qpu_flag {
  my ($self, $name, $default) = @_;

  my $value = $self->{input}->{$name};

  if (defined $value) {
    if ($value =~ /^(!?[+-?X]+)$/) {
      return $1;
    }
  }

  return $default if defined $default;
  return;
}

# Check whether a user is allow to edit ack flags
sub can_grant_acks {
  my $self    = shift;
  my $release = shift || undef;
  my $flags   = shift || [];
  my %can_grant;
  my $user = Bugzilla->user();

  if ($release && @{$flags}) {
    my $grant_groups = $self->get_grant_group($release, $flags);
    foreach my $ack (keys %$grant_groups) {
      if ($user->in_group($grant_groups->{$ack}{'group_name'})) {
        $can_grant{$ack} = 1;
      }
    }
  }
  return \%can_grant;
}

# Get grant groups for all the ack flags
sub get_grant_group {
  my $self    = shift;
  my $release = shift || undef;
  my $flags   = shift || [];

  # Bugzilla allows duplicate flag name. For example, there are
  # multiple devel_ack flags in bugzilla but each of them is used by
  # different products. To determine which ack flags to be used for a
  # release flag, we need to map the products of the release flag with
  # the products of the ack flags.

  if (!$release || !@{$flags}) {
    return {};
  }

  my $dbh = Bugzilla->dbh();

  # Get product inclusion list for the release
  my $includes
    = $self->get_flag_clusions({'flag_name' => $release, 'type' => 'include'});

  # Get product exclusion list for the release
  my $excludes
    = $self->get_flag_clusions({'flag_name' => $release, 'type' => 'exclude'});

  # Make sure the included products are not in the
  # exclude list
  foreach my $product (keys %{$includes}) {
    if (exists $excludes->{$product}) {
      delete $includes->{$product};
    }
  }

  my @sub_query;
  push @sub_query, "i.product_id IN (" . join(',', keys %{$includes}) . ")"
    if (%{$includes});
  push @sub_query, "i.product_id IS NULL";

  my $extra_sql = join(' OR ', @sub_query);

  my $mask = '?,' x @{$flags};
  chop($mask);

  my $ack_incl_sql = <<SQL;
       SELECT
           f.id AS flag_id,
           f.name AS flag_name,
           g.name AS group_name,
           i.product_id AS product_id
         FROM flaginclusions AS i
           LEFT JOIN flagtypes AS f ON f.id=i.type_id
           LEFT JOIN groups AS g ON g.id = f.grant_group_id
         WHERE f.name IN ($mask) AND ($extra_sql)
SQL

  my $ack_incl = $dbh->selectall_arrayref($ack_incl_sql, {}, @{$flags});

  my %grant_groups;
  my %ack_excludes;
  foreach my $row (@{$ack_incl}) {
    my $flag_id    = $row->[0];
    my $flag_name  = $row->[1];
    my $group_name = $row->[2];
    my $product_id = $row->[3];

    next if $grant_groups{$flag_name};

    # Get all products that are excluded by the flag
    if (!defined $ack_excludes{$flag_id}) {
      $ack_excludes{$flag_id}
        = $self->get_flag_clusions({'flag_id' => $flag_id, 'type' => 'exclude'});
    }

    my $all_products = $self->get_all_products();
    my $all_excludes = $ack_excludes{$flag_id};

    # Skip if both inclusion and exlucsion list = all products.
    if (!defined $product_id && %{$all_products} == %{$all_excludes}) {
      next;
    }

    # Skip if a product is in both inclusion and exclusion list
    elsif (exists $ack_excludes{$flag_id}->{$product_id}) {
      next;
    }

    $grant_groups{$flag_name}{'group_name'} = $group_name;
  }

  return \%grant_groups;
}

# Get products that are included or excluded by a flag
# and return their ids.
sub get_flag_clusions {
  my $self = shift;
  my $args = shift || undef;

  my $logger = Bugzilla->logger;

  if (!$args || ref $args ne 'HASH') {
    $logger->logdie('get_clusions requires a hash as argument.');
  }

  my $flag_name = delete $args->{'flag_name'} || undef;
  my $flag_id   = delete $args->{'flag_id'}   || undef;
  my $type      = delete $args->{'type'}      || 'include';
  my $where     = '';
  my $value     = '';
  my $table     = 'flaginclusions';

  if (!defined $flag_name && !defined $flag_id) {
    $logger->logdie('get_clusions requires either a flag name or id as argument.');
  }
  elsif (defined $flag_name) {
    $where = 'f.name = ?';
    $value = $flag_name;
  }
  elsif (defined $flag_id) {
    $where = 'f.id = ?';
    $value = $flag_id;
  }

  if ($type eq 'exclude') {
    $table = 'flagexclusions';
  }

  my $include_sql = <<SQL;
       SELECT DISTINCT
           i.product_id AS product_id
         FROM $table AS i
           LEFT JOIN flagtypes AS f ON f.id=i.type_id
         WHERE $where
SQL

  my $dbh  = Bugzilla->dbh();
  my $rows = $dbh->selectcol_arrayref($include_sql, {}, $value);

  # If flaginclusions/fagexclusions table returns a NULL row,
  # that means the flag is include/exclude all products.
  my %clusions;
  if (@$rows == 1 && !defined $rows->[0]) {
    map { $clusions{$_} = 1 } keys %{$self->get_all_products()};
  }
  else {
    map { $clusions{$_} = 1 } @{$rows};
  }

  return \%clusions;
}

# Get all products in bugzilla and return their ids.
sub get_all_products {
  my $self = shift;

  if (!defined $self->{'cached'}{'all_products'}) {

    # Get all products
    my $products_sql = <<SQL;
            SELECT id,
                   name
              FROM products
            WHERE isactive = 1
SQL
    my $dbh = Bugzilla->dbh();
    $self->{'cached'}{'all_products'}
      = $dbh->selectall_hashref($products_sql, 'id');
  }

  return $self->{'cached'}{'all_products'};
}

sub is_bugzilla_group {
  my $group = shift || undef;
  my $user  = Bugzilla->user;

  if (defined $user && defined $group && $user->in_group($group)) {
    return 1;
  }

  return 0;
}

# A function for sorting hash in template toolkit
sub sort_hash_by {
  my $self = shift;

  return sub {
    my $hash       = shift;
    my $sort_field = shift;
    return (sort(
      { $hash->{$a}->{$sort_field} <=> $hash->{$b}->{$sort_field} } keys %$hash));
  };
}

#################################################
sub has_member {
  my $member = shift;
  my @array  = @_;

  my $found = 0;
  $found = any { $_ eq $member } @array if (@array);

  return $found;
}

sub get_description {
  my ($self, $name) = @_;

  my $dbh = Bugzilla->dbh();
  my $sth = $dbh->prepare("SELECT description FROM fielddefs WHERE name = ?");
  $sth->execute($name);

  my ($description) = $sth->fetchrow_array;

  return $description || $name;
}

# Get the sort key of a priority or bug severity
sub get_sort_key {
  my ($self, $value, $type) = @_;
  my $list = {};

  if ($type eq 'priority') {
    $list = $self->get_priorities();
  }
  else {
    $list = $self->get_severities();
  }

  return $list->{$value}{'sortkey'} || 0;
}

# Get all type of priorities in bugzilla
sub get_priorities {
  my $self = shift;

  if (!defined $self->{'cached'}{'priorities'}) {
    my $dbh            = Bugzilla->dbh();
    my $priorities_sql = "SELECT value, sortkey FROM priority WHERE isactive = 1";

    $self->{'cached'}{'priorities'}
      = $dbh->selectall_hashref($priorities_sql, 'value');
  }

  return $self->{'cached'}{'priorities'};
}

# Get all type of bug severities in bugzilla
sub get_severities {
  my $self = shift;

  if (!defined $self->{'cached'}{'severities'}) {
    my $dbh = Bugzilla->dbh();
    my $severities_sql
      = "SELECT value, sortkey FROM bug_severity WHERE isactive = 1";

    $self->{'cached'}{'severities'}
      = $dbh->selectall_hashref($severities_sql, 'value');
  }

  return $self->{'cached'}{'severities'};
}

1;    # Magic true value required at end of module
__END__

=head1 NAME

Bugzilla::Extension::Releases::ACKList - [One line description of module's purpose here]


=head1 VERSION

This document describes Bugzilla::Extension::Releases::ACKList version 0.0.1


=head1 SYNOPSIS

    use Bugzilla::Extension::Releases::ACKList;

=for author to fill in:
    Brief code example(s) here showing commonest usage(s).
    This section will be as far as many users bother reading
    so make it as educational and exeplary as possible.


=head1 DESCRIPTION

=for author to fill in:
    Write a full description of the module and its features here.
    Use subsections (=head2, =head3) as appropriate.


=head1 INTERFACE 

=for author to fill in:
    Write a separate section listing the public components of the modules
    interface. These normally consist of either subroutines that may be
    exported, or methods that may be called on objects belonging to the
    classes provided by the module.


=head1 DIAGNOSTICS

=for author to fill in:
    List every single error and warning message that the module can
    generate (even the ones that will "never happen"), with a full
    explanation of each problem, one or more likely causes, and any
    suggested remedies.

=over

=item C<< Error message here, perhaps with %s placeholders >>

[Description of error here]

=item C<< Another error message here >>

[Description of error here]

[Et cetera, et cetera]

=back


=head1 CONFIGURATION AND ENVIRONMENT

=for author to fill in:
    A full explanation of any configuration system(s) used by the
    module, including the names and locations of any configuration
    files, and the meaning of any environment variables or properties
    that can be set. These descriptions must also include details of any
    configuration language used.

Bugzilla::Extension::Releases::ACKList requires no configuration files or environment variables.


=head1 DEPENDENCIES

=for author to fill in:
    A list of all the other modules that this module relies upon,
    including any restrictions on versions, and an indication whether
    the module is part of the standard Perl distribution, part of the
    module's distribution, or must be installed separately. ]

None.


=head1 INCOMPATIBILITIES

=for author to fill in:
    A list of any modules that this module cannot be used in conjunction
    with. This may be due to name conflicts in the interface, or
    competition for system or program resources, or due to internal
    limitations of Perl (for example, many modules that use source code
    filters are mutually incompatible).

None reported.


=head1 BUGS AND LIMITATIONS

=for author to fill in:
    A list of known problems with the module, together with some
    indication Whether they are likely to be fixed in an upcoming
    release. Also a list of restrictions on the features the module
    does provide: data types that cannot be handled, performance issues
    and the circumstances in which they may arise, practical
    limitations on the size of data sets, special cases that are not
    (yet) handled, etc.

No bugs have been reported.

Please report any bugs or feature requests to
C<engineering-services@redhat.com>, or through the web interface at
L<https://engineering.redhat.com/rt3>.


=head1 AUTHOR

Bernd Groh  C<< <bgroh@redhat.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2007, Red Hat, Inc. All rights reserved.



=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
