# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::FlagGroupDetail;

use strict;
use warnings;
use 5.10.1;

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::FlagType;
use Bugzilla::Util;
use Bugzilla::Error;

use Bugzilla::Extension::RuleEngine::FlagGroup;

use base qw(Bugzilla::Object);

use JSON;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'rh_flag_group_detail';
use constant LIST_ORDER => 'creation_ts, id';

use constant DB_COLUMNS => qw(
  id
  flag_group_id
  creation_ts
  creator_id
  theregexp
);

# Nothing gets updated. Ever.
use constant UPDATE_COLUMNS => qw( );

use constant VALIDATORS => {
  creation_ts   => \&_check_creation_ts,
  creator_id    => \&_check_creator_id,
  theregexp     => \&_check_theregexp,
  flag_group_id => \&_check_flag_group_id,
};

###############################
####      Validators       ####
###############################

sub _check_creation_ts {
  return Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
}

sub _check_creator_id {
  my $invocant = shift;

  # The creator is never changed
  return ref $invocant ? $invocant->creator->id : Bugzilla->user->id;
}

sub _check_theregexp {
  my ($invocant, $regexp) = @_;
  $regexp = trim($regexp);

  eval {qr/$regexp/};
  if ($@) {
    my $err = $@;
    $err =~ s/ at [\/\.\w]+ line \d+\.\s+$//;
    ThrowUserError('rh_invalid_regexp', {regexp => $regexp, msg => $err});
  }

  return $regexp;
}

sub _check_flag_group_id {
  my ($invocant, $id) = @_;
  $id = trim($id);

  Bugzilla::Extension::RuleEngine::FlagGroup->check({id => $id});

  return $id;
}

###############################
####       Methods         ####
###############################

sub create {
  my $class  = shift;
  my $params = shift;
  my $dbh    = Bugzilla->dbh;

  my $flag_ids = delete $params->{flag_ids};
  my $self     = $class->SUPER::create($params);

  # Now add the flag type mappings
  my $query = q{
        INSERT INTO rh_flag_group_map (flag_group_detail_id, flag_type_id)
        VALUES (?, ?)
    };

  foreach my $flag_id (@$flag_ids) {
    $dbh->do($query, undef, $self->id, $flag_id);
  }

  return $self;
}

sub flag_group {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{flag_group}) {
    $self->{flag_group}
      = Bugzilla::Extension::RuleEngine::FlagGroup->new($self->flag_group_id);
  }
  return $self->{flag_group};
}

sub creator {
  my ($self) = @_;
  return $self->{creator} if exists $self->{creator};
  $self->{creator} = new Bugzilla::User($self->creator_id);
  return $self->{creator};
}

###############################
####      Accessors        ####
###############################

sub flag_group_id { return $_[0]->{flag_group_id}; }
sub creation_ts   { return $_[0]->{creation_ts}; }
sub creator_id    { return $_[0]->{creator_id}; }
sub theregexp     { return $_[0]->{theregexp}; }

sub flag_types {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{flag_types}) {
    my $ids = $dbh->selectcol_arrayref(
      q{
            SELECT flag_type_id FROM rh_flag_group_map
            WHERE flag_group_detail_id = ?}, undef, $self->id
    );
    $self->{flag_types} = Bugzilla::FlagType->new_from_list($ids);
  }

  return $self->{flag_types};
}

sub flag_ids {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{flag_ids}) {
    $self->{flag_ids} = $dbh->selectcol_arrayref(
      q{
            SELECT flag_type_id FROM rh_flag_group_map
            WHERE flag_group_detail_id = ?}, undef, $self->id
    );
  }

  return @{$self->{flag_ids}};
}

###############################
####       Helpers         ####
###############################

1;

__END__

=head1 NAME

Bugzilla::Extension::RuleEngine::FlagGroupDetail - Bugzilla Flag Group Detail class.

=head1 SYNOPSIS

    use Bugzilla::Extension::RuleEngine::FlagGroupDetail;

    my $flag_group_detail = new Bugzilla::Extension::RuleEngine::FlagGroupDetail(1);

    my $id          = $rule_group->id;
    my $flag_group  = $flag_group_detail->flag_group;
    my $creation_ts = $flag_group_detail->creation_ts;
    my $creator     = $flag_group_detail->creator;
    my $theregexp   = $flag_group_detail->theregexp;
    my $flag_types  = $flag_group_detail->flag_types;

=head1 DESCRIPTION

Bugzilla::Extension::RuleEngine::FlagGroupDetail represents a version of a
flag group object. It is an implementation of L<Bugzilla::Object>, and thus
provides all methods that L<Bugzilla::Object> provides.

The methods that are specific to L<Bugzilla::Extension::RuleEngine::FlagGroupDetail>
are listed below.

I<theregexp> is a Perl regular expression string used to include any matching
flags.

e.g. to create a flag group that would include all flags that start with
C<bugzilla_> you could set this to:

  ^bugzilla_.*

=head1 METHODS

=over

=item C<creator>

=over

=item B<Description>

The person that created this version of the flag group.

=item B<Params>

none.

=item B<Returns>

A Bugzilla::User object.

=back

=item C<flag_group>

=over

=item B<Description>

Returns the flag_group that this detail belongs to.

=item B<Params>

none.

=item B<Returns>

A Bugzilla::Extension::RuleEngine::FlagGroup object.

=back

=item C<flag_types>

=over

=item B<Description>

Returns the flag types for this flag group detail

=item B<Params>

none.

=item B<Returns>

An arrayref of Bugzilla::FlagType object.

=back

=item C<flag_ids>

=over

=item B<Description>

Returns the flag type ids for this flag group detail

=item B<Params>

none.

=item B<Returns>

An array of integers (representing Bugzilla::FlagType's).

=back

=back

=cut
