# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::RuleState;

use strict;
use warnings;
use 5.10.1;

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Mailer;

use List::MoreUtils qw(uniq);

use base qw(Bugzilla::Object);

###############################
####    Initialization     ####
###############################

use constant DB_TABLE => 'rh_rule_state';
use constant LIST_ORDER =>
  'CASE WHEN when_enabled IS NULL THEN 0 ELSE 1 END, id';

use constant KILLSWITCH_ON_TMPL  => 'email/rule-killswitch-enabled.txt.tmpl';
use constant KILLSWITCH_OFF_TMPL => 'email/rule-killswitch-disabled.txt.tmpl';

use constant DB_COLUMNS => qw(
  id
  disabled_by
  when_disabled
  reason
  enabled_by
  when_enabled
  rule_id
);

use constant UPDATE_COLUMNS => qw( );

use constant VALIDATORS =>
  {reason => \&_check_reason, when_disabled => \&_check_when_disabled,};

###############################
####      Validators       ####
###############################

sub _check_when_disabled {
  return Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
}

sub _check_reason {
  my ($invocant, $reason) = @_;

  $reason = trim($reason);
  $reason || ThrowUserError('rule_state_not_specified');

  return $reason;
}

###############################
####       Methods         ####
###############################

sub create {
  my $class = shift;

  my $self = $class->SUPER::create(@_);

  $self->_killswitch_notify(KILLSWITCH_ON_TMPL);

  return $self;
}

sub enable {
  my $self = shift;

  my $query = q{
        UPDATE rh_rule_state
        SET when_enabled = NOW(), enabled_by = ?
        WHERE id = ?
    };

  Bugzilla->dbh->do($query, undef, Bugzilla->user->id, $self->id);

  $self->_killswitch_notify(KILLSWITCH_OFF_TMPL);

  # Reload myself
  $self = $self->new($self->id);

  return;
}

sub disabled_user {
  my ($self) = @_;
  return $self->{disabled_user} if exists $self->{disabled_user};
  $self->{disabled_user} = new Bugzilla::User($self->disabled_by);
  return $self->{disabled_user};
}

sub enabled_user {
  my ($self) = @_;
  return $self->{enabled_user} if exists $self->{enabled_user};
  $self->{enabled_user} = new Bugzilla::User($self->enabled_by);
  return $self->{enabled_user};
}

###############################
####      Accessors        ####
###############################

sub disabled_by   { return $_[0]->{'disabled_by'}; }
sub when_disabled { return $_[0]->{'when_disabled'}; }
sub reason        { return $_[0]->{'reason'}; }
sub enabled_by    { return $_[0]->{'enabled_by'}; }
sub when_enabled  { return $_[0]->{'when_enabled'}; }
sub rule_id       { return $_[0]->{'rule_id'}; }

sub rule {
  my $self = shift;

  return $self->{'rule'} if exists $self->{'rule'};

  if (defined $self->rule_id) {
    $self->{'rule'} = Bugzilla::Extension::RuleEngine::Rule->new($self->rule_id);
  }
  else {
    $self->{'rule'} = undef;
  }

  return $self->{'rule'};
}

###############################
####       Helpers         ####
###############################

sub active_list {
  my $class = shift;
  my $dbh   = Bugzilla->dbh;

  my $query
    = "SELECT id FROM rh_rule_state WHERE when_enabled IS NULL ORDER BY id";
  my $ids = $dbh->selectcol_arrayref($query);

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  my @objects;

  foreach my $obj (@{$class->new_from_list($ids)}) {
    if ( (!$obj->rule && $user->is_insider)
      || ($obj->rule && $obj->rule->rule_group->user_can_view($user)))
    {
      push(@objects, $obj);
    }
  }

  return (\@objects);
}

sub previous_list {
  my $class = shift;
  my $dbh   = Bugzilla->dbh;

  my $query = q{
        SELECT id
        FROM rh_rule_state
        ORDER BY CASE WHEN when_enabled IS NULL THEN 0 ELSE 1 END,
            when_enabled DESC,
            when_disabled DESC
        LIMIT 10
    };

  my $ids = $dbh->selectcol_arrayref($query);

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  my @objects;

  # We cannot use new_from_list, as this messes with the ordering
  foreach my $obj (map { $class->new($_) } @$ids) {
    if ( (!$obj->rule && $user->is_insider)
      || ($obj->rule && $obj->rule->rule_group->user_can_view($user)))
    {
      push(@objects, $obj);
    }

  }

  return (\@objects);
}

sub disabled_count {
  my ($self, $rule_group_id) = @_;

  my $count;
  my $query = "SELECT COUNT(*) FROM rh_rule_state WHERE enabled_by IS NULL";

  trick_taint($rule_group_id) if ($rule_group_id);

  if ($rule_group_id) {
    $query .= ' AND rule_id IN (SELECT id FROM rh_rule WHERE rule_group_id = ?)';
    $count = Bugzilla->dbh->selectrow_array($query, undef, $rule_group_id);
  }
  else {
    $query .= ' AND rule_id IS NULL';
    $count = Bugzilla->dbh->selectrow_array($query);
  }


  return $count;
}

# Send notifications if the kill switch was tripped.
# The maintainers and the rule owners (if any) will
# be notified.
sub _killswitch_notify {
  my $self          = shift;
  my $template_path = shift;

  my (@owners, @admins);
  if (defined $self->rule_id) {
    @owners = map { $_->login } @{$self->rule->rule_owners}
      if ($self->rule->rule_owners);
    @admins = map { $_->login } @{$self->rule->rule_group->maintainers}
      if ($self->rule->rule_group->maintainers);

  }

  push(@admins, Bugzilla->params->{mail_errors_to});

  my $timestamp = Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
  $timestamp = format_time($timestamp, '%a, %d %b %Y %T %z', 'UTC');

  my $rule         = $self->rule;
  my $creator      = undef;
  my $global_count = $self->disabled_count;
  my $group_count  = 0;

  # A list of email addresses to send notificaitons to
  my @email_addr = uniq(@owners, @admins);

  if (defined $rule) {

    # Sometimes a rule is not involved in a kill switch being set.
    # If it is, identify who created the rule.
    $creator = Bugzilla::User->new($self->rule->current->creator_id);
    push @email_addr, $creator->email;
    $group_count = $self->disabled_count($rule->rule_group_id);
  }

  foreach my $email (uniq(@email_addr)) {

    # Since these aren't going to users, we will send it in the sites language
    my $default_lang = Bugzilla::User->new()->setting('lang');
    my $template     = Bugzilla->template_inner($default_lang);

    my $message;
    my %vars = (
      kill_reason  => $self->reason,
      date         => $timestamp,
      user         => Bugzilla->user,
      to           => $email,
      group_count  => $group_count,
      global_count => $global_count,
    );

    if (defined $rule) {
      $vars{creator} = $creator;
      $vars{rule}    = $rule;
    }

    $template->process($template_path, \%vars, \$message)
      || ThrowTemplateError($template->error());

    MessageToMTA($message);
  }

  return;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::RuleEngine::RuleState - Bugzilla Rule State class.

=head1 SYNOPSIS

    use Bugzilla::Extension::RuleEngine::RuleState;

    my $rule_state = new Bugzilla::Extension::RuleEngine::RuleState(1);

    my $id            = $rule_state->id;
    my $when_disabled = $rule_state->when_disabled;
    my $disabled_by   = $rule_state->disabled_by;
    my $reason        = $rule_state->reason;
    my $when_enabled  = $rule_state->when_enabled;
    my $enabled_by    = $rule_state->enabled_by;


=head1 DESCRIPTION

Bugzilla::Extension::RuleEngine::RuleState represents an entry in the Rules
Engine state table. If there are any rows where when_enabled is not set, the
Rules Engine will not be run.  It is an implementation of L<Bugzilla::Object>,
and thus provides all methods that L<Bugzilla::Object> provides.

The methods that are specific to L<Bugzilla::Extension::RuleEngine::RuleState> are listed
below.

Note that the meaning of disabled and enable is in relation to the Rules
Engine itself, not the state of the kill switch (which is the opposite)

=head1 METHODS

=over

=item C<enable>

=over

=item B<Description>

Changes the switch to enabled. This is used to ensure both
enabled values are set correctly (since there is not set_ subs)

=item B<Params>

none.

=item B<Returns>

nothing.

=back

=item C<disabled_user>

=over

=item B<Description>

The person that disabled the Rules Engine.

=item B<Params>

none.

=item B<Returns>

A Bugzilla::User object.

=back

=item C<enabled_user>

=over

=item B<Description>

The person that enabled the Rules Engine.

=item B<Params>

none.

=item B<Returns>

A Bugzilla::User object.

=back

=item C<active_list>

=over

=item B<Description>

Returns all state rows which result in the BRE being disabled

=item B<Params>

none.

=item B<Returns>

An arrayref of L<Bugzilla::Extension::RuleEngine::RuleState> objects.

=back

=item C<previous_list>

=over

=item B<Description>

Returns the ten most recent state rows (currently active first)

=item B<Params>

none.

=item B<Returns>

An arrayref of L<Bugzilla::Extension::RuleEngine::RuleState> objects.

=back

=item C<disabled_count>

=over

=item B<Description>

Returns the number of rows where the Rules Engine is disabled

=item B<Params>

=over

=item C<$rule_group_id> (optional).

The RuleGroup id to limit the count to.

=back

=item B<Returns>

An integer.

=back

=back

=cut
