# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SAML2Auth;

use 5.10.1;
use strict;
use warnings;

use List::Util qw(first);

use parent qw(Bugzilla::Extension);
use Bugzilla;
use Bugzilla::Constants;
use File::Copy;

# This code for this is in ../extensions/SAML2Auth/lib/Util.pm
use Bugzilla::Extension::SAML2Auth::Util;
use Bugzilla::Extension::SAML2Auth::IDP;

our $VERSION = '0.01';

BEGIN {
# Monkey-patch can() on Bugzilla::Auth::Login::CGI so that our own fail_nodata gets called
# if SAML2Auth is enabled in user_info_class.
# Our fail_nodata behaves like CGI's, so this shouldn't be a problem for CGI-based logins.
  *Bugzilla::Auth::Login::CGI::can = sub {
    my ($stack, $method) = @_;
    if (Bugzilla->params->{user_info_class} !~ /\SAML2Auth\b/
      || $method ne 'fail_nodata')
    {
      return $stack->SUPER::can($method);
    }
    return;
  };
}

# See the documentation of Bugzilla::Hook ("perldoc Bugzilla::Hook"
# in the bugzilla directory) for a list of all available hooks.
sub install_update_db {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  if (!$dbh->bz_column_info('saml2auth_idp', 'ldap_mail_attrs')) {
    $dbh->bz_drop_table('saml2auth_idp');
    $dbh->bz_add_table('saml2auth_idp');
  }

  return;
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  $schema->{saml2auth_idp} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      name       => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      cacert     => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      icon       => {TYPE => 'LONGTEXT',     NOTNULL => 1, DEFAULT => "'ca-user'"},
      url        => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      userregexp => {TYPE => 'LONGTEXT',     NOTNULL => 1, DEFAULT => "''"},
      metadata   => {TYPE => 'LONGTEXT',     NOTNULL => 1, DEFAULT => "''"},
      expires    => {TYPE => 'DATETIME',     NOTNULL => 1, DEFAULT => "NOW()"},
      comment    => {TYPE => 'LONGTEXT',     NOTNULL => 1, DEFAULT => "''"},
      isactive   => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'},
      saml_extern_id_attr => {TYPE => 'LONGTEXT', NOTNULL => 1, DEFAULT => "''"},
      ldap_extern_id_attr => {TYPE => 'LONGTEXT', NOTNULL => 1, DEFAULT => "''"},
      ldap_mail_attrs     => {TYPE => 'LONGTEXT', NOTNULL => 1, DEFAULT => "''"},
    ],
    INDEXES =>
      [saml2auth_idp_name_idx => {FIELDS => [qw(name)], TYPE => 'UNIQUE'},],
  };

  # IDP groups specify who is NOT allowed to use it!
  $schema->{saml2auth_idp_groups} = {
    FIELDS => [
      idp_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'saml2auth_idp', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      saml2auth_idp_groups_idx => {FIELDS => [qw(idp_id group_id)], TYPE => 'UNIQUE'},
    ],
  };

  return;
}

sub install_filesystem {
  my ($self, $args) = @_;
  my $files = $args->{'files'};

  my $extensionsdir = bz_locations()->{'extensionsdir'};

  foreach my $script (
    qw(saml2_login.cgi saml2_logout.cgi saml2_acs.cgi saml2_metadata.cgi))
  {
    my $scriptname = $extensionsdir . "/" . __PACKAGE__->NAME . "/cgi/$script";

# Have to hack this as the pathing is *^%*$ so you can't have cgi scripts in sub dirs
    if (ON_WINDOWS) {
      copy($scriptname, $script);
      $files->{$script} = {perms => Bugzilla::Install::Filesystem::WS_EXECUTE};
    }
    else {
      $files->{$scriptname} = {perms => Bugzilla::Install::Filesystem::WS_EXECUTE};
      symlink($scriptname, $script);
    }
  }

  return;
}

sub template_before_create {
  my ($self, $args) = @_;

  return if Bugzilla->user->id && !Bugzilla->cgi->param('logout');

  $args->{config}{VARIABLES}{saml2auth} = {
    idps => sub {
      return (Bugzilla::Extension::SAML2Auth::IDP->match({isactive => 1}));
    },
  };

  return;
}

sub config_modify_panels {
  my ($self, $args) = @_;
  my $auth_panel_params = $args->{panels}{auth}{params};

  my $user_info_class
    = first { $_->{name} eq 'user_info_class' } @$auth_panel_params;
  if ($user_info_class) {
    push @{$user_info_class->{choices}}, "SAML2Auth,CGI";
  }

  my $user_verify_class
    = first { $_->{name} eq 'user_verify_class' } @$auth_panel_params;
  if ($user_verify_class) {
    unshift @{$user_verify_class->{choices}}, "SAML2Auth";
  }

  return;
}

sub config_add_panels {
  my ($self, $args) = @_;
  my $modules = $args->{panel_modules};
  $modules->{SAML2Auth} = "Bugzilla::Extension::SAML2Auth::Params";

  return;
}

sub auth_login_methods {
  my ($self, $args) = @_;
  my $modules = $args->{'modules'};
  if (exists $modules->{'SAML2Auth'}) {
    $modules->{'SAML2Auth'} = 'Bugzilla/Extension/SAML2Auth/Login.pm';
  }

  return;
}

## Is this required?
sub auth_verify_methods {
  my ($self, $args) = @_;
  my $modules = $args->{'modules'};
  if (exists $modules->{'SAML2Auth'}) {
    $modules->{'SAML2Auth'} = 'Bugzilla/Extension/SAML2Auth/Verify.pm';
  }

  return;
}

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};
  my $cgi  = Bugzilla->cgi;
  my $p    = $cgi->Vars;

  # If we aren't viewing a SAML page, we don't need to do anything
  return unless ($page =~ m{^saml2auth/});

  if ($page eq 'saml2auth/settings.html') {
    my $action = $cgi->param('action') || "";
    my $params;
    foreach my $key (keys(%$p)) {
      next if ($key eq 'id');
      $params->{$key} = $p->{$key};
    }
    delete($params->{action});

    if ($action eq 'create') {
      $params->{name} = delete($params->{idp_name});
      my $groups = delete $params->{groups};
      my $idp    = Bugzilla::Extension::SAML2Auth::IDP->create($params);
      $idp->set_groups($groups);
      $idp->update();
    }
    elsif ($action eq 'save') {
      $params->{isactive} = 0 unless defined $params->{isactive};
      my $id = delete($params->{idp_id});
      $params->{name} = delete($params->{idp_name});
      my $groups = delete $params->{groups};
      my $idp    = Bugzilla::Extension::SAML2Auth::IDP->new($id);
      $idp->set_all($params);
      $idp->set_groups($groups);
      $idp->update();
    }
    elsif ($action eq 'delete') {
      my $id  = delete($params->{idp_id});
      my $idp = Bugzilla::Extension::SAML2Auth::IDP->new($id);
      $idp->remove_from_db();
    }

    $vars->{'groups'} = [Bugzilla::Group->get_all];
    $args->{vars}->{idps} = Bugzilla::Extension::SAML2Auth::IDP->match();
    $args->{vars}->{'doc_section'}
      = 'extensions/SAML2Auth/index-admin.html#bugzilla-extension-saml2auth-identity-providers';

  }
  else {
    ThrowUserError('saml_no_such_page');
  }

  return;
}

sub persist_login {
  my ($self, $args) = @_;

  my $cgi        = $args->{'cgi'};
  my %cookieargs = %{$args->{'cookieargs'}};
  $cookieargs{'-expires'} = 'Fri, 01-Jan-2038 00:00:00 GMT';

  if (defined(Bugzilla->request_cache->{saml2_idp_name})) {
    $cgi->send_cookie(
      -name  => 'Bugzilla_saml_choice',
      -value => Bugzilla->request_cache->{saml2_idp_name},
      %cookieargs
    );
    delete Bugzilla->request_cache->{saml2_idp_name};
  }

  return;
}

__PACKAGE__->NAME;

__END__

=head1 Name

Bugzilla::Extension::SAML2Auth - SAML2 authentication support for Bugzilla

=head1 Version

Version 0.01

=head1 Description

This module will allow you to set up your Bugzilla server as an SP and configure
multiple IDPs to authenticate against.


=head1 Configuration and Environment

To enable SAML2 authentication you will need to configure your Service Provider and select SAML2Auth in both C<user_info_class> and C<user_verify_class> on the user authentication panel in the parameters UI.

This module uses OpenSSL to validate certificates and XML signatures, you will save yourself a B<lot> of pain if 
you ensure all CAs are imported in to your system certifcate chain. If you are using self signed certificates 
then you should sue one self signed certficate as a CA, imprt that,a nd use it to sign certficates for your SP and IDPs.

=head2 Service Provider

To configure your Bugzilla instance as a Service Provider (SP) you will need paths to a valid key, a valid cert, and a cacert, in data/params.json.

The defaults for these fields are for FHS Linux paths https://en.wikipedia.org/wiki/Filesystem_Hierarchy_Standard, you may need to change them for other Linux distros and other O/Ss.
You may also need to change these if you haven't imported your CA in to your systems certificate chain. There are cross
platform instructions on doing this at
L<http://kb.kerio.com/product/kerio-connect/server-configuration/ssl-certificates/adding-trusted-root-certificates-to-the-server-1605.html>

The metadata URL for your SP will be C<$SITE_BASE_URL/saml2_metadata.cgi>

=head2 Identity Providers

Identity Providers (IDPs) are configured on a new Administration page, you will need to be able to verify the IDPs certificates,
it is easiest to do this by importing the necessary CA's in the same way you did for your SP's certificates.
The defaults are again based on FHS Linux paths. The full list of field descriptions is available in L<Bugzilla::Extension::SAML2Auth::IDP>

If the IDP provides full and correct metadata you can add the URL to it in the IDP administration page
and it will be fetched and stored in the Database. Some IDPs do not provide full and correct metadata, and some
do not provide a URL for their metadata. For these services you will need to add a recored on the administration
page and then import the metadata XML directly using the provided C<import-metadata.pl> script.

=head1 Author

Jeff Fearn <jfearn@redhat.com>

