# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::WebService::GroupReview;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Object;
use Bugzilla::Extension::Workflows::GroupReview;

use constant PUBLIC_METHODS => qw(
  get_group
  start_reviews
);


=head1 NAME

Bugzilla::Extension::Workflows::WebService::GroupReview

=head1 DESCRIPTION

This part of the Bugzilla API allows you to manages group membership reviews.

=head1 METHODS

=head2 get_group

Lists direct members in a group.

You will need to have Bless privileges on the group named to be able to perform
this action.

Users in the admin group will not be listed.

Users who are in the I<insiders> group will only be listed for users who are
also in the I<insiders> group.

=head3 Params

group:
  B<string> The name of the group to fetch.

min_days:
  B<int> I<optional> The minimum numbers of days since the user last logged in.
  Users who have logged in more recently than this will not be listed.

=head3 Returns

Returns an array named I<members> containing hashes, one per member, with the
following fields:

days_since_login:
  The number of days since the user last logged in.

id:
  The users unique ID.

login:
  The users login mail address.

review_days:
  The number of days the users review bug has been open.

review_id:
  The bug id of a review for this user in this group.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "GroupReview.get_group",
      "id": 1,
      "params": [
        {
          "name": "thegroup",
          "min_days": 21
        }
      ]
  }

=head4 JSONRPC reply

  {
      "id": 1,
      "error": null,
      "result": {
          "members": [
            {
              "days_since_login": 138,
              "id": 11,
              "login": "user1@example.com"
              "review_days": "",
              "review_id": "",
            }, {
              "days_since_login": 63
              "id": 22,
              "login": "user42@example.com",
              "review_days": "31",
              "review_id": "123456",
            }
          ]
      }
  }


=cut

sub get_group {
  my ($self, $args) = validate(@_);

  return {members => get_group_membership($args)};
}


=head2 start_reviews

Start the group reviews for users.

You will need to have Bless privileges on the group named to be able to perform
this action.

Users in the admin group will not be processed by this API and will be silently
discarded from the user list.

Users who are in the I<insiders> group can only be processed by users who are
also in the I<insiders> group. Insiders supplied by outsiders will be silently
discarded from the user list.

=head3 Params

group:
  B<string> The name of the group to process.

ids:
  B<array> of B<int> The IDs of the users to review. User IDs that are not current
  members of the group will be ignored.

=head3 Returns

Returns an array named I<members> containing hashes, one for each member updated,
with the following fields:

days_since_login:
  The number of days since the user last logged in.

id:
  The users unique ID.

login:
  The users login mail address.

review_days:
  The number of days the users review bug has been open.

review_id:
  The bug id of a review for this user in this group.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "GroupReview.start_reviews",
      "id": 1,
      "params": [
        {
          "name": "thegroup",
          "ids": [ 11, 22, 42 ]
        }
      ]
  }

=head4 JSONRPC reply

  {
      "id": 1,
      "error": null,
      "result": {
          "members": [
            {
              "days_since_login": 138,
              "id": 11,
              "login": "user11@example.com"
              "review_days": "0",
              "review_id": "123457",
            }, {
              "days_since_login": 63
              "id": 22,
              "login": "user22@example.com",
              "review_days": "31",
              "review_id": "123456",
            }, {
              "days_since_login": 11
              "id": 42,
              "login": "user42@example.com",
              "review_days": "0",
              "review_id": "123458",
            }
          ]
      }
  }


=cut

sub start_reviews {
  my ($self, $args) = validate(@_);

  return {members => start_group_membership_reviews($args)};
}

1;

__END__


